/*
 *        spCanvas.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spContainer.h>

#include <sp/spDraw.h>
#include <sp/spImage.h>
#include <sp/spPluginCanvasP.h>
#include <sp/spSliderP.h>
#include <sp/spCanvasP.h>

static spParamTable sp_canvas_param_tables[] = {
    {SppBorderOn, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.border_on), SP_FALSE_STRING},
    {SppFocusable, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.focusable), SP_FALSE_STRING},
    {SppDrawBackground, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.draw_background), SP_FALSE_STRING},
    {SppUseTabKey, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.use_tab_key), SP_FALSE_STRING},
    {SppUseArrowKey, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.use_arrow_key), SP_FALSE_STRING},
    {SppDrawFunc, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spPrimitive, primitive.draw_func), NULL},
    {SppDrawData, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spPrimitive, primitive.draw_data), NULL},

    {SppAcceptWheelEvent, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.accept_wheel_event), SP_FALSE_STRING},
    {SppWheelTargetHorizontal, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.wheel_target_horizontal), SP_FALSE_STRING},
    {SppScrollLeftByWheelDown, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.scroll_left_by_wheel_down), SP_FALSE_STRING},
        
    {SppScrollHorizontal, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_scroll), SP_FALSE_STRING},
    {SppHorizontalValue, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_value), "0"},
    {SppHorizontalMinimum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_minimum), "0"},
    {SppHorizontalMaximum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_maximum), "100"},
    {SppHorizontalPageSize, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_page_size), "100"},
    {SppHorizontalIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_increment), "10"},
    {SppHorizontalPageIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_page_increment), "100"},
    {SppHorizontalWheelIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_wheel_increment), "5"},
    {SppHorizontalContentFactor, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.h_content_factor), "1.0"},

    {SppScrollVertical, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_scroll), SP_FALSE_STRING},
    {SppVerticalValue, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_value), "0"},
    {SppVerticalMinimum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_minimum), "0"},
    {SppVerticalMaximum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_maximum), "100"},
    {SppVerticalPageSize, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_page_size), "100"},
    {SppVerticalIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_increment), "10"},
    {SppVerticalPageIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_page_increment), "100"},
    {SppVerticalWheelIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_wheel_increment), "5"},
    {SppVerticalContentFactor, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.v_content_factor), "1.0"},

    {SppWheelSubstitutiveButton, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spCanvas, canvas.wheel_substitutive_button), SP_RBUTTON_STRING},
};

spCanvasClassRec SpCanvasClassRec = {
    /* spObjectClassPart */
    {
        SpCanvas,
        (spObjectClass)&SpPrimitiveClassRec,
        sizeof(spCanvasRec),
        spArraySize(sp_canvas_param_tables),
        sp_canvas_param_tables,
        spCanvasPartInit,
        spCanvasPartFree,
        SP_FALSE,
        NULL,
        NULL,
        spCanvasCreate,
        spCanvasDestroy,
        spCanvasSetParams,
        NULL,
    },
    /* spComponentClassPart */
    {
        SP_TRUE,
        SP_FALSE,
        SP_TRUE,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,

        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
        NULL,
    },
    /* spPrimitiveClassPart */
    {
        0,
    },
    /* spCanvasClassPart */
    {
        spCanvasRefresh,
        spCanvasRedraw,
    },
};

spComponentClass SpCanvasClass = (spComponentClass)&SpCanvasClassRec;

void spCanvasPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpComponentPart(component).orientation = SP_VERTICAL;
    SpComponentPart(component).margin_width = SP_DEFAULT_MARGIN;
    SpComponentPart(component).margin_height = SP_DEFAULT_MARGIN;
    SpComponentPart(component).spacing = SP_DEFAULT_SPACING;
    
    SpCanvasPart(component).prev_width = 0;
    SpCanvasPart(component).prev_height = 0;
    SpCanvasPart(component).capture_flag = SP_FALSE;
    SpCanvasPart(component).exposed = SP_FALSE;
    SpCanvasPart(component).refreshed = SP_FALSE;
    SpCanvasPart(component).size = 0;

    SpCanvasPart(component).h_scroll_coef = 1;
    SpCanvasPart(component).v_scroll_coef = 1;
    
    return;
}

void spCanvasPartFree(spObject object)
{
    return;
}

void spProcessFirstExposure(spComponent component)
{
    spDebug(30, "spProcessFirstExposure", "first exposure\n");
            
    spExposeCanvasCBArch(component);
    SpCanvasPart(component).exposed = SP_TRUE;

    /* get canvas size */
    spGetSize(component, NULL, NULL);
            
    /* draw canvas */
    spRedrawCanvas(component);
    
    spDebug(30, "spProcessFirstExposure", "done\n");
    
    return;
}

void spExposeCanvasCB(spComponent component, void *data)
{
    spDebug(50, "spExposeCanvasCB", "canvas exposed\n");
    
    if (spIsCreated(component) == SP_TRUE) {
        SpCanvasPart(component).refreshed = SP_FALSE;
        
        if (spIsCanvasExposed(component) == SP_FALSE) {
            spProcessFirstExposure(component);
        } else {
            spDebug(30, "spExposeCanvasCB", "exposure (not first)\n");
            
            spGetSize(component, NULL, NULL);

            if (SpCanvasPart(component).prev_width != SpComponentPart(component).client_width
                || SpCanvasPart(component).prev_height != SpComponentPart(component).client_height) {
                SpComponentPart(component).call_reason = SP_CR_RESIZE;
                spRedrawCanvas(component);
            } else {
                /* refresh canvas */
                spRefreshCanvas(component);
            }
        }
        SpCanvasPart(component).prev_width = SpComponentPart(component).client_width;
        SpCanvasPart(component).prev_height = SpComponentPart(component).client_height;
    }
            
    spDebug(50, "spExposeCanvasCB", "done\n");
    
    return;
}

void spResizeCanvasCB(spComponent component, void *data)
{
    spCallbackReason prev_reason;
    
    spDebug(30, "spResizeCanvasCB", "window resized\n");
    
    if (spIsCreated(component) == SP_TRUE) {
        SpCanvasPart(component).refreshed = SP_FALSE;
        
        prev_reason = SpComponentPart(component).call_reason;
        
        if (spIsCanvasExposed(component) == SP_FALSE) {
            spDebug(30, "spResizeCanvasCB", "call spProcessFirstExposure\n");
            SpComponentPart(component).call_reason = SP_CR_EXPOSE;
            spProcessFirstExposure(component);
        } else {
            spDebug(30, "spResizeCanvasCB", "normal resize process\n");
            SpComponentPart(component).call_reason = SP_CR_RESIZE;
        
            SpComponentPart(component).client_width = -1;
            SpComponentPart(component).client_height = -1;
        
            spGetSize(component, NULL, NULL);

            spUpdateCanvasContentFactor(component, 1.0, SP_TRUE);

            spRedrawCanvas(component);
        }
        SpCanvasPart(component).prev_width = SpComponentPart(component).client_width;
        SpCanvasPart(component).prev_height = SpComponentPart(component).client_height;

        SpComponentPart(component).call_reason = prev_reason;
    }

    spDebug(30, "spResizeCanvasCB", "done\n");
    
    return;
}

spBool spCanvasCreate(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    if (SpComponentPart(component).call_func != NULL) {
        SpPrimitivePart(component).draw_func = SpComponentPart(component).call_func;
        SpPrimitivePart(component).draw_data = SpComponentPart(component).call_data;
    }

    spCanvasSetDefaultSize(component);
    if (SpComponentPart(component).disable_initial_scaling == SP_FALSE)
        spPrimitiveScaleDefaultSizeArch(component);
    spCanvasCreateArch(component);
         
    spAddCallback(component, SP_EXPOSE_CALLBACK, spExposeCanvasCB, NULL);

    return SP_TRUE;
}

spBool spCanvasSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spDebug(80, "spCanvasSetParams", "h_content_factor = %f, v_content_factor = %f\n",
            SpCanvasPart(component).h_content_factor, SpCanvasPart(component).v_content_factor);
    spCanvasSetParamsArch(component);
         
    return SP_TRUE;
}

spBool spCanvasDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spCanvasDestroyArch(component);
    spPrimitiveDestroyArch(component);
         
    return SP_TRUE;
}

void spCanvasSetDefaultSize(spComponent component)
{
    if (spIsSubClass(component, SpCanvas) == SP_FALSE) return;

    SpComponentPart(component).x = 0;
    SpComponentPart(component).y = 0;
    SpComponentPart(component).current_width = SpComponentPart(component).width;
#if 0
    if (SpCanvasPart(component).h_scroll == SP_TRUE) {
        if (SpComponentPart(component).width > 0) {
            SpComponentPart(component).current_width += spGetScrollBarDefaultWidth();
        }
    }
#endif
    SpComponentPart(component).current_height = SpComponentPart(component).height;
#if 0
    if (SpCanvasPart(component).v_scroll == SP_TRUE) {
        if (SpComponentPart(component).height > 0) {
            SpComponentPart(component).current_height += spGetScrollBarDefaultWidth();
        }
    }
#endif

    return;
}

spBool spIsCanvasExposed(spComponent component)
{
    if (spIsCanvas(component) == SP_FALSE || SpCanvasPart(component).exposed == SP_FALSE) return SP_FALSE;

    return spIsCanvasExposedArch(component);
}

spBool spIsCanvas(spComponent component)
{
    return spIsSubClass(component, SpCanvas);
}

spBool spIsScrollCanvas(spComponent component)
{
    if (spIsCanvas(component) == SP_TRUE
        && (SpCanvasPart(component).h_scroll == SP_TRUE || SpCanvasPart(component).v_scroll == SP_TRUE)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spComponent spCreateCanvas(spComponent parent, const char *name,
                           int canvas_width, int canvas_height, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
        return NULL;
    
    va_start(argp, canvas_height);
    spGetArgs(argp, args, num_arg);
    va_end(argp);
    
    spSetArg(args[num_arg], SppInitialWidth, canvas_width); num_arg++;
    spSetArg(args[num_arg], SppInitialHeight, canvas_height); num_arg++;

    return spCreateComponentArg(SpCanvasClass, NULL, name, parent, args, num_arg);
}

spBool spRefreshCanvas(spComponent component)
{
    spCanvasClass canvas_class;

    if (component == NULL) return SP_FALSE;
    
    spDebug(80, "spRefreshCanvas", "in: component = %ld\n", (long)component);
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
        spDebug(80, "spRefreshCanvas", "plugin canvas\n");
        if (SpPluginCanvasDrawingRec(component)->refresh == NULL) return SP_TRUE;
        
        if (spIsPluginCanvasDrawable(component) == SP_FALSE) return SP_FALSE;
        
        return SpPluginCanvasDrawingRec(component)->refresh(SpPluginCanvasInstance(component));
    } else {
        spDebug(80, "spRefreshCanvas", "non plugin canvas\n");
        
        if (/*spIsDrawable(component) == SP_FALSE ||*/ spIsCanvas(component) == SP_FALSE
            || (SpComponentPart(component).client_width <= 0 || SpComponentPart(component).client_height <= 0)) {
            spDebug(50, "spRefreshCanvas", "not drawable: %s: width = %d, height = %d\n", 
                    SpGetName(component), SpComponentPart(component).client_width, SpComponentPart(component).client_height);
            return SP_FALSE;
        }
    
        spDebug(50, "spRefreshCanvas", "width = %d, height = %d\n", 
                SpComponentPart(component).client_width, SpComponentPart(component).client_height);
 
        canvas_class = SpGetCanvasClass(component);

        while (canvas_class != NULL) {
            if (canvas_class->canvas.refresh != NULL) {
                spDebug(80, "spRefreshCanvas", "call class refresh function\n");
                return canvas_class->canvas.refresh(component);
            }
            if (streq(canvas_class->object.class_name, SpCanvas)) {
                break;
            }
        
            canvas_class = (spCanvasClass)canvas_class->object.super_class;
        }
    }
    
    spDebug(80, "spRefreshCanvas", "failed\n");
    
    return SP_FALSE;
}

spBool spCanvasRefresh(spComponent component)
{
    if (spIsDrawable(component) == SP_FALSE) return SP_FALSE;
    
    spDebug(80, "spCanvasRefresh", "call spRefreshCanvasArch: %s\n", SpGetName(component));
    
    SpCanvasPart(component).refreshed = SP_TRUE;
    
    return spRefreshCanvasArch(component);
}

spBool spCanvasRedraw(spComponent component)
{
    if (spRedrawCanvasArch(component) == SP_TRUE) {
        SpCanvasPart(component).refreshed = SP_FALSE;
        
        spDebug(80, "spCanvasRedraw", "before spDrawImage\n");
        spDrawImage(component);
        spDebug(80, "spCanvasRedraw", "after spDrawImage: refreshed = %d\n",
                SpCanvasPart(component).refreshed);
        
        if (SpCanvasPart(component).refreshed == SP_FALSE) {
            spDebug(80, "spCanvasRedraw", "call spRefreshCanvas\n");
            return spRefreshCanvas(component);
        } else {
            return SP_TRUE;
        }
    }

    spDebug(80, "spCanvasRedraw", "failed\n");
    
    return SP_FALSE;
}
     
spBool spRedrawCanvas(spComponent component)
{
    spCanvasClass canvas_class;
    
    spDebug(30, "spRedrawCanvas", "in\n");
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
        spDebug(30, "spRedrawCanvas", "for plugin canvas\n");
        if (spStartDrawing(component, spGetName(component)) == SP_TRUE) {
            spEndDrawing(component);
            return spRefreshCanvas(component);
        }
        spDebug(30, "spRedrawCanvas", "spStartDrawing failed\n");
    } else {
        spDebug(30, "spRedrawCanvas", "client_width = %d, client_height = %d\n",
                SpComponentPart(component).client_width, SpComponentPart(component).client_height);

        if (/*spIsPrimitive(component)*/spIsCanvas(component) == SP_FALSE
            || (SpComponentPart(component).client_width <= 0 || SpComponentPart(component).client_height <= 0)) {
            spDebug(30, "spRedrawCanvas", "failed: client size is 0\n");
            return SP_FALSE;
        }
    
        canvas_class = SpGetCanvasClass(component);

        while (canvas_class != NULL) {
            if (canvas_class->canvas.redraw != NULL) {
                spDebug(30, "spRedrawCanvas", "call canvas_class->canvas.redraw\n");
                return canvas_class->canvas.redraw(component);
            }
            if (streq(canvas_class->object.class_name, SpCanvas)) {
                break;
            }
        
            canvas_class = (spCanvasClass)canvas_class->object.super_class;
        }
    }

    spDebug(30, "spRedrawCanvas", "failed: no redraw function\n");
    
    return SP_FALSE;
}

spBool spSetCanvasCapture(spComponent component, spBool flag)
{
    if (spIsCreated(component) == SP_FALSE
        || spIsCanvas(component) == SP_FALSE) return SP_FALSE;

    SpCanvasPart(component).capture_flag = flag;

    return spSetCanvasCaptureArch(component);
}

spBool spSetCanvasCursor(spComponent component, spCursor cursor)
{
    if (spIsCreated(component) == SP_FALSE
        || spIsCanvas(component) == SP_FALSE) {
        return SP_FALSE;
    }
    if (cursor == NULL) {
        return spUnsetCanvasCursor(component);
    }

    return spSetCanvasCursorArch(component, cursor);
}

spBool spUnsetCanvasCursor(spComponent component)
{
    if (spIsCreated(component) == SP_FALSE
        || spIsCanvas(component) == SP_FALSE) {
        return SP_FALSE;
    }

    return spUnsetCanvasCursorArch(component);
}

spBool spIsCanvasFocused(spComponent component)
{
    spBool focused;
    
    if (spIsCreated(component) == SP_FALSE
        || spIsCanvas(component) == SP_FALSE
        || SpCanvasPart(component).focusable == SP_FALSE) return SP_FALSE;

    focused = spIsCanvasFocusedArch(component);
    spDebug(100, "spIsCanvasFocused", "focused = %d\n", focused);

    return focused;
}

spBool spIsCanvasCleared(spComponent component)
{
    if (spIsCanvas(component) == SP_FALSE) return SP_FALSE;

#if defined(ANDROID)
    return spIsCanvasClearedArch(component);
#else
    return SP_FALSE;
#endif
}

spBool spPrintCanvasCreate(spComponent component)
{
#if defined(SP_SUPPORT_PRINT_CANVAS)
    if (spIsCreated(component) == SP_FALSE) {
        if (spPrintCanvasCreateArch(component) == SP_FALSE) {
            return SP_FALSE;
        }
        SpObjectPart(component).create_flag = SP_TRUE;
    }

    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

double spZoomScrollCanvasPageSize(int *page_size, int *value, int minimum, int maximum, int client_size,
                                  double content_factor, double relative_factor, spBool limit_flag)
{
    int range;
    double factor;
    double new_content_factor;
    double page_size_d;

    factor = content_factor * relative_factor;
    spDebug(30, "spZoomScrollCanvasPageSize", "original content_factor = %f, original relative_factor = %f, factor = %f\n",
            content_factor, relative_factor, factor);
    
    range = maximum - minimum;
    page_size_d = (double)client_size / factor;
    if (limit_flag) {
        page_size_d = MIN(page_size_d, (double)range);
        *page_size = (int)spRound(page_size_d);
        *value = MIN(*value, range - *page_size); 
        *value = MAX(*value, 0);
    } else {
        *page_size = (int)spRound(page_size_d);
    }

    /* updated value */
    if (relative_factor != 1.0) {
        new_content_factor = (double)client_size / page_size_d;
        relative_factor = new_content_factor / content_factor;

        spDebug(30, "spZoomScrollCanvasPageSize", "new content_factor = %f, new relative_factor = %f, page_size_d = %f\n",
                new_content_factor, relative_factor, page_size_d);
    }
    
    return relative_factor;
}

spBool spUpdateScrollCanvasPageSize(spComponent component, double relative_factor, spBool limit_flag)
{
    spBool flag = SP_FALSE;
    double h_relative_factor = 1.0;
    double v_relative_factor = 1.0;

    spDebug(100, "spUpdateScrollCanvasPageSize", "in: relative_factor = %f\n", relative_factor);
    
    if (SpCanvasPart(component).h_scroll == SP_TRUE) {
        h_relative_factor
            = spZoomScrollCanvasPageSize(&SpCanvasPart(component).h_page_size, &SpCanvasPart(component).h_value,
                                         SpCanvasPart(component).h_minimum, SpCanvasPart(component).h_maximum,
                                         SpComponentPart(component).client_width, SpCanvasPart(component).h_content_factor,
                                         relative_factor, limit_flag);
        flag = SP_TRUE;
    }
    if (SpCanvasPart(component).v_scroll == SP_TRUE) {
        v_relative_factor
             = spZoomScrollCanvasPageSize(&SpCanvasPart(component).v_page_size, &SpCanvasPart(component).v_value,
                                          SpCanvasPart(component).v_minimum, SpCanvasPart(component).v_maximum,
                                          SpComponentPart(component).client_height, SpCanvasPart(component).v_content_factor,
                                          relative_factor, limit_flag);
        flag = SP_TRUE;
    }
    spDebug(30, "spUpdateScrollCanvasPageSize", "relative_factor = %f, h_relative_factor = %f, v_relative_factor = %f\n",
            relative_factor, h_relative_factor, v_relative_factor);

    if (h_relative_factor < v_relative_factor) {
        if (h_relative_factor != 1.0) {
            if (SpCanvasPart(component).h_scroll == SP_TRUE)
                SpCanvasPart(component).h_content_factor *= h_relative_factor;
        
            if (SpCanvasPart(component).v_scroll == SP_TRUE)
                SpCanvasPart(component).v_content_factor *= h_relative_factor;
        }
    } else {
        if (v_relative_factor != 1.0) {
            if (SpCanvasPart(component).h_scroll == SP_TRUE)
                SpCanvasPart(component).h_content_factor *= v_relative_factor;
        
            if (SpCanvasPart(component).v_scroll == SP_TRUE)
                SpCanvasPart(component).v_content_factor *= v_relative_factor;
        }
    }

    spDebug(30, "spUpdateScrollCanvasPageSize", "flag = %d, h_content_factor = %f, v_content_factor = %f\n",
            flag, SpCanvasPart(component).h_content_factor, SpCanvasPart(component).v_content_factor);
    
    if (flag == SP_TRUE) {
        spUpdateCanvasScrollBarArch(component);
    }

    spDebug(100, "spUpdateScrollCanvasPageSize", "done: relative_factor = %f, flag = %d\n", relative_factor, flag);
    
    return flag;
}

spBool spUpdateCanvasContentFactor(spComponent component, double relative_factor, spBool limit_flag)
{
    if (spIsScrollCanvas(component) == SP_TRUE) {
        return spUpdateScrollCanvasPageSize(component, relative_factor, limit_flag);
    } else {
        spDebug(30, "spUpdateCanvasContentFactor", "relative_factor = %f\n", relative_factor);
        
        spDebug(30, "spUpdateCanvasContentFactor", "original: h_content_factor = %f, v_content_factor = %f\n",
                SpCanvasPart(component).h_content_factor, SpCanvasPart(component).v_content_factor);
        SpCanvasPart(component).h_content_factor *= relative_factor;
        SpCanvasPart(component).v_content_factor *= relative_factor;
        spDebug(30, "spUpdateCanvasContentFactor", "updated: h_content_factor = %f, v_content_factor = %f\n",
                SpCanvasPart(component).h_content_factor, SpCanvasPart(component).v_content_factor);
        return SP_TRUE;
    }
}

spBool spZoomScrollCanvasLimit(spComponent component, double relative_factor, spBool limit_flag)
{
    if (spIsScrollCanvas(component) == SP_FALSE) return SP_FALSE;
    
    spDebug(80, "spZoomScrollCanvasLimit", "relative_factor = %f, limit_flag = %d\n", relative_factor, limit_flag);
    
    if (spUpdateCanvasContentFactor(component, relative_factor, limit_flag) == SP_TRUE) {
        if (spZoomScrollCanvasArch(component) == SP_TRUE) {
            SpComponentPart(component).call_reason = SP_CR_ZOOM;
            spDebug(80, "spZoomScrollCanvasLimit", "call spRedrawCanvas\n");
            spRedrawCanvas(component);
            SpComponentPart(component).call_reason = SP_CR_NONE;
        
            return SP_TRUE;
        }
        spDebug(80, "spZoomScrollCanvasLimit", "spZoomScrollCanvasArch failed\n");
    } else {
        spDebug(80, "spZoomScrollCanvasLimit", "spUpdateCanvasContentFactor failed\n");
    }
    
    return SP_FALSE;
}

spBool spZoomScrollCanvas(spComponent component, double relative_factor)
{
    return spZoomScrollCanvasLimit(component, relative_factor, SP_TRUE);
}

void spZoomInScrollCanvasCB(spComponent component, spComponent canvas)
{
    spZoomScrollCanvas(canvas, 2.0);

    return;
}

void spZoomOutScrollCanvasCB(spComponent component, spComponent canvas)
{
    spZoomScrollCanvas(canvas, 0.5);

    return;
}

void spMoveCanvasScrollBarValue(int *value, double delta, int minimum, int maximum, int page_size, int increment)
{
    *value = (int)spRound((double)*value + delta * (double)increment);
    spCorrectScrollBarValue(value, minimum, maximum, page_size);

    return;
}

#if defined(SP_USE_DEFAULT_CALLBACK_STRUCT)
spCallback *spFindDrawCallbackDefault(spComponent component, long *index, spBool resize_flag)
{
    long k;

    for (k = *index; k < SpPrimitivePart(component).num_callback; k++) {
        if ((resize_flag == SP_TRUE && (SpPrimitivePart(component).callbacks[k].call_type & SP_RESIZE_CALLBACK))
            || (resize_flag == SP_FALSE && (SpPrimitivePart(component).callbacks[k].call_type & SP_EXPOSE_CALLBACK))) {
            *index = k;
            return &SpPrimitivePart(component).callbacks[k];
        }
    }

    return NULL;
}
#endif

spCallback *spFindDrawCallback(spComponent component, long *index, spBool resize_flag)
{
    if (component == NULL || index == NULL) return NULL;

#if defined(SP_USE_DEFAULT_CALLBACK_STRUCT)
    return spFindDrawCallbackDefault(component, index, resize_flag);
#else
    return spFindDrawCallbackArch(component, index, resize_flag);
#endif
}

spBool spEmitDrawCallback(spComponent component)
{
    long k;
    spCallback *callback;
    spBool resize_flag = SP_FALSE;
    spBool flag = SP_FALSE;

    if (SpComponentPart(component).call_reason == SP_CR_RESIZE) {
        resize_flag = SP_TRUE;
    }
    
    if (SpPrimitivePart(component).draw_func != NULL) {
        spDebug(30, "spEmitDrawCallback", "call draw_func\n");
        SpPrimitivePart(component).draw_func(component, SpPrimitivePart(component).draw_data);
        flag = SP_TRUE;
    }

    for (k = 0;; k++) {
        callback = spFindDrawCallback(component, &k, resize_flag);
        if (callback == NULL) {
            spDebug(80, "spEmitDrawCallback", "spFindDrawCallback not found: k = %ld\n", k);
            break;
        }
        if (callback->call_func == spExposeCanvasCB
            || callback->call_func == spResizeCanvasCB) {
            continue;
        }
        spDebug(80, "spEmitDrawCallback", "k = %ld, function found\n", k);

        callback->call_func(component, callback->call_data);
        flag = SP_TRUE;
    }

    return flag;
}
