/*
 *        spCanvas_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <gdk/gdkkeysyms.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spPrimitiveP.h>
#include <sp/spSliderP.h>
#include <sp/spFrameP.h>
#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>
#include <sp/spImageP.h>
#include <sp/spCanvasP.h>

extern spTopLevel sp_toplevel;
extern GtkWidget *sp_grabbed_widget;

static GtkWidget *sp_grabbed_canvas = NULL;
static GtkWidget *sp_focused_canvas = NULL;


void spGetCanvasOriginGtk(spComponent component, int *x, int *y)
{
    if (SpCanvasArch(component).h_adjustment != NULL) {
        /* *x = SpCanvasPart(component).h_value;*/
#if GTK_CHECK_VERSION(3,0,0)
        *x = (int)spRound(gtk_adjustment_get_value(SpCanvasArch(component).h_adjustment));
#else
        *x = (int)spRound(GTK_ADJUSTMENT(SpCanvasArch(component).h_adjustment)->value);
#endif
    } else {
        *x = 0;
    }
    if (SpCanvasArch(component).v_adjustment != NULL) {
        /* *y = SpCanvasPart(component).v_value;*/
#if GTK_CHECK_VERSION(3,0,0)
        *y = (int)spRound(gtk_adjustment_get_value(SpCanvasArch(component).v_adjustment));
#else
        *y = (int)spRound(GTK_ADJUSTMENT(SpCanvasArch(component).v_adjustment)->value);
#endif
    } else {
        *y = 0;
    }

    return;
}

#if GTK_CHECK_VERSION(3,0,0)
spBool spRefreshCanvasGtk(spComponent component, cairo_t *cr)
{
    int x, y;
    
    if (cr == NULL || SpPrimitiveArch(component).pixmap == NULL) return SP_FALSE;

    spGetCanvasOriginGtk(component, &x, &y);
    spDebug(50, "spRefreshCanvasGtk", "x = %d, y = %d\n", x, y);
    
    cairo_set_source_surface(cr, SpPrimitiveArch(component).pixmap, x, y);
    cairo_rectangle(cr, x, y,
                        SpComponentPart(component).client_width, SpComponentPart(component).client_height);
    cairo_fill(cr);

    return SP_TRUE;
}
#endif

spBool spIsCanvasExposedArch(spComponent component)
{
    if (SpCanvasPart(component).exposed
#if GTK_CHECK_VERSION(3,0,0)
        && gtk_widget_get_window(SpPrimitiveArch(component).widget) != NULL
#else
        && SpPrimitiveArch(component).widget->window != NULL
#endif
        ) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

void spExposeCanvasCBArch(spComponent component)
{
    return;
}

spBool spUpdateScrollCanvasContentSizeGtk(spComponent component)
{
    if (SpCanvasArch(component).h_adjustment != NULL || SpCanvasArch(component).v_adjustment != NULL) {
        int width, height;

        width = (int)spRound(SpCanvasPart(component).h_content_factor *
                             (double)(SpCanvasPart(component).h_maximum - SpCanvasPart(component).h_minimum));
        height = (int)spRound(SpCanvasPart(component).v_content_factor *
                              (double)(SpCanvasPart(component).v_maximum - SpCanvasPart(component).v_minimum));
    
        spSetWidgetSizeGtk(SpPrimitiveArch(component).widget, width, height);

        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spUpdateCanvasScrollBarArch(spComponent component)
{
    spBool flag = SP_FALSE;
    spDebug(50, "spUpdateCanvasScrollBarArch", "in\n");

    spUpdateScrollCanvasContentSizeGtk(component);
    
    if (SpCanvasArch(component).h_adjustment != NULL) {
        spDebug(50, "spUpdateCanvasScrollBarArch", "spSetAdjustmentParamsGtk for h_adjustment\n");
        spSetAdjustmentParamsGtk(GTK_ADJUSTMENT(SpCanvasArch(component).h_adjustment),
                                 SpCanvasPart(component).h_minimum,
                                 SpCanvasPart(component).h_maximum,
                                 SpCanvasPart(component).h_value,
                                 SpCanvasPart(component).h_increment,
                                 SpCanvasPart(component).h_page_increment,
                                 SpCanvasPart(component).h_page_size,
                                 SpCanvasPart(component).h_content_factor,
                                 SP_FALSE);
#if GTK_CHECK_VERSION(3,0,0)
        g_signal_emit_by_name(SpCanvasArch(component).h_adjustment, "changed");
#else
        gtk_signal_emit_by_name(GTK_OBJECT(SpCanvasArch(component).h_adjustment), "changed");
#endif
        flag = SP_TRUE;
    }
    
    if (SpCanvasArch(component).v_adjustment != NULL) {
        spDebug(50, "spUpdateCanvasScrollBarArch", "spSetAdjustmentParamsGtk for v_adjustment\n");
        spSetAdjustmentParamsGtk(GTK_ADJUSTMENT(SpCanvasArch(component).v_adjustment),
                                 SpCanvasPart(component).v_minimum,
                                 SpCanvasPart(component).v_maximum,
                                 SpCanvasPart(component).v_value,
                                 SpCanvasPart(component).v_increment,
                                 SpCanvasPart(component).v_page_increment,
                                 SpCanvasPart(component).v_page_size,
                                 SpCanvasPart(component).v_content_factor,
                                 SP_FALSE);
#if GTK_CHECK_VERSION(3,0,0)
        g_signal_emit_by_name(SpCanvasArch(component).v_adjustment, "changed");
#else
        gtk_signal_emit_by_name(GTK_OBJECT(SpCanvasArch(component).v_adjustment), "changed");
#endif
        flag = SP_TRUE;
    }

    return flag;
}

static gint grabPointerCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    static GtkWidget *prev_grabbed = NULL;
    
    switch (event->type) {
      case GDK_BUTTON_PRESS:
      case GDK_2BUTTON_PRESS:
      case GDK_3BUTTON_PRESS:
        spDebug(20, "spGrabPointerGtkCB", "button press\n");
        if (
#if GTK_CHECK_VERSION(3,0,0)
            !gtk_widget_has_grab(widget)
#else
            !GTK_WIDGET_HAS_GRAB(widget)
#endif
            ) {
            sp_grabbed_canvas = widget;
            prev_grabbed = spAddGrabGtk(sp_grabbed_canvas);

            return FALSE;
        }
        break;
      case GDK_BUTTON_RELEASE:
        if (sp_grabbed_canvas != NULL &&
#if GTK_CHECK_VERSION(3,0,0)
            gtk_widget_has_grab(sp_grabbed_canvas)
#else
            GTK_WIDGET_HAS_GRAB(sp_grabbed_canvas)
#endif
            ) {
            spDebug(50, "spGrabPointerGtkCB", "ungrab pointer\n");
            spRemoveGrabGtk(sp_grabbed_canvas);
            sp_grabbed_canvas = NULL;
            
            spAddGrabGtk(prev_grabbed);
            prev_grabbed = NULL;
            return FALSE;
        } else {
#if GTK_CHECK_VERSION(3,0,0)
            g_signal_stop_emission_by_name(widget, "button_release_event");
#else
            gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "button_release_event");
#endif
            return TRUE;
        }
        break;
      case GDK_MOTION_NOTIFY:
        if (event->motion.state) {
            if (sp_grabbed_canvas != NULL) {
                return FALSE;
            } else {
#if GTK_CHECK_VERSION(3,0,0)
                g_signal_stop_emission_by_name(widget, "motion_notify_event");
#else
                gtk_signal_emit_stop_by_name(GTK_OBJECT(widget), "motion_notify_event");
#endif
                return TRUE;
            }
        }
        break;
      case GDK_ENTER_NOTIFY:
        gtk_widget_grab_focus(widget);
        break;
      default:
        break;
    }
    
    return FALSE;
}

static gint focusChangeCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    spComponent component = (spComponent)data;
    
    if (event->type == GDK_FOCUS_CHANGE) {
        if (spIsCreated(component) == SP_TRUE
#if GTK_CHECK_VERSION(3,0,0)
            && gtk_widget_get_window(SpPrimitiveArch(component).widget) != NULL
#else
            && SpPrimitiveArch(component).widget->window != NULL
#endif
            && SpCanvasPart(component).focusable == SP_TRUE) {
            if (event->focus_change.in) {
                sp_focused_canvas = widget;
            } else {
                sp_focused_canvas = NULL;
            }
            spRedrawCanvas(component);
            return TRUE;
        }
    }
    
    return FALSE;
}

static gint keyPressCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    spComponent component = (spComponent)data;
    
    if (event->type== GDK_KEY_PRESS && spIsCreated(component) == SP_TRUE) {
        spDebug(100, "keyPressCB", "key press event received\n");
        if ((SpCanvasPart(component).use_tab_key == SP_TRUE
#if GTK_CHECK_VERSION(3,0,0)
             && event->key.keyval == GDK_KEY_Tab
#else
             && event->key.keyval == GDK_Tab
#endif
                )
            || (SpCanvasPart(component).use_arrow_key == SP_TRUE
#if GTK_CHECK_VERSION(3,0,0)
                && (event->key.keyval == GDK_KEY_Left
                    || event->key.keyval == GDK_KEY_Up
                    || event->key.keyval == GDK_KEY_Right
                    || event->key.keyval == GDK_KEY_Down)
#else
                && (event->key.keyval == GDK_Left
                    || event->key.keyval == GDK_Up
                    || event->key.keyval == GDK_Right
                    || event->key.keyval == GDK_Down)
#endif
                )) {
            spDebug(100, "keyPressCB", "return TRUE\n");
            return TRUE;
        }
    }
    
    spDebug(100, "keyPressCB", "return FALSE\n");
    
    return FALSE;
}

void spCanvasCreateArch(spComponent component)
{
    gint events;
    
    SpCanvasArch(component).h_adjustment = NULL;
    SpCanvasArch(component).v_adjustment = NULL;
#if GTK_CHECK_VERSION(3,0,0)
    SpCanvasArch(component).draw_cr = NULL;
    SpCanvasArch(component).wait_refresh = SP_FALSE;
    SpCanvasArch(component).in_draw_func = SP_FALSE;
    SpCanvasArch(component).skip_draw_func = SP_FALSE;
#endif
    
    if (SpCanvasPart(component).h_scroll == SP_TRUE || SpCanvasPart(component).v_scroll == SP_TRUE) {
        SpPrimitiveArch(component).top_widget = gtk_scrolled_window_new(NULL, NULL);
        gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(SpPrimitiveArch(component).top_widget),
                                       SpCanvasPart(component).h_scroll == SP_TRUE ? GTK_POLICY_ALWAYS: GTK_POLICY_NEVER,
                                       SpCanvasPart(component).v_scroll == SP_TRUE ? GTK_POLICY_ALWAYS: GTK_POLICY_NEVER);
        spContainerAddGtk(component, SpPrimitiveArch(component).top_widget);
        
        if (SpCanvasPart(component).h_scroll == SP_TRUE) {
            SpCanvasArch(component).h_adjustment
                = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(SpPrimitiveArch(component).top_widget));
        }
        if (SpCanvasPart(component).v_scroll == SP_TRUE) {
            SpCanvasArch(component).v_adjustment
                = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(SpPrimitiveArch(component).top_widget));
        }
        
        if (SpComponentPart(component).visible_flag == SP_TRUE) {
            gtk_widget_show(SpPrimitiveArch(component).top_widget);
        }
    } else if (SpCanvasPart(component).border_on == SP_TRUE) {
        spPrimitiveCreateBorderGtk(component);
    }
    
#if GTK_CHECK_VERSION(3,16,0)
    if (spIsSubClass(component, "GLCanvas") == SP_TRUE) {
        SpPrimitiveArch(component).widget = gtk_gl_area_new();
        spDebug(50, "spCanvasCreateArch", "gtk_gl_area_new result: %lx\n", (long)SpPrimitiveArch(component).widget);
    } else {
        SpPrimitiveArch(component).widget = gtk_drawing_area_new();
    }
#else
    SpPrimitiveArch(component).widget = gtk_drawing_area_new();
#endif

    spSetStyleGtk(component, SpPrimitiveArch(component).widget);
    events = GDK_BUTTON_PRESS_MASK
        | GDK_BUTTON_RELEASE_MASK
        | GDK_BUTTON_MOTION_MASK
        | GDK_BUTTON1_MOTION_MASK
        | GDK_BUTTON2_MOTION_MASK
        | GDK_BUTTON3_MOTION_MASK
        | GDK_POINTER_MOTION_MASK
        | GDK_KEY_PRESS_MASK
        | GDK_KEY_RELEASE_MASK
        | GDK_ENTER_NOTIFY_MASK
        | GDK_LEAVE_NOTIFY_MASK
        | GDK_STRUCTURE_MASK
        | GDK_EXPOSURE_MASK
        | GDK_FOCUS_CHANGE_MASK;
#if GTK_CHECK_VERSION(2,0,0)
    if (SpCanvasPart(component).accept_wheel_event == SP_TRUE) {
        events |= GDK_SCROLL_MASK;
    }
#endif
    gtk_widget_set_events(SpPrimitiveArch(component).widget, events);
    
    spPrimitiveContainerAddGtk(component);

    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "button_press_event", grabPointerCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "button_release_event", grabPointerCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "motion_notify_event", grabPointerCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "enter_notify_event", grabPointerCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "leave_notify_event", grabPointerCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "focus_in_event", focusChangeCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "focus_out_event", focusChangeCB, component);
#if GTK_CHECK_VERSION(3,0,0)
    g_signal_connect_after(SpPrimitiveArch(component).widget,
                           "key_press_event", G_CALLBACK(keyPressCB), component);
#else
    gtk_signal_connect_after(GTK_OBJECT(SpPrimitiveArch(component).widget),
                             "key_press_event", GTK_SIGNAL_FUNC(keyPressCB), component);
#endif

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)
    spDebug(50, "spCanvasCreateArch", "spAddCallback of spResizeCanvasCB\n");
    spAddCallback(component, SP_RESIZE_CALLBACK, spResizeCanvasCB, NULL);
#endif

    if (SpPrimitivePart(component).draw_func != NULL
        && (SpCanvasArch(component).h_adjustment != NULL || SpCanvasArch(component).v_adjustment != NULL)) {
        spAddCallback(component, SP_SCROLL_CALLBACK | SP_WHEEL_CALLBACK | SP_ZOOM_CALLBACK,
                      SpPrimitivePart(component).draw_func, SpPrimitivePart(component).draw_data);
    }

#if GTK_CHECK_VERSION(3,0,0)
    gtk_widget_set_can_focus(SpPrimitiveArch(component).widget, TRUE);
#else
    GTK_WIDGET_SET_FLAGS(SpPrimitiveArch(component).widget, GTK_CAN_FOCUS);
#endif
    
#if GTK_CHECK_VERSION(3,0,0)
    if (SpCanvasPart(component).h_scroll == SP_TRUE || SpCanvasPart(component).v_scroll == SP_TRUE) {
        gtk_widget_set_hexpand(SpPrimitiveArch(component).widget, FALSE);
        gtk_widget_set_vexpand(SpPrimitiveArch(component).widget, FALSE);
    }
#endif
        
    return;
}

void spCanvasSetParamsArch(spComponent component)
{
    spUpdateCanvasScrollBarArch(component);
    
    return;
}
    
void spCanvasDestroyArch(spComponent component)
{
    return;
}
    
spBool spRefreshCanvasArch(spComponent component)
{
    int x, y;
    
    spDebug(50, "spRefreshCanvasArch", "in\n");

    spGetCanvasOriginGtk(component, &x, &y);
    spDebug(50, "spRefreshCanvasArch", "x = %d, y = %d\n", x, y);
    
#if GTK_CHECK_VERSION(3,0,0)
    {
        GdkWindow *gdkwin;
        
        if ((gdkwin = gtk_widget_get_window(SpPrimitiveArch(component).widget)) == NULL) return SP_FALSE;

        if (spIsSubClass(component, "GLCanvas") == SP_FALSE && SpCanvasArch(component).draw_cr != NULL) {
            spDebug(50, "spRefreshCanvasArch", "in draw (expose) callback\n");
            
            /* in draw (expose) callback */
            if (spRefreshCanvasGtk(component, SpCanvasArch(component).draw_cr) == SP_FALSE) {
                spDebug(50, "spRefreshCanvasArch", "spRefreshCanvasGtk failed\n");
                return SP_FALSE;
            }
        } else if (SpCanvasArch(component).in_draw_func == SP_FALSE) {
            spDebug(50, "spRefreshCanvasArch", "wait refresh\n");
            SpCanvasArch(component).wait_refresh = SP_TRUE;
            gtk_widget_queue_draw_area(SpPrimitiveArch(component).widget, x, y,
                                       SpComponentPart(component).client_width, SpComponentPart(component).client_height);
        }
    }
#else
    if (SpPrimitiveArch(component).widget->window == NULL) return SP_FALSE;
    
    gdk_draw_pixmap(SpPrimitiveArch(component).widget->window,
                    SpTopLevelArch(sp_toplevel).fg_gc,
                    SpPrimitiveArch(component).pixmap,
                    0, 0,
                    x, y,
                    SpComponentPart(component).client_width,
                    SpComponentPart(component).client_height);
#endif
    
    spDebug(30, "spRefreshCanvasArch", "width = %d, height = %d\n", 
            SpComponentPart(component).client_width, 
            SpComponentPart(component).client_height);
    
    return SP_TRUE;
}

spBool spRedrawCanvasArch(spComponent component)
{
    spRedrawImageArch(component);
    
    spDebug(50, "spRedrawCanvasArch", "client_width = %d, client_height = %d\n",
            SpComponentPart(component).client_width, SpComponentPart(component).client_height);

    return SP_TRUE;
}

int captureCanvasCB(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    int x, y;
    GtkWidget *gtkwin;
    spComponent window;
    spComponent component = (spComponent)data;
    static int prev_dx = 0, prev_dy = 0;
    static int prev_button = 0;
    static GtkWidget *prev_widget = NULL;

    spDebug(50, "captureCanvasCB", "prev_dx = %d, prev_dy = %d, prev_widget = %ld\n",
            prev_dx, prev_dy, (long)prev_widget);
    
    if (component == NULL
        || SpCanvasPart(component).capture_flag == SP_FALSE) {
        spDebug(50, "captureCanvasCB", "capture is not required\n");
        return FALSE;
    }
    
    window = SpGetWindow(component);
    gtkwin = SpFrameArch(window).toplevel;
    
    spDebug(50, "captureCanvasCB", "event->type = %d\n", event->type);
    
    switch (event->type) {
      case GDK_MOTION_NOTIFY:
        if (widget != prev_widget
            || !(((event->motion.state & GDK_BUTTON1_MASK) && prev_button == 1)
                 || ((event->motion.state & GDK_BUTTON2_MASK) && prev_button == 2)
                 || ((event->motion.state & GDK_BUTTON3_MASK) && prev_button == 3))) {
            prev_dx = 0; prev_dy = 0;
            prev_button = 0;
            prev_widget = NULL;
            return FALSE;            
        }
        
        SpComponentPart(window).x = event->motion.x_root + prev_dx;
        SpComponentPart(window).y = event->motion.y_root + prev_dy;

#if GTK_CHECK_VERSION(2,0,0)
        gtk_window_move(GTK_WINDOW(gtkwin), SpComponentPart(window).x, SpComponentPart(window).y);
#else
        gdk_window_move(spGetGdkWindowGtk(SpFrameArch(window).toplevel), SpComponentPart(window).x, SpComponentPart(window).y);
#endif
        break;
      case GDK_BUTTON_PRESS:
#if GTK_CHECK_VERSION(2,0,0)
        gtk_window_get_position(GTK_WINDOW(gtkwin), &x, &y);
#else
        gdk_window_get_position(spGetGdkWindowGtk(SpFrameArch(window).toplevel), &x, &y);
#endif
        prev_dx = x - event->button.x_root;
        prev_dy = y - event->button.y_root;
        prev_button = event->button.button;
        prev_widget = widget;
        break;
      case GDK_BUTTON_RELEASE:
        prev_dx = 0; prev_dy = 0;
        prev_button = 0;
        prev_widget = NULL;
        break;
      default:
        break;
    }
    
    spDebug(50, "captureCanvasCB", "done\n");
    
    return FALSE;
}

spBool spSetCanvasCaptureArch(spComponent component)
{
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "button_press_event", captureCanvasCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "button_release_event", captureCanvasCB, component);
    spSignalConnectEventGtk(SpPrimitiveArch(component).widget,
                            "motion_notify_event", captureCanvasCB, component);
    
    return SP_TRUE;
}

spBool spSetCanvasCursorArch(spComponent component, spCursor cursor)
{
    GdkWindow *gdkwin;
    
#if GTK_CHECK_VERSION(3,0,0)
    gdkwin = gtk_widget_get_window(SpPrimitiveArch(component).widget);
#else
    gdkwin = SpPrimitiveArch(component).widget->window;
#endif
    
    gdk_window_set_cursor(gdkwin, SpCursorArch(cursor).cursor);
    gdk_flush();
    
    return SP_TRUE;
}

spBool spUnsetCanvasCursorArch(spComponent component)
{
    GdkWindow *gdkwin;
    
#if GTK_CHECK_VERSION(3,0,0)
    gdkwin = gtk_widget_get_window(SpPrimitiveArch(component).widget);
#else
    gdkwin = SpPrimitiveArch(component).widget->window;
#endif
    
    gdk_window_set_cursor(gdkwin, NULL);
    gdk_flush();
    
    return SP_TRUE;
}

spBool spIsCanvasFocusedArch(spComponent component)
{
    if (SpPrimitiveArch(component).widget == sp_focused_canvas) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPrintCanvasCreateArch(spComponent component)
{
    return SP_TRUE;
}

spBool spZoomScrollCanvasArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)
    spUpdateScrollCanvasContentSizeGtk(component);
#endif
    return SP_TRUE;
}

spCallback *spFindDrawCallbackArch(spComponent component, long *index, spBool resize_flag)
{
    long k;
#if GTK_CHECK_VERSION(3,0,0)
    static const char *expose_name = "draw";
#else
    static const char *expose_name = "expose_event";
#endif

    for (k = *index; k < SpPrimitivePart(component).num_callback; k++) {
        if ((resize_flag == SP_TRUE && streq(SpPrimitivePart(component).callbacks[k].call_name, "configure_event"))
            || (resize_flag == SP_FALSE && streq(SpPrimitivePart(component).callbacks[k].call_name, expose_name))) {
            *index = k;
            return &SpPrimitivePart(component).callbacks[k];
        }
    }

    return NULL;
}
