/* device independent bitmap using plug-in */
#include <sp/spBaseLib.h>
#include <sp/spPluginP.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

int spGetDIBitmapPluginCount(void);
const char *spGetDIBitmapPluginName(int index);
void *spInitDIBitmapInstancePlugin(const char *name);
spBool spFreeDIBitmapInstancePlugin(void *instance);
int spGetDIBitmapPriorityPlugin(void *instance);
const char **spGetDIBitmapFileTypeListPlugin(void *instance);
const char **spGetDIBitmapFileDescListPlugin(void *instance);
const char **spGetDIBitmapFileFilterListPlugin(void *instance);
spBool spIsDIBitmapSupportedPlugin(void *instance, const char *filename);
spBool spSetDIBitmapFileTypePlugin(void *instance, int index);
int spGetDIBitmapFileTypePlugin(void *instance);
unsigned long spGetDIBitmapMetaInfoMaskPlugin(void *instance);
spBool spSetDIBitmapMetaInfoPlugin(void *instance, spMetaInfo *meta_info);
spBool spGetDIBitmapMetaInfoPlugin(void *instance, spMetaInfo *meta_info);
const char *spGetDIBitampBestSuffixPlugin(void *instance);
unsigned long spGetDIBitmapCapsPlugin(void *instance);

spBool spOpenDIBitmapFilePlugin(void *instance, const char *filename, const char *mode);
spBool spCloseDIBitmapFilePlugin(void *instance);
spBool spSetDIBitmapFileInfoPlugin(void *instance, spDIBitmapInfo *file_info, unsigned long *strict_mask);
spBool spSetDIBitmapInfoPlugin(void *instance, spDIBitmapInfoInternal *info);
spBool spGetDIBitmapFileInfoPlugin(void *instance, spDIBitmapInfo *file_info);
spBool spGetDIBitmapInfoPlugin(void *instance, spDIBitmapInfoInternal *info, unsigned long *strict_mask);
void *spReadDIBitmapPlugin(void *instance, spDIBitmapInfoInternal *info, void *bitmap);
spBool spWriteDIBitmapPlugin(void *instance, spDIBitmapInfoInternal *info, void *bitmap);
spBool spReadDIBitmapRGBAPlugin(void *instance, int *r, int *g, int *b, int *a);
spBool spWriteDIBitmapRGBAPlugin(void *instance, int r, int g, int b, int a);

static spDIBitmapFuncList sp_plugin_dibitmap_func_list =
{
    spGetDIBitmapPluginCount,
    spGetDIBitmapPluginName,
	
    spInitDIBitmapInstancePlugin,
    spFreeDIBitmapInstancePlugin,
	
    spGetDIBitmapPriorityPlugin,
    spGetDIBitmapFileTypeListPlugin,
    spGetDIBitmapFileDescListPlugin,
    spGetDIBitmapFileFilterListPlugin,
    spIsDIBitmapSupportedPlugin,
    spSetDIBitmapFileTypePlugin,
    spGetDIBitmapFileTypePlugin,
    spGetDIBitmapMetaInfoMaskPlugin,
    spSetDIBitmapMetaInfoPlugin,
    spGetDIBitmapMetaInfoPlugin,
    spGetDIBitampBestSuffixPlugin,
	
    spGetDIBitmapCapsPlugin,
	
    spOpenDIBitmapFilePlugin,
    spCloseDIBitmapFilePlugin,
	
    spSetDIBitmapFileInfoPlugin,
    spSetDIBitmapInfoPlugin,
    spGetDIBitmapFileInfoPlugin,
    spGetDIBitmapInfoPlugin,
	
    spReadDIBitmapPlugin,
    spWriteDIBitmapPlugin,
	
    spReadDIBitmapRGBAPlugin,
    spWriteDIBitmapRGBAPlugin,
};

spDIBitmapFuncList *spGetPluginDIBitmapFuncList(void)
{
    return &sp_plugin_dibitmap_func_list;
}

extern spBool spAddBmpDIBitmapPlugin(void);
extern spBool spRemoveBmpDIBitmapPlugin(void);
#if defined(GTK) || defined(USE_MOTIF)
extern spBool spAddPngDIBitmapPlugin(void);
extern spBool spRemovePngDIBitmapPlugin(void);
extern spBool spAddJpegDIBitmapPlugin(void);
extern spBool spRemoveJpegDIBitmapPlugin(void);
#elif defined(_WIN32)
extern spBool spAddSusieDIBitmapPlugin(void);
extern spBool spRemoveSusieDIBitmapPlugin(void);
#endif

spBool spSearchDIBitmapPluginFunc(spPlugin *plugin, void *data)
{
    spDebug(100, "spSearchDIBitmapPluginFunc", "name = %s\n", spGetPluginName(plugin));
    return spIsDIBitmapPlugin(plugin);
}

int spGetDIBitmapPluginCount(void)
{
    spAddBmpDIBitmapPlugin();
#if defined(GTK) || defined(USE_MOTIF)
#if defined(SP_DIBITMAP_PLUGIN_USE_PNG)
    spAddPngDIBitmapPlugin();
#endif
#if defined(SP_DIBITMAP_PLUGIN_USE_JPEG)
    spAddJpegDIBitmapPlugin();
#endif
#elif defined(_WIN32)
    spAddSusieDIBitmapPlugin();
#endif

    return spGetNumSpecificPlugin(spSearchDIBitmapPluginFunc, NULL);
}

static void spFreeDIBitmapPluginNameList(void *list)
{
    int i;
    char **name_list = (char **)list;

    if (name_list != NULL) {
	for (i = 0; name_list[i] != NULL; i++) {
	    xfree(name_list[i]);
	}
	xfree(name_list);
    }

    spRemoveBmpDIBitmapPlugin();
#if defined(GTK) || defined(USE_MOTIF)
#if defined(SP_DIBITMAP_PLUGIN_USE_PNG)
    spRemoveJpegDIBitmapPlugin();
#endif
#if defined(SP_DIBITMAP_PLUGIN_USE_JPEG)
    spRemovePngDIBitmapPlugin();
#endif
#elif defined(_WIN32)
    spRemoveSusieDIBitmapPlugin();
#endif
    
    return;
}

static char **sp_dibitmap_plugin_name_list = NULL;

const char *spGetDIBitmapPluginName(int index)
{
    if (sp_dibitmap_plugin_name_list == NULL) {
	int i, count;
	int index = 0;

	count = spGetDIBitmapPluginCount();
	
	sp_dibitmap_plugin_name_list = xalloc(count + 1, char *);

	for (i = 0; i < count; i++) {
	    sp_dibitmap_plugin_name_list[i] = xspSearchSpecificPluginFile(&index, spSearchDIBitmapPluginFunc, NULL);
	    spDebug(100, "spSearchDIBitmapPluginFunc", "sp_dibitmap_plugin_name_list[%d] = %s\n",
		    i, sp_dibitmap_plugin_name_list[i]);
	    
	    if (index < 0) {
		++i;
		break;
	    }
	    ++index;
	}
	
	sp_dibitmap_plugin_name_list[i] = NULL;
	
	spAddExitCallback(spFreeDIBitmapPluginNameList, sp_dibitmap_plugin_name_list);
    }
    
    return sp_dibitmap_plugin_name_list[index];
}

void *spInitDIBitmapInstancePlugin(const char *name)
{
    spPlugin *plugin = NULL;

    spDebug(80, "spInitDIBitmapInstancePlugin", "in: name = %s\n", name);
    
    if ((plugin = spLoadPlugin(name)) != NULL) {
	if (spIsDIBitmapPlugin(plugin) == SP_TRUE
	    && spInitPluginInstance(plugin) == SP_TRUE) {
	    spDebug(80, "spInitDIBitmapInstancePlugin", "successfully done\n");
	    return plugin;
	} else {
	    spDebug(10, "spInitDIBitmapInstancePlugin", "spIsDIBitmapPlugin or spInitPluginInstance failed\n");
	}
	spFreePlugin(plugin);
    } else {
	spDebug(10, "spInitDIBitmapInstancePlugin", "spLoadPlugin failed\n");
    }

    return NULL;
}

spBool spFreeDIBitmapInstancePlugin(void *instance)
{
    spPlugin *plugin = (spPlugin *)instance;
    spFreePlugin(plugin);
    return SP_TRUE;
}

int spGetDIBitmapPriorityPlugin(void *instance)
{
    spPlugin *plugin = (spPlugin *)instance;
    
    return spGetPluginPriority(plugin);
}

const char **spGetDIBitmapFileTypeListPlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    return rec->file_type_list;
}

const char **spGetDIBitmapFileDescListPlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    return rec->file_desc_list;
}

const char **spGetDIBitmapFileFilterListPlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    return rec->file_filter_list;
}

spBool spIsDIBitmapSupportedPlugin(void *instance, const char *filename)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    spDebug(100, "spIsDIBitmapSupportedPlugin", "filename = %s\n", filename);

    if (rec->is_supported == NULL) return SP_FALSE;
    
    return rec->is_supported(filename);
}

spBool spSetDIBitmapFileTypePlugin(void *instance, int index)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->set_file_type == NULL) return SP_FALSE;
    
    return rec->set_file_type(plugin->instance, index);
}

int spGetDIBitmapFileTypePlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->get_file_type == NULL) return -1;
    
    return rec->get_file_type(plugin->instance);
}

unsigned long spGetDIBitmapMetaInfoMaskPlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    return rec->meta_info_mask;
}

spBool spSetDIBitmapMetaInfoPlugin(void *instance, spMetaInfo *meta_info)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->set_meta_info == NULL) return SP_FALSE;
    
    return rec->set_meta_info(plugin->instance, meta_info);
}

spBool spGetDIBitmapMetaInfoPlugin(void *instance, spMetaInfo *meta_info)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->get_meta_info == NULL) return SP_FALSE;
    
    return rec->get_meta_info(plugin->instance, meta_info);
}

const char *spGetDIBitampBestSuffixPlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->get_best_suffix == NULL) return NULL;
    
    return rec->get_best_suffix(plugin->instance);
}

unsigned long spGetDIBitmapCapsPlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->get_dib_caps == NULL) return 0L;
    
    return rec->get_dib_caps(plugin->instance);
}

spBool spOpenDIBitmapFilePlugin(void *instance, const char *filename, const char *mode)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;

    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->open(plugin->instance, filename, mode) == SP_PLUGIN_ERROR_SUCCESS) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spCloseDIBitmapFilePlugin(void *instance)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    return rec->close(plugin->instance);
}

spBool spSetDIBitmapFileInfoPlugin(void *instance, spDIBitmapInfo *file_info, unsigned long *strict_mask)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->set_file_dib_info == NULL) return SP_TRUE;
    
    return rec->set_file_dib_info(plugin->instance, file_info, strict_mask);
}

spBool spSetDIBitmapInfoPlugin(void *instance, spDIBitmapInfoInternal *info)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->set_dib_info == NULL) return SP_TRUE;
    
    return rec->set_dib_info(plugin->instance, info);
}

spBool spGetDIBitmapFileInfoPlugin(void *instance, spDIBitmapInfo *file_info)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->get_file_dib_info == NULL) return SP_TRUE;
    
    return rec->get_file_dib_info(plugin->instance, file_info);
}

spBool spGetDIBitmapInfoPlugin(void *instance, spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->get_dib_info == NULL) return SP_TRUE;
    
    return rec->get_dib_info(plugin->instance, info, strict_mask);
}

void *spReadDIBitmapPlugin(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->read == NULL) return NULL;
    
    if (rec->read(plugin->instance, info, info->data) == SP_TRUE) {
        return bitmap;
    } else {
        return NULL;
    }
}

spBool spWriteDIBitmapPlugin(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->write == NULL) return SP_FALSE;
    
    return rec->write(plugin->instance, info, info->data);
}

spBool spReadDIBitmapRGBAPlugin(void *instance, int *r, int *g, int *b, int *a)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->read_rgba == NULL) return SP_FALSE;
    
    return rec->read_rgba(plugin->instance, r, g, b, a);
}

spBool spWriteDIBitmapRGBAPlugin(void *instance, int r, int g, int b, int a)
{
    spDIBitmapPluginRec *rec;
    spPlugin *plugin = (spPlugin *)instance;
    
    rec = (spDIBitmapPluginRec *)plugin->rec;

    if (rec->write_rgba == NULL) return SP_FALSE;
    
    return rec->write_rgba(plugin->instance, r, g, b, a);
}
