/*
 *	spDraw_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spString.h>
#include <sp/spMemory.h>
#include <sp/spCanvas.h>
#include <sp/spDIBitmap.h>

#include <sp/spLocale.h>
#include <sp/spFrameP.h>
#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

spBool spIsAlphaSupported(spComponent component)
{
    return SP_FALSE;
}

spBool spIsDrawable(spComponent component)
{
    if (spIsPrimitive(component) == SP_TRUE
#if GTK_CHECK_VERSION(3,0,0)
	&& SpPrimitiveArch(component).pixmap_cr != NULL
#else
	&& SpPrimitiveArch(component).pixmap != NULL
#endif
	) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void *spGetImageHandleArch(spComponent component)
{
    return component;
}

spBool spCopyImageArch(spComponent src, spComponent dest,
		       int src_x, int src_y, int width, int height,
		       int dest_x, int dest_y)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(dest).pixmap_cr != NULL) {
	cairo_set_source_surface(SpPrimitiveArch(dest).pixmap_cr,
				 SpPrimitiveArch(src).pixmap, dest_x - src_x, dest_y - src_y);
	cairo_rectangle(SpPrimitiveArch(dest).pixmap_cr, dest_x, dest_y, width, height);
	cairo_fill(SpPrimitiveArch(dest).pixmap_cr);
    } else {
	return SP_FALSE;
    }
#else
    gdk_draw_pixmap(SpPrimitiveArch(dest).pixmap,
		    SpTopLevelArch(sp_toplevel).fg_gc,
		    SpPrimitiveArch(src).pixmap,
		    src_x, src_y,
		    dest_x, dest_y,
		    MIN(SpComponentPart(src).client_width - src_x, width),
		    MIN(SpComponentPart(src).client_height - src_y, height));
#endif

    return SP_TRUE;
}

spBool spStretchImageArch(spComponent src, spComponent dest,
			  int src_x, int src_y, int src_width, int src_height,
			  int dest_x, int dest_y, int dest_width, int dest_height)
{
#if GTK_CHECK_VERSION(3,0,0)
    double sx, sy;

    spDebug(80, "spStretchImageArch", "src_x = %d, src_y = %d, dest_x = %d, dest_y = %d\n",
	    src_x, src_y, dest_x, dest_y);
    
    sx = (double)dest_width / (double)src_width;
    sy = (double)dest_height / (double)src_height;
    spDebug(80, "spStretchImageArch", "sx = %f, sy = %f\n", sx, sy);

    if (SpPrimitiveArch(dest).pixmap_cr != NULL && SpPrimitiveArch(src).pixmap != NULL) {
	spDebug(80, "spStretchImageArch", "before cairo_save\n");
	cairo_save(SpPrimitiveArch(dest).pixmap_cr);
	
	if (sx != 1.0 || sy != 1.0) {
	    spDebug(80, "spStretchImageArch", "before cairo_scale\n");
	    cairo_scale(SpPrimitiveArch(dest).pixmap_cr, sx, sy);
	}
	
	spDebug(80, "spStretchImageArch", "before cairo_set_source_surface\n");
	
	cairo_set_source_surface(SpPrimitiveArch(dest).pixmap_cr,
				 SpPrimitiveArch(src).pixmap, dest_x - src_x, dest_y - src_y);
	spDebug(80, "spStretchImageArch", "before cairo_rectangle\n");
	cairo_rectangle(SpPrimitiveArch(dest).pixmap_cr, dest_x, dest_y, (double)dest_width / sx, (double)dest_height / sy);
	spDebug(80, "spStretchImageArch", "before cairo_fill\n");
	cairo_fill(SpPrimitiveArch(dest).pixmap_cr);
	
	spDebug(80, "spStretchImageArch", "before cairo_restore\n");
	cairo_restore(SpPrimitiveArch(dest).pixmap_cr);
    } else {
	spDebug(80, "spStretchImageArch", "failed: pixmap_cr == NULL\n");
	return SP_FALSE;
    }
    
    spDebug(80, "spStretchImageArch", "done\n");
    
    return SP_TRUE;
#else
    return spStretchImageDefault(src, dest, src_x, src_y, src_width, src_height,
				 dest_x, dest_y, dest_width, dest_height);
#endif
}

void spFillRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	spUpdateColorCairo(SpPrimitiveArch(component).pixmap_cr, graphics);
	cairo_rectangle(SpPrimitiveArch(component).pixmap_cr, x - 0.5, y - 0.5, width + 1.0, height + 1.0);
	cairo_fill(SpPrimitiveArch(component).pixmap_cr);
    }
#else
    gdk_draw_rectangle(SpPrimitiveArch(component).pixmap,
		       (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
			SpGraphicsArch(graphics).gc), TRUE, 
		       x, y, width, height);
#endif

    return;
}

void spDrawRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	spUpdateLineContextCairo(SpPrimitiveArch(component).pixmap_cr, graphics);
	cairo_rectangle(SpPrimitiveArch(component).pixmap_cr, x - 0.5, y - 0.5, width + 1.0, height + 1.0);
	cairo_stroke(SpPrimitiveArch(component).pixmap_cr);
    }
#else
    gdk_draw_rectangle(SpPrimitiveArch(component).pixmap,
		       (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
			SpGraphicsArch(graphics).gc), FALSE, 
		       x, y, width, height);
#endif

    return;
}

#if GTK_CHECK_VERSION(3,0,0)
void spEllipseCairo(cairo_t *cr, double x, double y, double width, double height, double angle1, double angle2,
		    spBool close_flag)
{
    spDebug(80, "spEllipseCairo", "x = %f, y = %f, width = %f, height = %f\n", x, y, width, height);
    cairo_save(cr);
    cairo_translate(cr, x + width / 2.0, y + height / 2.0);
    cairo_scale(cr, width / 2.0, height / 2.0);
    if (angle2 < 0.0) {
	cairo_arc(cr, 0.0, 0.0, 1.0, -angle1 * PI / 180.0, (-angle1 - angle2) * PI / 180.0);
    } else {
	cairo_arc_negative(cr, 0.0, 0.0, 1.0, -angle1 * PI / 180.0, (-angle1 - angle2) * PI / 180.0);
    }
    if (close_flag) {
	cairo_line_to(cr, 0.0, 0.0);
    }
    cairo_restore(cr);
    spDebug(80, "spEllipseCairo", "done\n");

    return;
}
#endif

void spFillArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	spUpdateColorCairo(SpPrimitiveArch(component).pixmap_cr, graphics);
	spEllipseCairo(SpPrimitiveArch(component).pixmap_cr, x - 0.5, y - 0.5, width + 1.0, height + 1.0,
		       angle1, angle2, SP_TRUE);
	cairo_fill(SpPrimitiveArch(component).pixmap_cr);
    }
#else
    gdk_draw_arc(SpPrimitiveArch(component).pixmap,
		 (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		  SpGraphicsArch(graphics).gc), TRUE, 
		 x, y, width, height, (int)spRound(angle1 * 64.0), (int)spRound(angle2 * 64.0));
#endif

    return;
}

void spDrawArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	spUpdateLineContextCairo(SpPrimitiveArch(component).pixmap_cr, graphics);
	spEllipseCairo(SpPrimitiveArch(component).pixmap_cr, x - 0.5, y - 0.5, width + 1.0, height + 1.0,
		       angle1, angle2, SP_FALSE);
	cairo_stroke(SpPrimitiveArch(component).pixmap_cr);
    }
#else
    gdk_draw_arc(SpPrimitiveArch(component).pixmap,
		 (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		  SpGraphicsArch(graphics).gc), FALSE, 
		 x, y, width, height, (int)spRound(angle1 * 64.0), (int)spRound(angle2 * 64.0));
#endif

    return;
}

void spDrawLineArch(spComponent component, spGraphics graphics,
		    int x1, int y1, int x2, int y2)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	spDebug(100, "spDrawLineArch", "line from (%d, %d) --> (%d, %d)\n", x1, y1, x2, y2);
	spUpdateLineContextCairo(SpPrimitiveArch(component).pixmap_cr, graphics);
	cairo_move_to(SpPrimitiveArch(component).pixmap_cr, x1, y1);
	cairo_line_to(SpPrimitiveArch(component).pixmap_cr, x2, y2);
	cairo_stroke(SpPrimitiveArch(component).pixmap_cr);
	spFillRectangleArch(component, graphics, x2, y2, /*1*/0, /*1*/0);
    }
#else
    gdk_draw_line(SpPrimitiveArch(component).pixmap,
		  (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		   SpGraphicsArch(graphics).gc),
		  x1, y1, x2, y2);
    gdk_draw_point(SpPrimitiveArch(component).pixmap,
		   (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		    SpGraphicsArch(graphics).gc),
		   x2, y2);
#endif

    return;
}

void spDrawPointArch(spComponent component, spGraphics graphics,
		     int x, int y)
{
#if GTK_CHECK_VERSION(3,0,0)
    spFillRectangleArch(component, graphics, x, y, /*1*/0, /*1*/0);
#else
    gdk_draw_point(SpPrimitiveArch(component).pixmap,
		   (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		    SpGraphicsArch(graphics).gc),
		   x, y);
#endif
    
    return;
}

#if GTK_CHECK_VERSION(2,0,0)
spBool spGetFontDescAscentGtk(PangoLayout *layout, PangoFontDescription *font_desc,
			      int *ascent, int *descent)
{
    PangoFontMetrics *metrics;
    
    if ((metrics = pango_context_get_metrics(pango_layout_get_context(layout),
					     font_desc, gtk_get_default_language())) != NULL) {
	if (ascent != NULL) {
	    *ascent = PANGO_PIXELS(pango_font_metrics_get_ascent(metrics));
	}
	if (descent != NULL) {
	    *descent = PANGO_PIXELS(pango_font_metrics_get_descent(metrics));
	}
    
	pango_font_metrics_unref(metrics);
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}
#endif

#if GTK_CHECK_VERSION(2,0,0)
static char *localeToUTF8(const char *string)
{
    char *utf8string;

    if (spEqLanguage(spGetLanguage(), "C") == SP_TRUE) {
	return g_strdup(string);
    } else {
	if ((utf8string = (char *)g_locale_to_utf8(string, -1, NULL, NULL, NULL)) == NULL) {
	    return NULL;
	}
    }

    return utf8string;
}
#endif

void spDrawStringArch(spComponent component, spGraphics graphics,
		      int x, int y, const char *string)
{
#if GTK_CHECK_VERSION(2,6,0)
    {
	int ascent, descent;
	double x_shift = 0.0;
	char *tstring;
	PangoRectangle rect;
	PangoFontDescription *font_desc;
	spAlignment alignment;
    
	if ((tstring = localeToUTF8(string)) != NULL) {
	    font_desc = (graphics == NULL ? SpTopLevelArch(sp_toplevel).font_desc :
			 SpGraphicsArch(graphics).font_desc);
	    if (font_desc != NULL) {
		pango_layout_set_font_description(SpPrimitiveArch(component).layout, font_desc);
	    }
	    
	    pango_layout_set_text(SpPrimitiveArch(component).layout, tstring, -1);
	    pango_layout_get_pixel_extents(SpPrimitiveArch(component).layout, NULL, &rect);

#if GTK_CHECK_VERSION(3,0,0) || PANGO_VERSION_CHECK(1,6,0)
	    alignment = SpGraphicsPart(graphics).string_alignment;
#else
	    if (SpGraphicsPart(graphics).string_rotation != 0) {
		alignment = SpGraphicsPart(graphics).rotation_failed_alignment;
	    } else {
		alignment = SpGraphicsPart(graphics).string_alignment;
	    }
#endif
	    
	    if (alignment == SP_ALIGNMENT_CENTER) {
		x_shift = rect.width / 2.0;
	    } else if (alignment == SP_ALIGNMENT_END) {
		x_shift = rect.width;
	    }

            ascent = descent = 0;
	    spGetFontDescAscentGtk(SpPrimitiveArch(component).layout, font_desc, &ascent, &descent);

#if GTK_CHECK_VERSION(3,0,0)
	    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
		spUpdateColorCairo(SpPrimitiveArch(component).pixmap_cr, graphics);
		
		spDebug(80, "spDrawStringArch", "cairo rotation: x_shift = %f, rotation = %d\n",
			x_shift, SpGraphicsPart(graphics).string_rotation);
		
		if (SpGraphicsPart(graphics).string_rotation != 0) {
		    cairo_save(SpPrimitiveArch(component).pixmap_cr);
		    cairo_translate(SpPrimitiveArch(component).pixmap_cr, x, y - rect.height + descent);
		    cairo_rotate(SpPrimitiveArch(component).pixmap_cr,
				 PI * (double)(-SpGraphicsPart(graphics).string_rotation) / 180.0);
		    cairo_move_to(SpPrimitiveArch(component).pixmap_cr, -x_shift, 0.0);
		    pango_cairo_show_layout(SpPrimitiveArch(component).pixmap_cr, SpPrimitiveArch(component).layout);
		    cairo_restore(SpPrimitiveArch(component).pixmap_cr);    
		} else {
		    cairo_move_to(SpPrimitiveArch(component).pixmap_cr, x - x_shift, y - rect.height + descent);
		    pango_cairo_show_layout(SpPrimitiveArch(component).pixmap_cr, SpPrimitiveArch(component).layout);
		}
	    }
#else
#if PANGO_VERSION_CHECK(1,6,0)
	    if (SpGraphicsPart(graphics).string_rotation != 0) {
		GdkScreen *screen;
		PangoRenderer *renderer;
		PangoContext *context;
		PangoMatrix matrix = PANGO_MATRIX_INIT;

		screen = gdk_drawable_get_screen(SpPrimitiveArch(component).pixmap);
		renderer = gdk_pango_renderer_get_default(screen);
		gdk_pango_renderer_set_drawable(GDK_PANGO_RENDERER(renderer),
						SpPrimitiveArch(component).pixmap);
		gdk_pango_renderer_set_gc(GDK_PANGO_RENDERER(renderer),
					  (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
					   SpGraphicsArch(graphics).gc));

		spDebug(80, "spDrawStringArch", "pango rotation: x_shift = %f, rotation = %d\n",
			x_shift, SpGraphicsPart(graphics).string_rotation);
		
		context = pango_layout_get_context(SpPrimitiveArch(component).layout);
		
		pango_matrix_translate(&matrix, x, y);
		pango_matrix_rotate(&matrix, SpGraphicsPart(graphics).string_rotation);

		pango_context_set_matrix(context, &matrix);
		pango_layout_context_changed(SpPrimitiveArch(component).layout);

		pango_renderer_draw_layout(renderer, SpPrimitiveArch(component).layout,
					   -x_shift * PANGO_SCALE, (-rect.height + descent) * PANGO_SCALE);
		
		pango_context_set_matrix(context, NULL);
		pango_layout_context_changed(SpPrimitiveArch(component).layout);
	    }
	    else
#endif
	    {
		gdk_draw_layout(SpPrimitiveArch(component).pixmap,
				(graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
				 SpGraphicsArch(graphics).gc),
				x - x_shift, y - rect.height + descent, SpPrimitiveArch(component).layout);
	    }
#endif
	    g_free(tstring);
	}
    }
#else
    {
	spAlignment alignment;
	GdkFont *font;
	int x_shift = 0;
    
	font = (graphics == NULL ? SpTopLevelArch(sp_toplevel).font :
		SpGraphicsArch(graphics).font);

	if (SpGraphicsPart(graphics).string_rotation != 0) {
	    alignment = SpGraphicsPart(graphics).rotation_failed_alignment;
	} else {
	    alignment = SpGraphicsPart(graphics).string_alignment;
	}
	
	if (alignment == SP_ALIGNMENT_CENTER) {
	    x_shift = gdk_string_measure(font, string) / 2;
	} else if (alignment == SP_ALIGNMENT_END) {
	    x_shift = gdk_string_measure(font, string);
	}
	
	gdk_draw_string(SpPrimitiveArch(component).pixmap, font,
			(graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
			 SpGraphicsArch(graphics).gc),
			x - x_shift, y, string);
    }
#endif
    
    return;
}

spBool spGetStringExtentArch(spComponent component, spGraphics graphics, const char *string,
			     int *x, int *y, int *width, int *height, int *next_x)
{
#if GTK_CHECK_VERSION(2,0,0)
    {
	int ascent, descent;
	char *tstring;
	PangoLayout *layout;
	PangoFontDescription *font_desc;
	PangoRectangle rect, lrect;
	
	if ((tstring = localeToUTF8(string)) == NULL) {
	    return SP_FALSE;
	}

	if (component != NULL && SpPrimitiveArch(component).layout != NULL) {
	    layout = SpPrimitiveArch(component).layout;
	} else {
	    layout = pango_layout_new(SpTopLevelArch(sp_toplevel).context);
	}
	
	font_desc = (graphics == NULL ? SpTopLevelArch(sp_toplevel).font_desc :
		     SpGraphicsArch(graphics).font_desc);
	if (font_desc != NULL) {
	    pango_layout_set_font_description(layout, font_desc);
	}
	pango_layout_set_text(layout, tstring, -1);
	pango_layout_get_pixel_extents(layout, &rect, &lrect);

        ascent = descent = 0;
	spGetFontDescAscentGtk(layout, font_desc, &ascent, &descent);
	
	if (x != NULL) *x = lrect.x;
	if (y != NULL) *y = -ascent;
	if (width != NULL) *width = lrect.width;
	if (height != NULL) *height = lrect.height;
	if (next_x != NULL) {
	    *next_x = lrect.x + lrect.width;
	}
	spDebug(50, "spGetStringExtent",
		"rect.x = %d, rect.y = %d, rect.width = %d, rect.height = %d, ascent = %d, descent = %d\n",
		rect.x, rect.y, rect.width, rect.height, ascent, descent);
	spDebug(50, "spGetStringExtent",
		"lrect.x = %d, lrect.y = %d, lrect.width = %d, lrect.height = %d\n",
		lrect.x, lrect.y, lrect.width, lrect.height);

	if (component != NULL && layout != SpPrimitiveArch(component).layout) {
	    g_object_unref(layout);
	}
	g_free(tstring);
    }
#else
    {
	GdkFont *font;
    
	font = (graphics == NULL ? SpTopLevelArch(sp_toplevel).font :
		SpGraphicsArch(graphics).font);
    
	if (x != NULL) *x = -1;
	if (y != NULL) {
	    *y = -font->ascent - 1;
	}
	if (width != NULL) {
	    *width = gdk_string_measure(font, string) + 1;
	}
	if (height != NULL) {
	    *height = font->ascent + font->descent + 1;
	}
	if (next_x != NULL) {
	    *next_x = gdk_string_width(font, string);;
	}
    }
#endif

    return SP_TRUE;
}

spBool spCreatePolygonArch(spPolygon polygon)
{
    SpPolygonArch(polygon).points = xalloc(polygon->max_point, GdkPoint);
    return SP_TRUE;
}

spBool spDestroyPolygonArch(spPolygon polygon)
{
    xfree(SpPolygonArch(polygon).points);
    return SP_TRUE;
}

spBool spStartPolygonPointArch(spPolygon polygon, int x, int y)
{
    SpPolygonArch(polygon).points[0].x = x;
    SpPolygonArch(polygon).points[0].y = y;
    
    return SP_TRUE;
}

spBool spAddPolygonPointArch(spPolygon polygon, int x, int y)
{
    SpPolygonArch(polygon).points[polygon->num_point].x = x;
    SpPolygonArch(polygon).points[polygon->num_point].y = y;
    
    return SP_TRUE;
}

spBool spAddPolygonRelativePointArch(spPolygon polygon, int dx, int dy)
{
    SpPolygonArch(polygon).points[polygon->num_point].x
	= polygon->px + dx;
    SpPolygonArch(polygon).points[polygon->num_point].y
	= polygon->py + dy;
    
    return SP_TRUE;
}

spBool spEndPolygonPointArch(spPolygon polygon)
{
    return SP_TRUE;
}

spBool spGetPolygonPointArch(spPolygon polygon, int index, int *x, int *y)
{
    if (x != NULL) *x = SpPolygonArch(polygon).points[index].x;
    if (y != NULL) *y = SpPolygonArch(polygon).points[index].y;
    
    return SP_TRUE;
}

spBool spMovePolygonArch(spPolygon polygon, int dx, int dy)
{
    int i;

    for (i = 0; i < polygon->num_point; i++) {
	SpPolygonArch(polygon).points[i].x += dx;
	SpPolygonArch(polygon).points[i].y += dy;
    }
    
    return SP_TRUE;
}

spBool spDrawPolygonGtk(spComponent component, spGraphics graphics, spPolygon polygon, spBool fill_flag)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	int i;

	cairo_move_to(SpPrimitiveArch(component).pixmap_cr,
		      SpPolygonArch(polygon).points[0].x, SpPolygonArch(polygon).points[0].y);
	for (i = 1; i < polygon->num_point; i++) {
	    cairo_line_to(SpPrimitiveArch(component).pixmap_cr,
			  SpPolygonArch(polygon).points[i].x, SpPolygonArch(polygon).points[i].y);
	}
	cairo_line_to(SpPrimitiveArch(component).pixmap_cr,
		      SpPolygonArch(polygon).points[0].x, SpPolygonArch(polygon).points[0].y);

	if (fill_flag) {
	    cairo_fill(SpPrimitiveArch(component).pixmap_cr);
	} else {
	    cairo_stroke(SpPrimitiveArch(component).pixmap_cr);
	}
    } else {
	return SP_FALSE;
    }
#else
    gdk_draw_polygon(SpPrimitiveArch(component).pixmap,
		     (graphics == NULL ? SpTopLevelArch(sp_toplevel).fg_gc :
		      SpGraphicsArch(graphics).gc),
		     fill_flag,
		     SpPolygonArch(polygon).points, polygon->num_point);
#endif
    
    return SP_TRUE;
}

spBool spFillPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    return spDrawPolygonGtk(component, graphics, polygon, SP_TRUE);
}

spBool spDrawPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    return spDrawPolygonGtk(component, graphics, polygon, SP_FALSE);
}

unsigned long spGetDrawingCapsArch(spComponent component)
{
    unsigned long caps;
    
    caps = SP_DRAWING_CAPS_COPY_IMAGE
	| SP_DRAWING_CAPS_FILL_RECTANGLE | SP_DRAWING_CAPS_DRAW_RECTANGLE
	| SP_DRAWING_CAPS_FILL_ARC | SP_DRAWING_CAPS_DRAW_ARC
	| SP_DRAWING_CAPS_FILL_POLYGON | SP_DRAWING_CAPS_DRAW_POLYGON
	| SP_DRAWING_CAPS_DRAW_LINE
	| SP_DRAWING_CAPS_STRING_EXTENT | SP_DRAWING_CAPS_STRING_EXTENT_ACCURATE
	| SP_DRAWING_CAPS_DRAW_WIDE_LINE | SP_DRAWING_CAPS_DRAW_STYLED_LINE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE
	| SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE | SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE
	| SP_DRAWING_CAPS_DRAW_WIDE_POLYGON | SP_DRAWING_CAPS_DRAW_STYLED_POLYGON | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_POLYGON
	| SP_DRAWING_CAPS_DRAW_WIDE_ARC | SP_DRAWING_CAPS_DRAW_STYLED_ARC | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC
	;

#if GTK_CHECK_VERSION(2,6,0) && PANGO_VERSION_CHECK(1,6,0)
    caps |= SP_DRAWING_CAPS_STRING_ROTATION;
#endif

    return caps;
}
