/*
 *	spPluginCanvas.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spPlugin.h>

#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>
#include <sp/spCanvasP.h>

#include <sp/spDrawingPluginP.h>
#include <sp/spPluginCanvasP.h>

static spParamTable sp_plugin_canvas_param_tables[] = {
    {SppDrawFunc, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spPluginCanvas, plugin_canvas.draw_func), NULL},
    {SppDrawData, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spPluginCanvas, plugin_canvas.draw_data), NULL},
    {SppPluginName, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spPluginCanvas, plugin_canvas.plugin_name), NULL},
};

spPluginCanvasClassRec SpPluginCanvasClassRec = {
    /* spObjectClassPart */
    {
	SpPluginCanvas,
	(spObjectClass)&SpComponentClassRec,
	sizeof(spPluginCanvasRec),
	spArraySize(sp_plugin_canvas_param_tables),
	sp_plugin_canvas_param_tables,
	spPluginCanvasPartInit,
	spPluginCanvasPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spPluginCanvasCreate,
	spPluginCanvasDestroy,
	NULL,
	NULL,
    },
    /* spComponentClassPart */
    {
	SP_FALSE,
	SP_FALSE,
	SP_FALSE,
	NULL,
	NULL,
	spPluginCanvasSetSize,
	spPluginCanvasGetSize,
	spPluginCanvasGetClientSize,
	spPluginCanvasSetSensitive,
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spPluginCanvasClassPart */
    {
	0,
    },
};

spComponentClass SpPluginCanvasClass = (spComponentClass)&SpPluginCanvasClassRec;

void spPluginCanvasPartInit(spObject object)
{
    spComponent component = (spComponent)object;

    SpPluginCanvasPart(component).plugin = NULL;
    SpPluginCanvasPart(component).current_page = 0;
    SpPluginCanvasPart(component).num_page = 1;
    SpPluginCanvasPart(component).drawing_started = SP_FALSE;
    
    SpPluginCanvasPart(component).plugin_name = NULL;
    SpPluginCanvasPart(component).draw_func = NULL;
    SpPluginCanvasPart(component).draw_data = NULL;
    SpPluginCanvasPart(component).paper_size = SP_PAPER_SIZE_A4;
    SpPluginCanvasPart(component).paper_orientation = SP_PAPER_ORIENTATION_PORTRAIT;
    SpPluginCanvasPart(component).left_margin_mm = 0.0;
    SpPluginCanvasPart(component).right_margin_mm = 0.0;
    SpPluginCanvasPart(component).top_margin_mm = 0.0;
    SpPluginCanvasPart(component).bottom_margin_mm = 0.0;
    
    return;
}

void spPluginCanvasPartFree(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (SpPluginCanvasPart(component).plugin_name != NULL) {
	xfree(SpPluginCanvasPart(component).plugin_name);
    }
    
    return;
}

spBool spPluginCanvasHostCallback(spPluginRec *rec, spPluginHostCallbackReason reason, void *callback_data, void *user_data)
{
    spComponent component;

    spDebug(30, "spPluginCanvasHostCallback", "in\n");
    
    if ((component = (spComponent)user_data) != NULL) {
        if (SpPluginCanvasPart(component).drawing_started == SP_FALSE
            && (reason == SP_DRAWING_PLUGIN_HOST_CR_DRAW_PAGE || reason == SP_DRAWING_PLUGIN_HOST_CR_DRAW)) {
            spDebug(30, "spPluginCanvasHostCallback", "draw callback in drawing_started == SP_FALSE\n");
        } else {
            if (reason == SP_DRAWING_PLUGIN_HOST_CR_DRAW_PAGE) {
                SpPluginCanvasPart(component).current_page = (int)callback_data;
                spDebug(30, "spPluginCanvasHostCallback", "SP_DRAWING_PLUGIN_HOST_CR_DRAW_PAGE: current_page = %d\n",
                        SpPluginCanvasPart(component).current_page);
            }

            if (SpPluginCanvasPart(component).draw_func != NULL) {
                spDebug(30, "spPluginCanvasHostCallback", "call draw_func of spPluginCanvas\n");
                SpPluginCanvasPart(component).draw_func(component, SpPluginCanvasPart(component).draw_data);
            }
        
            if (reason == SP_DRAWING_PLUGIN_HOST_CR_DRAW) {
                ++SpPluginCanvasPart(component).current_page;
                spDebug(30, "spPluginCanvasHostCallback", "SP_DRAWING_PLUGIN_HOST_CR_DRAW: updated: current_page = %d\n",
                        SpPluginCanvasPart(component).current_page);
            }
        }
    }
    
    spDebug(30, "spPluginCanvasHostCallback", "done\n");
    
    return SP_TRUE;
}

spBool spPluginCanvasCreate(spObject object)
{
    char *filename;
    spBool status;
    spComponent parent;
    spComponent component = (spComponent)object;

    if (component == NULL) return SP_FALSE;

    if (!strnone(SpPluginCanvasPart(component).plugin_name)) {
	filename = SpPluginCanvasPart(component).plugin_name;
    } else {
	filename = SpGetName(component);
    }
    spDebug(50, "spPluginCanvasCreate", "filename = %s\n", filename);

    SpPluginCanvasPart(component).plugin = spLoadPlugin(filename);

    status = SP_FALSE;
    if (SpPluginCanvasPart(component).plugin != NULL) {
	if (spIsDrawingPlugin(SpPluginCanvasPart(component).plugin) == SP_TRUE) {
	    if (spInitPluginInstance(SpPluginCanvasPart(component).plugin) == SP_TRUE) {
		spBool create_status;
		
		spDebug(10, "spPluginCanvasCreate", "instantiated\n");

		if (spGetPluginCaps(SpPluginCanvasPart(component).plugin) & SP_DRAWING_PLUGIN_CAPS_CALLBACK_DRAW) {
		    spSetPluginHostCallback(SpPluginCanvasPart(component).plugin,
					    spPluginCanvasHostCallback, component);
		}
		
		parent = spGetParent(component);
		
		if (parent != NULL && SpPluginCanvasDrawingRec(component)->set_parent != NULL) {
		    SpPluginCanvasDrawingRec(component)->set_parent(SpPluginCanvasInstance(component), parent);
		}
		
		if (SpPluginCanvasDrawingRec(component)->create_component != NULL) {
		    int num_arg = 0;
		    spArg args[SP_MAX_NUM_ARG];
		    void *parent_handle = NULL;
		    if (parent != NULL) parent_handle = spGetPrimitiveHandleArch(parent);
		    
		    create_status  = SpPluginCanvasDrawingRec(component)->create_component(SpPluginCanvasInstance(component),
											   parent_handle,
											   args, num_arg);
		} else {
		    create_status = SP_TRUE;
		}
		
		if (create_status == SP_TRUE) {
		    spDebug(50, "spPluginCanvasCreate", "call_func = %lx, draw_func = %lx\n",
                            SpComponentPart(component).call_func, SpPluginCanvasPart(component).draw_func);
		    if (SpComponentPart(component).call_func != NULL) {
			SpPluginCanvasPart(component).draw_func = SpComponentPart(component).call_func;
			SpPluginCanvasPart(component).draw_data = SpComponentPart(component).call_data;
		    }
		    status = SP_TRUE;
		} else {
		    spDebug(10, "spPluginCanvasCreate", "error: cannot create component\n");
		}
	    } else {
		spDebug(10, "spPluginCanvasCreate", "error: cannot instantiate plugin\n");
	    }
	} else {
	    spDebug(10, "spPluginCanvasCreate", "error: %s is not drawing plugin\n", filename);
	}

	if (status == SP_FALSE) {
	    spFreePlugin(SpPluginCanvasPart(component).plugin);
	    SpPluginCanvasPart(component).plugin = NULL;
	}
    } else {
	spDebug(50, "spPluginCanvasCreate", "cannot open plugin %s\n", filename);
    }
    
    spDebug(50, "spPluginCanvasCreate", "done: status = %d\n", status);
    
    return status;
}

spBool spPluginCanvasDestroy(spObject object)
{
    spComponent component = (spComponent)object;

    if (SpPluginCanvasPart(component).plugin != NULL) {
	if (SpPluginCanvasDrawingRec(component)->destroy_component != NULL) {
	    SpPluginCanvasDrawingRec(component)->destroy_component(SpPluginCanvasInstance(component));
	}

	if (spGetPluginRefCount(SpPluginCanvasPart(component).plugin) <= 1) {
	    spDestroyPluginPolygonInstanceInPlugin(SpPluginCanvasPart(component).plugin);
	}
	
	spFreePlugin(SpPluginCanvasPart(component).plugin);
    }
    
    return SP_TRUE;
}

spBool spPluginCanvasSetSize(spComponent component, int width, int height)
{
    if (spIsPluginCanvasCreated(component) == SP_FALSE
	|| SpPluginCanvasDrawingRec(component)->set_size == NULL) return SP_FALSE;
    
    return SpPluginCanvasDrawingRec(component)->set_size(SpPluginCanvasInstance(component),
							 width, height);
}

spBool spPluginCanvasGetSize(spComponent component, int *width, int *height)
{
    return spPluginCanvasGetClientSize(component, width, height);
}

spBool spPluginCanvasGetClientSize(spComponent component, int *width, int *height)
{
    if (spIsPluginCanvasDrawable(component) == SP_FALSE
	|| SpPluginCanvasDrawingRec(component)->get_client_size == NULL) return SP_FALSE;
    
    spDebug(80, "spPluginCanvasGetClientSize", "current size: %d, %d\n",
	    SpComponentPart(component).client_width, SpComponentPart(component).client_height);
    
    return SpPluginCanvasDrawingRec(component)->get_client_size(SpPluginCanvasInstance(component),
								width, height);
}

spBool spPluginCanvasSetSensitive(spComponent component, spBool flag)
{
    return SP_FALSE;
}

spBool spIsPluginCanvas(spComponent component)
{
    return spIsSubClass(component, SpPluginCanvas);
}

spBool spIsPluginCanvasCreated(spComponent component)
{
    if (spIsSubClass(component, SpPluginCanvas) == SP_TRUE
	&& SpPluginCanvasPart(component).plugin != NULL
	&& SpPluginCanvasInstance(component) != NULL) {
	spDebug(100, "spIsPluginCanvasCreated", "already created\n");
	return SP_TRUE;
    } else {
	spDebug(100, "spIsPluginCanvasCreated", "not created\n");
	return SP_FALSE;
    }
}

spBool spIsPluginCanvasDrawable(spComponent component)
{
    if (spIsSubClass(component, SpPluginCanvas) == SP_TRUE
	&& SpPluginCanvasPart(component).plugin != NULL
	&& SpPluginCanvasInstance(component) != NULL
	&& SpPluginCanvasDrawingRec(component)->is_drawable != NULL
	&& SpPluginCanvasDrawingRec(component)->is_drawable(SpPluginCanvasInstance(component)) == SP_TRUE) {
	spDebug(100, "spIsPluginCanvasDrawable", "drawable\n");
	return SP_TRUE;
    } else {
	spDebug(100, "spIsPluginCanvasDrawable", "non-drawable\n");
	return SP_FALSE;
    }
}

spBool spIsPluginCanvasDrawingStarted(spComponent component)
{
    if (spIsPluginCanvasDrawable(component) == SP_TRUE
	&& (SpPluginCanvasDrawingRec(component)->start_drawing == NULL
	    || SpPluginCanvasPart(component).drawing_started == SP_TRUE)) {
	spDebug(100, "spIsPluginCanvasDrawingStarted", "started\n");
	return SP_TRUE;
    } else {
	spDebug(100, "spIsPluginCanvasDrawingStarted", "not started\n");
	return SP_FALSE;
    }
}

const char *spGetPluginCanvasName(spComponent component)
{
    char *name = NULL;
    
    if (spIsSubClass(component, SpPluginCanvas) == SP_TRUE) {
	if (!strnone(SpPluginCanvasPart(component).plugin_name)) {
	    name = SpPluginCanvasPart(component).plugin_name;
	} else {
	    name = SpGetName(component);
	}
    }

    return name;
}

spComponent spCreatePluginCanvasArg(spComponent parent, const char *name, spArg *args, int num_arg)
{
    spDebug(80, "spCreatePluginCanvasArg", "parent = %ld\n", (long)parent);
#if 0
    if (spIsCreated(parent) == SP_FALSE)
	return NULL;
#endif
    
    return spCreateComponentArg(SpPluginCanvasClass, NULL, name, parent, args, num_arg);
}

spComponent spCreatePluginCanvas(spComponent parent, const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;

#if 0
    if (spIsCreated(parent) == SP_FALSE)
	return NULL;
#endif
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);
    
    return spCreateComponentArg(SpPluginCanvasClass, NULL, name, parent, args, num_arg);
}

struct _spPluginPolygonInstance {
    spPolygon polygon;
    
    char *name;
    void *instance;
    spPluginPolygonInstance prev;
    spPluginPolygonInstance next;

    /* to prevent plugin from leaking memory */
    spDrawingPluginRec *plugin_rec;
    spPluginPolygonInstance prev_in_plugin;
    spPluginPolygonInstance next_in_plugin;
};

void *spGetPluginPolygonInstance(spPolygon polygon, spPlugin *plugin, spBool *new_flag)
{
    spPluginPolygonInstance instance, last;
    void *plugin_instance;
    spDrawingPluginRec *rec;
    
    if (polygon == NULL || spIsDrawingPlugin(plugin) == SP_FALSE) return NULL;

    spDebug(80, "spGetPluginPolygonInstance", "in\n");
    
    rec = (spDrawingPluginRec *)plugin->rec;
    if (rec->init_polygon_instance == NULL) {
	if (new_flag != NULL) *new_flag = SP_FALSE;
	spDebug(80, "spGetPluginPolygonInstance", "rec->init_polygon_instance == NULL, return\n");
	return polygon;
    }
    
    spDebug(80, "spGetPluginPolygonInstance", "before loop\n");
    
    last = polygon->instance;
    while (last != NULL) {
	if (last->plugin_rec == rec) {
	    /* already added */
	    if (new_flag != NULL) *new_flag = SP_FALSE;
	    spDebug(80, "spGetPluginPolygonInstance", "already added, return\n");
	    return last->instance;
	} else if (last->next == NULL) {
	    break;
	}
	last = last->next;
    }

    spDebug(80, "spGetPluginPolygonInstance", "after loop: last = %ld\n", (long)last);
    
    if ((plugin_instance = rec->init_polygon_instance(polygon))	== NULL) {
	spDebug(80, "spGetPluginPolygonInstance", "rec->init_polygon_instance returns NULL, return\n");
	return NULL;
    }

    spDebug(80, "spGetPluginPolygonInstance", "call rec->init_polygon_instance done: plugin_instance = %ld\n",
	    (long)plugin_instance);
    
    instance = xalloc(1, struct _spPluginPolygonInstance);
    instance->polygon = polygon;
    instance->name = rec->plugin_id;
    instance->instance = plugin_instance;
    instance->prev = last;
    instance->next = NULL;
    instance->plugin_rec = rec;
    instance->prev_in_plugin = NULL;
    instance->next_in_plugin = NULL;

    if (last == NULL) {
	polygon->instance = instance;
    } else {
	last->next = instance;
    }

    /* add this to global polygon instance */
    last = (spPluginPolygonInstance)rec->polygon_host_data;
    spDebug(80, "spGetPluginPolygonInstance", "add this to global polygon instance: last = %ld\n", (long)last);
    if (last == NULL) {
	rec->polygon_host_data = instance;
    } else {
	while (last != NULL) {
	    spDebug(80, "spGetPluginPolygonInstance", "in loop: last = %ld\n", (long)last);
	    if (last->next_in_plugin == NULL) {
		spDebug(80, "spGetPluginPolygonInstance", "last->next_in_plugin == NULL, break\n");
		last->next_in_plugin = instance;
		instance->prev_in_plugin = last;
		break;
	    }
	    last = last->next_in_plugin;
	}
    }

    if (new_flag != NULL) *new_flag = SP_TRUE;
    
    spDebug(80, "spGetPluginPolygonInstance", "done\n");
    
    return plugin_instance;
}

static spBool freePluginPolygonInstance(spPluginPolygonInstance instance)
{
    spBool flag = SP_FALSE;

    if (instance != NULL) {
	if (instance->prev != NULL) {
	    instance->prev->next = instance->next;
	} else if (instance->polygon != NULL) {
	    instance->polygon->instance = instance->next;
	}
	if (instance->next != NULL) {
	    instance->next->prev = instance->prev;
	}

	if (instance->prev_in_plugin != NULL) {
	    instance->prev_in_plugin->next_in_plugin = instance->next_in_plugin;
	} else {
	    instance->plugin_rec->polygon_host_data = instance->next_in_plugin;
	}
	if (instance->next_in_plugin != NULL) {
	    instance->next_in_plugin->prev_in_plugin = instance->prev_in_plugin;
	}

	if (instance->plugin_rec->free_polygon_instance != NULL) {
	    flag = instance->plugin_rec->free_polygon_instance(instance->instance);
	}

	xfree(instance);
    }

    return flag;
}

spBool spDestroyPluginPolygonInstanceInPlugin(spPlugin *plugin)
{
    spDrawingPluginRec *rec;
    spPluginPolygonInstance next, last;

    if (spIsDrawingPlugin(plugin) == SP_FALSE) return SP_FALSE;
    
    rec = (spDrawingPluginRec *)plugin->rec;

    last = (spPluginPolygonInstance)rec->polygon_host_data;
    while (last != NULL) {
	next = last->next_in_plugin;
	freePluginPolygonInstance(last);
	last = next;
    }

    return SP_TRUE;
}

spBool spDestroyPluginPolygonInstance(spPolygon polygon)
{
    spPluginPolygonInstance next, last;

    if (polygon == NULL) return SP_FALSE;
    
    last = polygon->instance;
    while (last != NULL) {
	next = last->next;
	freePluginPolygonInstance(last);
	last = next;
    }

    return SP_TRUE;
}

void *spGetPluginCanvasPolygonInstance(spComponent component, spPolygon polygon)
{
    int i;
    int x, y;
    void *polygon_instance;
    spBool new_flag = SP_FALSE;
    
    if (polygon == NULL || polygon->end_flag == SP_FALSE
	|| spIsPluginCanvas(component) == SP_FALSE) return NULL;

    spDebug(80, "spGetPluginCanvasPolygonInstance", "in\n");
    
    if ((polygon_instance = spGetPluginPolygonInstance(polygon,
					SpPluginCanvasPart(component).plugin, &new_flag)) == NULL) {
	spDebug(10, "spGetPluginCanvasPolygonInstance", "spGetPluginPolygonInstance failed\n");
	return NULL;
    }

    spDebug(80, "spGetPluginCanvasPolygonInstance", "new_flag = %d\n", new_flag);
    
    if (new_flag == SP_TRUE) {
	if (SpPluginCanvasDrawingRec(component)->polygon_event != NULL) {
	    spDebug(50, "spGetPluginCanvasPolygonInstance", "num_point = %d\n", polygon->num_point);
	    for (i = 0; i < polygon->num_point; i++) {
		if (spGetPolygonPoint(polygon, i, &x, &y, SP_TRUE) == SP_FALSE) {
		    break;
		}
		spDebug(50, "spGetPluginCanvasPolygonInstance", "i = %d, x = %d, y = %d\n", i, x, y);
		if (i == 0) {
		    SpPluginCanvasDrawingRec(component)->polygon_event(SpPluginCanvasInstance(component), polygon_instance,
								       SP_POLYGON_EVENT_START_POINT, x, y);
		} else {
		    SpPluginCanvasDrawingRec(component)->polygon_event(SpPluginCanvasInstance(component), polygon_instance,
								       SP_POLYGON_EVENT_ADD_POINT, x, y);
		}
	    }
    
	    SpPluginCanvasDrawingRec(component)->polygon_event(SpPluginCanvasInstance(component), polygon_instance,
							       SP_POLYGON_EVENT_END_POINT, 0, 0);
	}
    }

    if (SpPluginCanvasDrawingRec(component)->polygon_event != NULL
	&& (polygon->move_dx != 0 || polygon->move_dy != 0)) {
	spDebug(50, "spGetPluginCanvasPolygonInstance", "MOVE: dx = %d, dy = %d\n", polygon->move_dx, polygon->move_dy);
	SpPluginCanvasDrawingRec(component)->polygon_event(SpPluginCanvasInstance(component), polygon_instance,
							   SP_POLYGON_EVENT_MOVE, polygon->move_dx, polygon->move_dy);
    }
    
    return polygon_instance;
}
