/*
 *        spPrimitive_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#undef GTK_DISABLE_DEPRECATED

#include <gdk/gdkkeysyms.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spThread.h>

#include <sp/spComponent.h>
#include <sp/spFrame.h>
#include <sp/spDialog.h>
#include <sp/spButton.h>
#include <sp/spMenuItem.h>
#include <sp/spMenu.h>
#include <sp/spToolBar.h>
#include <sp/spTabBox.h>

#include <sp/spTopLevelP.h>
#include <sp/spFrameP.h>
#include <sp/spComboBoxP.h>
#include <sp/spCanvasP.h>
#include <sp/spContainerP.h>
#include <sp/spGraphicsP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spSliderP.h>
#include <sp/spToolBarP.h>
#include <sp/spToolItemP.h>
#include <sp/spStatusBarP.h>
#include <sp/spParamFieldP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>

extern spTopLevel sp_toplevel;

static spCallbackTable sp_callback_tables[] = {
    {SP_NO_CALLBACK, NULL, NULL, 0},
    {SP_KEY_PRESS_CALLBACK, SpCanvas, "key_press_event", GDK_KEY_PRESS_MASK},
    {SP_KEY_PRESS_CALLBACK, SpCanvas, "button_press_event", GDK_KEY_PRESS_MASK},
    {SP_KEY_RELEASE_CALLBACK, SpCanvas, "key_release_event", GDK_KEY_RELEASE_MASK},
    {SP_BUTTON_PRESS_CALLBACK, SpPrimitive, "button_press_event", GDK_BUTTON_PRESS_MASK},
    {SP_BUTTON_RELEASE_CALLBACK, SpPrimitive, "button_release_event", GDK_BUTTON_RELEASE_MASK},
    {SP_POINTER_MOTION_CALLBACK, SpCanvas, "motion_notify_event", GDK_POINTER_MOTION_MASK},
    {SP_BUTTON_MOTION_CALLBACK, SpCanvas, "motion_notify_event", GDK_BUTTON_MOTION_MASK},
#if 0
    {SP_ENTER_WINDOW_CALLBACK, SpPrimitive, "enter_notify_event", GDK_ENTER_NOTIFY_MASK},
    {SP_LEAVE_WINDOW_CALLBACK, SpPrimitive, "leave_notify_event", GDK_LEAVE_NOTIFY_MASK},
#endif
    /**/
    {SP_DESTROY_CALLBACK, SpFrame, SppDestroyCallback, 0},
    {SP_CLOSE_CALLBACK, SpFrame, SppCloseCallback, 0},
    {SP_CLOSE_CALLBACK, SpFrame, "destroy", 0},
    {SP_ACTIVATE_CALLBACK, SpPushButton, "clicked", 0},
    {SP_ACTIVATE_CALLBACK, SpMenuItem, "activate", 0},
    {SP_ACTIVATE_CALLBACK, SpToolItem, "clicked", 0},
    {SP_ACTIVATE_CALLBACK, SpTextField, "activate", 0},
    {SP_ACTIVATE_CALLBACK, SpList, "button_press_event", GDK_BUTTON_PRESS_MASK},
    {SP_VALUE_CHANGED_CALLBACK, SpCheckBox, "toggled", 0},
    {SP_VALUE_CHANGED_CALLBACK, SpRadioButton, "toggled", 0},
    {SP_VALUE_CHANGED_CALLBACK, SpCheckBoxMenuItem, "toggled", 0},
    {SP_VALUE_CHANGED_CALLBACK, SpRadioButtonMenuItem, "toggled", 0},
    {SP_VALUE_CHANGED_CALLBACK, SpCheckToolItem, "toggled", 0},
    {SP_VALUE_CHANGED_CALLBACK, SpText, "changed", 0},
#if GTK_CHECK_VERSION(3,0,0)
    {SP_VALUE_CHANGED_CALLBACK, SpSlider, "button-release-event", 0},
#endif
    {SP_VALUE_CHANGED_CALLBACK, SpSlider, SP_SLIDER_VALUE_CHANGED_CALLBACK_NAME, 0},
#if GTK_CHECK_VERSION(3,0,0)
    {SP_VALUE_CHANGED_CALLBACK, SpList, "cursor-changed", 0},
#else
    {SP_VALUE_CHANGED_CALLBACK, SpList, "selection_changed", 0},
#endif
#if 0
    {SP_MAP_CALLBACK, SpPrimitive, "map", 0},
    {SP_UNMAP_CALLBACK, SpPrimitive, "unmap", 0},
#endif
#if GTK_CHECK_VERSION(3,0,0)
    {SP_EXPOSE_CALLBACK, SpCanvas, "draw", 0},
#else
    {SP_EXPOSE_CALLBACK, SpCanvas, "expose_event", GDK_EXPOSURE_MASK},
#endif
    {SP_RESIZE_CALLBACK, SpCanvas, "configure_event", GDK_STRUCTURE_MASK},
    {SP_RESIZE_CALLBACK, SpFrame, "configure_event", GDK_STRUCTURE_MASK},
    /**/
    {SP_OK_CALLBACK, SpPushButton, "clicked", 0},
    {SP_CANCEL_CALLBACK, SpPushButton, "clicked", 0},
    {SP_HELP_CALLBACK, SpPushButton, "clicked", 0},
#if 0
    {SP_APPLY_CALLBACK, SpPushButton, "clicked", 0},
#endif
    {SP_SCROLL_CALLBACK, SpCanvas, "value_changed", 1},
    {SP_SCROLL_CALLBACK, SpCanvas, "value_changed", 2},
#if GTK_CHECK_VERSION(2,0,0)
    {SP_WHEEL_CALLBACK, SpCanvas, "scroll_event", GDK_SCROLL_MASK},
    {SP_ZOOM_CALLBACK, SpCanvas, SppZoomCallback, GDK_SCROLL_MASK},
#endif
};
static int sp_num_callback_table = 0;

gboolean spIsWidgetMappedGtk(GtkWidget *widget)
{
    gboolean flag;

    if (widget == NULL) return FALSE;
    
#if GTK_CHECK_VERSION(2,20,0)
    flag = gtk_widget_get_mapped(widget);
#else
    flag = GTK_WIDGET_MAPPED(widget);
#endif

    return flag;
}

gboolean spIsWidgetRealizedGtk(GtkWidget *widget)
{
    gboolean flag;
    
#if GTK_CHECK_VERSION(2,20,0)
    flag = gtk_widget_get_realized(widget);
#else
    flag = GTK_WIDGET_REALIZED(widget);
#endif

    return flag;
}

gboolean spWidgetHasGrabGtk(GtkWidget *widget)
{
    gboolean flag;
    
#if GTK_CHECK_VERSION(2,18,0)
    flag = gtk_widget_has_grab(widget);
#else
    flag = GTK_WIDGET_HAS_GRAB(widget);
#endif

    return flag;
}

GdkWindow *spGetGdkWindowGtk(GtkWidget *widget)
{
    GdkWindow *window;
    
#if GTK_CHECK_VERSION(2,14,0)
    window = gtk_widget_get_window(widget);
#else
    window = widget->window;
#endif

    return window;
}

gint spGetAllocationWidthGtk(GtkWidget *widget)
{
#if GTK_CHECK_VERSION(3,0,0)
#if 0
    gint width, minimum_width, natural_width;
    
    width = gtk_widget_get_allocated_width(widget);
    minimum_width = natural_width = 0;
    gtk_widget_get_preferred_width(widget, &minimum_width, &natural_width);
    spDebug(100, "spGetAllocationWidthGtk", "width = %d, minimum_width = %d, natural_width = %d\n",
            width, minimum_width, natural_width);
    
    if (width <= 1 && natural_width > 1) {
        return natural_width;
    } else {
        return width;
    }
#else
    return gtk_widget_get_allocated_width(widget);
#endif
#else
    return widget->allocation.width;
#endif
}

gint spGetAllocationHeightGtk(GtkWidget *widget)
{
#if GTK_CHECK_VERSION(3,0,0)
#if 0
    gint height, minimum_height, natural_height;
    
    height = gtk_widget_get_allocated_height(widget);
    minimum_height = natural_height = 0;
    gtk_widget_get_preferred_height(widget, &minimum_height, &natural_height);
    spDebug(100, "spGetAllocationHeightGtk", "height = %d, minimum_height = %d, natural_height = %d\n",
            height, minimum_height, natural_height);
    
    if (height <= 1 && natural_height > 1) {
        return natural_height;
    } else {
        return height;
    }
#else
    return gtk_widget_get_allocated_height(widget);
#endif
#else
    return widget->allocation.height;
#endif
}

gint spSignalConnectCallbackGtk(GtkWidget *widget, const gchar *call_name, spCallbackFuncGtk func, gpointer data)
{
#if GTK_CHECK_VERSION(3,0,0)
    return g_signal_connect(widget, call_name, G_CALLBACK(func), data);
#elif GTK_CHECK_VERSION(2,0,0)
    return g_signal_connect(GTK_OBJECT(widget), call_name, G_CALLBACK(func), data);
#else
    return gtk_signal_connect(GTK_OBJECT(widget), call_name, GTK_SIGNAL_FUNC(func), data);
#endif
}

gint spSignalConnectEventGtk(GtkWidget *widget, const gchar *call_name, spEventFuncGtk func, gpointer data)
{
#if GTK_CHECK_VERSION(3,0,0)
    return g_signal_connect(widget, call_name, G_CALLBACK(func), data);
#elif GTK_CHECK_VERSION(2,0,0)
    return g_signal_connect(GTK_OBJECT(widget), call_name, G_CALLBACK(func), data);
#else
    return gtk_signal_connect(GTK_OBJECT(widget), call_name, GTK_SIGNAL_FUNC(func), data);
#endif
}

void spEmitComponentCallbackFuncGtk(spCallback *callback, GdkEvent *event)
{
    long component_id;
    
    component_id = SpGetComponentId(callback->component);
    SpPrimitiveArch(callback->component).call_name = callback->call_name;
    SpPrimitiveArch(callback->component).event = event;
            
    spComponentCallbackFunc(callback);
            
    if (component_id != spGetDestroyedComponentId()
        && callback != NULL && callback->component != NULL) {
        SpPrimitiveArch(callback->component).call_name = NULL;
        SpPrimitiveArch(callback->component).event = NULL;
    }

    return;
}

static void callbackFunc(GtkWidget *widget, gpointer data)
{
    gdouble value;
    spCallback *callback = (spCallback *)data;

    if (callback != NULL && callback->call_func != NULL
        && callback->component != NULL && callback->call_name != NULL) {
        spDebug(50, "callbackFunc", "callbacked: %s\n", callback->call_name);

        if (streq(callback->call_name, "destroy")
            && spIsCreated(callback->component) == SP_FALSE) {
            /* not closing window, but destroying window */
            return;
        } else if (streq(callback->call_name, "activate")
                   || streq(callback->call_name, "toggled")) {
            if (spIsMenuItem(callback->component) == SP_TRUE) {
                spDebug(80, "callbackFunc", "activate signal of menu item received\n");
                if (spIsSensitive(callback->component) == SP_FALSE) {
                    /* sometimes gtk send messages to insensitive widgets through shortcuts */
                    return;
                }
            }
        } else if (spIsCanvas(callback->component) == SP_TRUE && streq(callback->call_name, "value_changed")) {
            spDebug(80, "callbackFunc", "scroll callback received\n");
            if (callback->event_mask == 1 && SpCanvasArch(callback->component).h_adjustment != NULL) {
                spDebug(80, "callbackFunc", "horizontal scroll callback received\n");
#if GTK_CHECK_VERSION(2,0,0)
                value = gtk_adjustment_get_value(SpCanvasArch(callback->component).h_adjustment);
#else
                value = GTK_ADJUSTMENT(SpCanvasArch(callback->component).h_adjustment)->value;
#endif
                SpCanvasPart(callback->component).h_value = (int)spRound(value / SpCanvasPart(callback->component).h_content_factor);
            } else if (callback->event_mask == 2 && SpCanvasArch(callback->component).v_adjustment != NULL) {
                spDebug(80, "callbackFunc", "vertical scroll callback received\n");
#if GTK_CHECK_VERSION(2,0,0)
                value = gtk_adjustment_get_value(SpCanvasArch(callback->component).v_adjustment);
#else
                value = GTK_ADJUSTMENT(SpCanvasArch(callback->component).v_adjustment)->value;
#endif
                SpCanvasPart(callback->component).v_value = (int)spRound(value / SpCanvasPart(callback->component).v_content_factor);
            } else {
                return;
            }
        } else if (spIsSlider(callback->component) == SP_TRUE && streq(callback->call_name, SP_SLIDER_VALUE_CHANGED_CALLBACK_NAME)) {
            int value;

            value = spGetSliderValueArch(callback->component);
            if (SpPrimitiveArch(callback->component).prev_value == value) {
                return;
            }
            SpPrimitiveArch(callback->component).prev_value = value;
        }

        if (!SpPrimitiveArch(callback->component).state) {
            spEmitComponentCallbackFuncGtk(callback, NULL);
            SpPrimitiveArch(callback->component).state = 0;
        }
        /*spDebug(50, "callbackFunc", "done: %s\n", callback->call_name);*/
    }
    
    return;
}

static gint eventFunc(GtkWidget *widget, GdkEvent *event, gpointer data)
{
    gint flag = FALSE;
    spCallback *callback = (spCallback *)data;

     if (callback != NULL && callback->call_func != NULL
        && callback->component != NULL && callback->call_name != NULL) {
        spDebug(50, "eventFunc", "callbacked: %s %s, event->type = %d\n",
                SpGetClassName(callback->component), callback->call_name, event->type);
        if (streq(callback->call_name, "motion_notify_event")
            && callback->event_mask == GDK_BUTTON_MOTION_MASK
            && !event->motion.state) {
            return FALSE;
        } else if (streq(callback->call_name, "button_press_event")
                   && callback->event_mask == GDK_KEY_PRESS_MASK
                   && !(event->button.button == 4 || event->button.button == 5)) {
            return FALSE;
        }
        
        if (event->type == GDK_CONFIGURE) {
            spDebug(50, "eventFunc", "width = %d, height = %d, geometry_flag = %d\n",
                    event->configure.width, event->configure.height,
                    SpComponentPart(callback->component).geometry_flag);
            if (SpComponentPart(callback->component).geometry_flag == SP_TRUE) {
                spDebug(50, "eventFunc", "geometry_flag == SP_TRUE, return FALSE\n");
                return FALSE;
            } else if (spIsCanvas(callback->component) == SP_TRUE) {
                spDebug(50, "eventFunc", "widget is canvas\n");
                if (!spIsWidgetRealizedGtk(SpFrameArch(SpGetWindow(callback->component)).toplevel)) {
                    spDebug(50, "eventFunc", "canvas has not been realized, return FALSE\n");
                    return FALSE;
                }
            }
#if GTK_CHECK_VERSION(2,0,0)
        } else if (event->type == GDK_SCROLL) {
            spDebug(50, "eventFunc", "GDK_SCROLL received: state = %x, call_name = %s\n",
                    event->scroll.state, callback->call_name);
            
            if (event->scroll.state & GDK_CONTROL_MASK) {
                double relative_factor;
                
                spDebug(50, "eventFunc", "GDK_SCROLL with control key\n");
                
                if (streq(callback->call_name, "scroll_event")) {
                    spDebug(50, "eventFunc", "GDK_SCROLL with control key, but callback is not for zoom event\n");
                    return FALSE;
                }
                if (event->scroll.direction == GDK_SCROLL_UP || event->scroll.direction == GDK_SCROLL_RIGHT) {
                    relative_factor = 2.0;
                } else {
                    relative_factor = 0.5;
                }
                spUpdateCanvasContentFactor(callback->component, relative_factor, SP_TRUE);
                flag = TRUE;
            } else {
                GtkAdjustment *adjustment;
                int increment;
                int *pvalue;
                gdouble factor;
                gdouble old_value, new_value;
                int horizontal_flag;
                
                if (streq(callback->call_name, SppZoomCallback)
                    || (SpCanvasPart(callback->component).accept_wheel_event == SP_FALSE
                        && spIsScrollCanvas(callback->component) == SP_FALSE)) {
                    spDebug(50, "eventFunc", "GDK_SCROLL for SppZoomCallback, return\n");
                    return FALSE;
                }

                horizontal_flag = (event->scroll.state & GDK_SHIFT_MASK);
                if (SpCanvasPart(callback->component).wheel_target_horizontal == SP_TRUE) {
                    spDebug(80, "eventFunc", "WHEEL_TARGET IS HORIZONTAL\n");
                    horizontal_flag = !horizontal_flag;
                }
                
                if (horizontal_flag) {
                    adjustment = SpCanvasArch(callback->component).h_adjustment;
                    increment = SpCanvasPart(callback->component).h_wheel_increment;
                    pvalue = &SpCanvasPart(callback->component).h_value;
                    factor = SpCanvasPart(callback->component).h_content_factor;
                    if (SpCanvasPart(callback->component).scroll_left_by_wheel_down == SP_TRUE) {
                        increment *= -1;
                    }
                } else {
                    adjustment = SpCanvasArch(callback->component).v_adjustment;
                    increment = SpCanvasPart(callback->component).v_wheel_increment;
                    pvalue = &SpCanvasPart(callback->component).v_value;
                    factor = SpCanvasPart(callback->component).v_content_factor;
                }
                
                if (adjustment != NULL) {
                    old_value = gtk_adjustment_get_value(adjustment);
                
                    if (event->scroll.direction == GDK_SCROLL_UP || event->scroll.direction == GDK_SCROLL_LEFT) {
                        new_value = old_value - increment;
                    } else if (event->scroll.direction == GDK_SCROLL_RIGHT || event->scroll.direction == GDK_SCROLL_DOWN) {
                        new_value = old_value + increment;
                    } else {
                        return FALSE;
                    }

                    spCorrectAdjustmentValueGtk(adjustment, &new_value);
                    spDebug(50, "eventFunc", "old_value = %f, new_value = %f\n", old_value, new_value);
                
                    *pvalue = (int)spRound(new_value / factor);
                    gtk_adjustment_set_value(adjustment, new_value);
                    flag = TRUE;
                }
            }
#endif
        }
        
        if (spIsSubClass(callback->component, SpList) == SP_FALSE
            || (streq(callback->call_name, "button_press_event") && event->type == GDK_2BUTTON_PRESS)) {
            spEmitComponentCallbackFuncGtk(callback, event);
        }
    }
    
     spDebug(100, "eventFunc", "done\n");
     
    return flag;
}

#if GTK_CHECK_VERSION(3,0,0)
#if GTK_CHECK_VERSION(3,16,0)
static gboolean resizeFunc(GtkGLArea *area, gint width, gint height, gpointer data)
{
    spCallback *callback = (spCallback *)data;
    
    if (callback != NULL && callback->call_func != NULL
        && callback->component != NULL && callback->call_name != NULL) {
        spDebug(50, "resizeFunc", "callbacked: %s\n", callback->call_name);

        /* resizeFunc will be called just before renderFunc */
        SpCanvasArch(callback->component).in_draw_func = SP_TRUE;
        SpCanvasArch(callback->component).skip_draw_func = SP_TRUE;
        
        spEmitComponentCallbackFuncGtk(callback, NULL);
        
        SpCanvasArch(callback->component).in_draw_func = SP_FALSE;
        
        spDebug(50, "resizeFunc", "processed: %s\n", callback->call_name);
    }

    return TRUE;
}

static gboolean renderFunc(GtkGLArea *area, GdkGLContext *context, gpointer data)
{
    spCallback *callback = (spCallback *)data;
    
    if (callback != NULL && callback->call_func != NULL
        && callback->component != NULL && callback->call_name != NULL) {
        spDebug(50, "renderFunc", "callbacked: %s, skip_draw_func = %d\n",
                callback->call_name, SpCanvasArch(callback->component).skip_draw_func);

        SpCanvasArch(callback->component).in_draw_func = SP_TRUE;
        
        if (SpCanvasArch(callback->component).skip_draw_func == SP_FALSE) {
            spEmitComponentCallbackFuncGtk(callback, NULL);
        }

        SpCanvasArch(callback->component).skip_draw_func = SP_FALSE;
        SpCanvasArch(callback->component).in_draw_func = SP_FALSE;
        
        spDebug(50, "renderFunc", "processed: %s\n", callback->call_name);
    }

    return TRUE;
}
#endif

static gboolean drawFunc(GtkWidget *widget, cairo_t *cr, gpointer data)
{
    spCallback *callback = (spCallback *)data;
    
    if (callback != NULL && callback->call_func != NULL
        && callback->component != NULL && callback->call_name != NULL) {
        spDebug(50, "drawFunc", "callbacked: %s\n", callback->call_name);
#if 1
        spDebug(50, "drawFunc", "width = %d, height = %d\n", spGetAllocationWidthGtk(widget), spGetAllocationHeightGtk(widget));
#endif

        SpCanvasArch(callback->component).in_draw_func = SP_TRUE;
        
        if (SpCanvasArch(callback->component).wait_refresh == SP_TRUE) {
            spDebug(50, "drawFunc", "process waited refresh\n");
            spRefreshCanvasGtk(callback->component, cr);
            SpCanvasArch(callback->component).wait_refresh = SP_FALSE;
        } else {
            SpCanvasArch(callback->component).draw_cr = cr;
            spEmitComponentCallbackFuncGtk(callback, NULL);
            SpCanvasArch(callback->component).draw_cr = NULL;
        }
        
        SpCanvasArch(callback->component).in_draw_func = SP_FALSE;
    }

    return TRUE;
}

static gboolean changeValueFunc(GtkRange *range, GtkScrollType scroll,
                                gdouble value, gpointer data)
{
    spCallback *callback = (spCallback *)data;
    spComponent component;
    GdkModifierType mask;

    if (callback != NULL && callback->call_func != NULL
        && callback->call_name != NULL && spIsSlider(callback->component) == SP_TRUE) {
        spDebug(30, "changeValueFunc", "call_name = %s\n", callback->call_name);
        component = callback->component;

        if (SpSliderPart(component).track_call_on == SP_FALSE) {
            if (spGetMouseModifierMaskGtk(component, &mask) == SP_TRUE) {
                spDebug(30, "changeValueFunc", "type = %d (%d), value = %f, mask = %x\n",
                        (int)scroll, GTK_SCROLL_JUMP, value, (int)mask);
    
                if (scroll == GTK_SCROLL_JUMP && (mask & GDK_BUTTON1_MASK)) {
                    SpPrimitiveArch(component).state = 1;
                    spDebug(30, "changeValueFunc", "set tracking state\n");
                } else {
                    SpPrimitiveArch(component).state = 0;
                }
            }
        } else {
            SpPrimitiveArch(component).state = 0;
        }
        if (!SpPrimitiveArch(component).state) {
            spEmitComponentCallbackFuncGtk(callback, NULL);
            SpPrimitiveArch(component).state = 0;
        }
    }
    
    spDebug(30, "changeValueFunc", "done\n");
    
    return FALSE;
}

static gboolean buttonReleaseFunc(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    spCallback *callback = (spCallback *)user_data;

    if (callback != NULL && callback->call_func != NULL
        && callback->call_name != NULL && spIsSlider(callback->component) == SP_TRUE) {
        spDebug(30, "buttonReleaseFunc", "call_name = %s, state = %d\n",
                callback->call_name, SpPrimitiveArch(callback->component).state);
        if (SpPrimitiveArch(callback->component).state) {
            spEmitComponentCallbackFuncGtk(callback, NULL);
            SpPrimitiveArch(callback->component).state = 0;
        }
    }
    
    spDebug(30, "buttonReleaseFunc", "done\n");
    
    return FALSE;
}
#endif

static const gchar *getCallbackName(spComponent component, spCallbackType call_type,
                                    gint *event_mask, int *count)
{
    int i;
    const gchar *call_name = NULL;

    if (sp_num_callback_table <= 0) {
        sp_num_callback_table = spArraySize(sp_callback_tables);
    }
    
    if (call_type == SP_NO_CALLBACK) return NULL;

    if (*count < sp_num_callback_table) {
        for (i = *count; i < sp_num_callback_table; i++) {
            if (sp_callback_tables[i].component_type == NULL
                || spIsSubClass(component, sp_callback_tables[i].component_type) == SP_TRUE) {
                if (call_type & sp_callback_tables[i].call_type &&
                    sp_callback_tables[i].call_name != NULL) {
                    call_name = sp_callback_tables[i].call_name;
                    *event_mask = sp_callback_tables[i].event_mask;
    
                    *count = i + 1;
                    
                    break;
                }
            }
        }
        if (i >= sp_num_callback_table) {
            *count = sp_num_callback_table;
        }
    }

    return call_name;
}

static void addToolItemCallback(GtkWidget *widget, gpointer data)
{
    spCallback *callback = (spCallback *)data;
    
    spDebug(50, "addToolItemCallback", "add\n");
    
    if (callback != NULL
        && callback->component != NULL
        && SpPrimitiveArch(callback->component).widget != NULL) {
        callback->func_id = spSignalConnectCallbackGtk(SpPrimitiveArch(callback->component).widget,
                                                       callback->call_name, callbackFunc, (gpointer)callback);
    }

    return;
}

static spBool addCallback(spComponent component, spBool propagate, const gchar *call_name,
                          gint event_mask, spCallbackFunc call_func, void *call_data)
{
    long k;
    GtkWidget *widget = NULL;

    if (call_name != NULL && call_func != NULL) {
        for (k = 0; k < SpPrimitivePart(component).num_callback; k++) {
            if (SpPrimitivePart(component).callbacks[k].component == component &&
                SpPrimitivePart(component).callbacks[k].call_name == call_name &&
                SpPrimitivePart(component).callbacks[k].call_func == call_func &&
                SpPrimitivePart(component).callbacks[k].call_data == call_data &&
                SpPrimitivePart(component).callbacks[k].event_mask == event_mask) {
                return SP_FALSE;
            } else if (SpPrimitivePart(component).callbacks[k].call_func == NULL) {
                break;
            }
        }
        if (k >= SpPrimitivePart(component).num_callback) {
            spAllocCallbacks(component);
            SpPrimitivePart(component).num_callback++;
        }
        SpPrimitivePart(component).callbacks[k].component = component;
        SpPrimitivePart(component).callbacks[k].call_name = call_name;
        SpPrimitivePart(component).callbacks[k].call_func = call_func;
        SpPrimitivePart(component).callbacks[k].call_data = call_data;
        SpPrimitivePart(component).callbacks[k].event_mask = event_mask;
        SpPrimitivePart(component).callbacks[k].propagate = propagate;
        
        spDebug(40, "addCallback", "k = %ld, call_name = %s\n",
                k, SpPrimitivePart(component).callbacks[k].call_name);

        if (!streq(SpPrimitivePart(component).callbacks[k].call_name, SppDestroyCallback)
            && !streq(SpPrimitivePart(component).callbacks[k].call_name, SppCloseCallback)) {
            if (spIsFrame(component) == SP_TRUE
                && SpFrameArch(component).toplevel != NULL
                && streq(call_name, "destroy")) {
                widget = SpFrameArch(component).toplevel;
                spDebug(30, "addCallback", "add destroy callback\n");
#if GTK_CHECK_VERSION(3,0,0)
#if GTK_CHECK_VERSION(3,16,0)
            } else if (spIsSubClass(component, "GLCanvas") == SP_TRUE && streq(call_name, "configure_event")) {
                SpPrimitivePart(component).callbacks[k].func_id =
                    g_signal_connect(SpPrimitiveArch(component).widget, "resize",
                                     G_CALLBACK(resizeFunc), (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                return SP_TRUE;
#endif
            } else if (spIsCanvas(component) == SP_TRUE && streq(call_name, "draw")) {
#if GTK_CHECK_VERSION(3,16,0)
                if (spIsSubClass(component, "GLCanvas") == SP_TRUE) {
                    SpPrimitivePart(component).callbacks[k].func_id =
                        g_signal_connect(SpPrimitiveArch(component).widget, "render",
                                         G_CALLBACK(renderFunc), (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                    return SP_TRUE;
                }
#endif
                SpPrimitivePart(component).callbacks[k].func_id =
                    g_signal_connect(SpPrimitiveArch(component).widget, call_name,
                                     G_CALLBACK(drawFunc), (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                return SP_TRUE;
#endif
            } else if (spIsToolItem(component) == SP_TRUE) {
                spSignalConnectCallbackGtk(SpParentPrimitiveArch(component).widget, "realize",
                                           addToolItemCallback, (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                return SP_TRUE;
#if GTK_CHECK_VERSION(2,0,0)
            } else if (spIsSubClass(component, SpTextArea) == SP_TRUE
                       && GTK_IS_TEXT_VIEW(SpPrimitiveArch(component).widget)
                       && streq(call_name, "changed")) {
                SpPrimitivePart(component).callbacks[k].func_id =
                    g_signal_connect(gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget)),
                                     call_name, G_CALLBACK(callbackFunc),
                                     (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                return SP_TRUE;
#endif
            } else if (spIsCanvas(component) == SP_TRUE && streq(call_name, "value_changed")) {
                spDebug(30, "addCallback", "add scroll callback to canvas\n");
                if (event_mask == 1 && SpCanvasArch(component).h_adjustment != NULL) {
                    spDebug(30, "addCallback", "add horizontal scroll callback\n");
                    widget = (GtkWidget *)SpCanvasArch(component).h_adjustment;
                } else if (event_mask == 2 && SpCanvasArch(component).v_adjustment != NULL) {
                    spDebug(30, "addCallback", "add vertical scroll callback\n");
                    widget = (GtkWidget *)SpCanvasArch(component).v_adjustment;
                } else {
                    return SP_FALSE;
                }
                event_mask = 0;
            } else if (spIsSlider(component) == SP_TRUE
                       && streq(call_name, "value_changed")) { /* "change-value" is for SpPrimitiveArch(component).widget*/
                widget = (GtkWidget *)SpPrimitiveArch(component).adjustment;
#if !GTK_CHECK_VERSION(3,0,0)
            } else if (spIsComboBox(component) == SP_TRUE
                       && streq(call_name, "changed")) {
                widget = spGetComboBoxEntryGtk(component);
#endif
            } else if (SpPrimitiveArch(component).widget != NULL) {
                if (streq(call_name, SppZoomCallback)) {
                    call_name = "scroll_event";
                }
                
                widget = SpPrimitiveArch(component).widget;
            }
            
            if (widget != NULL) {
                if (event_mask) {
                    SpPrimitivePart(component).callbacks[k].func_id =
                        spSignalConnectEventGtk(widget, call_name, eventFunc,
                                                (gpointer)&(SpPrimitivePart(component).callbacks[k]));
#if GTK_CHECK_VERSION(3,0,0)
                } else if (spIsSlider(component) == SP_TRUE && streq(call_name, "change-value")) {
                    spDebug(30, "addCallback", "add scroll bar change-value (%s) callback\n", call_name);
                    SpPrimitivePart(component).callbacks[k].func_id =
                        g_signal_connect(widget, call_name, G_CALLBACK(changeValueFunc),
                                         (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                } else if (spIsSlider(component) == SP_TRUE && streq(call_name, "button-release-event")) {
                    spDebug(30, "addCallback", "add scroll bar button-release-event (%s) callback\n", call_name);
                    SpPrimitivePart(component).callbacks[k].func_id =
                        g_signal_connect(widget, call_name, G_CALLBACK(buttonReleaseFunc),
                                         (gpointer)&(SpPrimitivePart(component).callbacks[k]));
#endif
                } else {
                    SpPrimitivePart(component).callbacks[k].func_id =
                        spSignalConnectCallbackGtk(widget, call_name, callbackFunc,
                                                   (gpointer)&(SpPrimitivePart(component).callbacks[k]));
                }
                spDebug(30, "addCallback", "call_name = %s\n", call_name);
            }
        }
        
        return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool removeCallback(spComponent component, const gchar *call_name, gint event_mask,
                             spCallbackFunc call_func, void *call_data)
{
    long k;
    GtkWidget *widget = NULL;
    
    if (call_name != NULL && call_func != NULL) {
        spDebug(30, "removeCallback", "call_name = %s\n", call_name);
                
        for (k = 0; k < SpPrimitivePart(component).num_callback; k++) {
            spDebug(30, "removeCallback", "k = %ld\n",
                    k, SpPrimitivePart(component).callbacks[k].call_name);
            if (SpPrimitivePart(component).callbacks[k].component == component &&
                SpPrimitivePart(component).callbacks[k].call_name == call_name &&
                SpPrimitivePart(component).callbacks[k].call_func == call_func &&
                SpPrimitivePart(component).callbacks[k].call_data == call_data &&
                SpPrimitivePart(component).callbacks[k].event_mask == event_mask) {
                if (!streq(SpPrimitivePart(component).callbacks[k].call_name, SppDestroyCallback)
                    && !streq(SpPrimitivePart(component).callbacks[k].call_name, SppCloseCallback)) {
                    if (spIsFrame(component) == SP_TRUE
                        && SpFrameArch(component).toplevel != NULL
                        && streq(call_name, "destroy") == SP_TRUE) {
                        widget = SpFrameArch(component).toplevel;
                        /*SpFramePart(component).close_style = SP_DESTROY_CLOSE;*/
#if GTK_CHECK_VERSION(2,0,0)
                    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE
                               && GTK_IS_TEXT_VIEW(SpPrimitiveArch(component).widget)
                               && streq(call_name, "changed")) {
                        widget = NULL;
                        g_signal_handler_disconnect(gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget)),
                                                    SpPrimitivePart(component).callbacks[k].func_id);
#endif
                    } else if (spIsCanvas(component) == SP_TRUE && streq(call_name, "value_changed")) {
                        if (event_mask == 1 && SpCanvasArch(component).h_adjustment != NULL) {
                            spDebug(30, "removeCallback", "remove horizontal scroll callback\n");
                            widget = (GtkWidget *)SpCanvasArch(component).h_adjustment;
                        } else if (event_mask == 2 && SpCanvasArch(component).v_adjustment != NULL) {
                            spDebug(30, "removeCallback", "remove vertical scroll callback\n");
                            widget = (GtkWidget *)SpCanvasArch(component).v_adjustment;
                        } else {
                            return SP_FALSE;
                        }
                    } else if (spIsSlider(component) == SP_TRUE
                               && streq(call_name, "value_changed")) {
                        widget = (GtkWidget *)SpPrimitiveArch(component).adjustment;
                    } else if (SpPrimitiveArch(component).widget != NULL) {
                        widget = SpPrimitiveArch(component).widget;
                    }
                    
                    if (widget != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
                        g_signal_handler_disconnect(widget,
                                                    SpPrimitivePart(component).callbacks[k].func_id);
#elif GTK_CHECK_VERSION(2,0,0)
                        g_signal_handler_disconnect(GTK_OBJECT(widget),
                                                    SpPrimitivePart(component).callbacks[k].func_id);
#else
                        gtk_signal_disconnect(GTK_OBJECT(widget),
                                              SpPrimitivePart(component).callbacks[k].func_id);
#endif
                    }
                }
                    
                SpPrimitivePart(component).callbacks[k].component = NULL;
                SpPrimitivePart(component).callbacks[k].call_name = NULL;
                SpPrimitivePart(component).callbacks[k].call_func = NULL;
                SpPrimitivePart(component).callbacks[k].call_data = NULL;
                SpPrimitivePart(component).callbacks[k].func_id = -1;
                SpPrimitivePart(component).callbacks[k].event_mask = 0;
                SpPrimitivePart(component).callbacks[k].propagate = SP_FALSE;
                    
                return SP_TRUE;
            }
        }
    }

    return SP_FALSE;
}

spBool spPrimitiveAddCallbackArch(spComponent component, spBool propagate, spCallbackType call_type,
                                  spCallbackFunc call_func, void *call_data)
{
    int count = 0;
    gint event_mask = 0;
    spBool flag = SP_FALSE;
    const gchar *call_name = NULL;
    
    while ((call_name = getCallbackName(component, call_type, &event_mask, &count)) != NULL) {
        spDebug(40, "spPrimitiveAddCallbackArch", "call_name = %s\n", call_name);
        if (addCallback(component, propagate, call_name,
                        event_mask, call_func, call_data) == SP_TRUE) {
            flag = SP_TRUE;
        }
    }

    spDebug(50, "spPrimitiveAddCallbackArch", "done: flag = %d\n", flag);
    
    return flag;
}

spBool spPrimitiveRemoveCallbackArch(spComponent component, spCallbackType call_type,
                                     spCallbackFunc call_func, void *call_data)
{
    int count = 0;
    gint event_mask = 0;
    spBool flag = SP_FALSE;
    const gchar *call_name = NULL;
    
    while ((call_name = getCallbackName(component, call_type, &event_mask, &count)) != NULL) {
        if (removeCallback(component, call_name, event_mask,
                           call_func, call_data) == SP_TRUE) {
            flag = SP_TRUE;
        }
    }

    spDebug(50, "spRemoveCallback", "done: flag = %d\n", flag);
    
    return flag;
}

void spInitCallbackArch(spCallback *callback)
{
    callback->component = NULL;
    callback->call_func = NULL;
    callback->call_data = NULL;
    callback->propagate = SP_FALSE;
    
    callback->call_name = NULL;
    /*callback->event = NULL;*/
    callback->event_mask = 0;
    callback->func_id = -1;
    
    return;
}

spCallbackReason spPrimitiveGetCallbackReasonArch(spComponent component)
{
    spCallbackReason call_reason = SP_CR_NONE;
    GdkEvent *event;

    if (SpPrimitiveArch(component).event != NULL) {
        event = SpPrimitiveArch(component).event;
        spDebug(50, "spPrimitiveGetCallbackReasonArch", "event->type = %d\n", event->type);
        switch (event->type) {
          case GDK_KEY_PRESS:
            call_reason = SP_CR_KEY_PRESS;
            break;
          case GDK_KEY_RELEASE:
            call_reason = SP_CR_KEY_RELEASE;
            break;
          case GDK_2BUTTON_PRESS:
            if (spIsSubClass(component, SpList) == SP_TRUE) {
                call_reason = SP_CR_ACTIVATE;
            } else {
                call_reason = SP_CR_UNKNOWN;
            }
            break;
          case GDK_BUTTON_PRESS:
            spDebug(50, "spPrimitiveGetCallbackReasonArch",
                    "button press: event->button.button = %d\n", event->button.button);
            if (event->button.button == 1) {
                call_reason = SP_CR_LBUTTON_PRESS;
            } else if (event->button.button == 3) {
                call_reason = SP_CR_RBUTTON_PRESS;
            } else if (event->button.button == 2) {
                call_reason = SP_CR_MBUTTON_PRESS;
            } else if (event->button.button == 4) {
                call_reason = SP_CR_KEY_PRESS;
            } else if (event->button.button == 5) {
                call_reason = SP_CR_KEY_PRESS;
            }
            break;
          case GDK_BUTTON_RELEASE:
            if (event->button.button == 1) {
                call_reason = SP_CR_LBUTTON_RELEASE;
            } else if (event->button.button == 3) {
                call_reason = SP_CR_RBUTTON_RELEASE;
            } else if (event->button.button == 2) {
                call_reason = SP_CR_MBUTTON_RELEASE;
            }
            break;
          case GDK_MOTION_NOTIFY:
            if (event->motion.state & GDK_BUTTON1_MASK) {
                call_reason = SP_CR_LBUTTON_MOTION;
            } else if (event->motion.state & GDK_BUTTON3_MASK) {
                call_reason = SP_CR_RBUTTON_MOTION;
            } else if (event->motion.state & GDK_BUTTON2_MASK) {
                call_reason = SP_CR_MBUTTON_MOTION;
            } else {
                call_reason = SP_CR_POINTER_MOTION;
            }
            break;
            
          case GDK_EXPOSE:
            call_reason = SP_CR_EXPOSE;
            break;
          case GDK_CONFIGURE:
            call_reason = SP_CR_RESIZE;
            break;
            
          case GDK_ENTER_NOTIFY:
            call_reason = SP_CR_ENTER_WINDOW;
            break;
          case GDK_LEAVE_NOTIFY:
            call_reason = SP_CR_LEAVE_WINDOW;
            break;
            
#if GTK_CHECK_VERSION(2,0,0)
          case GDK_SCROLL:
            if (event->scroll.state & GDK_CONTROL_MASK) {
                call_reason = SP_CR_ZOOM;
            } else {
                call_reason = SP_CR_WHEEL;
            }
            break;
#endif
            
          default:
            call_reason = SP_CR_UNKNOWN;
            break;
        }
    } else if (!strnone(SpPrimitiveArch(component).call_name)) {
        spDebug(100, "spPrimitiveGetCallbackReasonArch", "call_name = %s\n", SpPrimitiveArch(component).call_name);
        if (streq(SpPrimitiveArch(component).call_name, "clicked")
            || streq(SpPrimitiveArch(component).call_name, "activate")) {
            if (spIsToggleButton(component) == SP_TRUE) {
                call_reason = SP_CR_VALUE_CHANGED;
            } else {
                call_reason = SP_CR_ACTIVATE;
            }
        } else if (streq(SpPrimitiveArch(component).call_name, SppDestroyCallback)) {
            call_reason = SP_CR_DESTROY;
        } else if (streq(SpPrimitiveArch(component).call_name, SppCloseCallback)) {
            call_reason = SP_CR_CLOSE;
        } else if (streq(SpPrimitiveArch(component).call_name, "destroy")) {
            call_reason = SP_CR_CLOSE;
        } else if (streq(SpPrimitiveArch(component).call_name, "value_changed")
                   && spIsCanvas(component) == SP_TRUE) {
            call_reason = SP_CR_SCROLL;
        } else if (streq(SpPrimitiveArch(component).call_name, "value_changed")
                   || streq(SpPrimitiveArch(component).call_name, "selection_changed")
                   || streq(SpPrimitiveArch(component).call_name, SP_SLIDER_VALUE_CHANGED_CALLBACK_NAME)) {
            call_reason = SP_CR_VALUE_CHANGED;
#if GTK_CHECK_VERSION(3,0,0)
        } else if (streq(SpPrimitiveArch(component).call_name, "configure_event")) {
            call_reason = SP_CR_RESIZE;
        } else if (streq(SpPrimitiveArch(component).call_name, "draw")) {
            call_reason = SP_CR_EXPOSE;
        } else if (spIsSlider(component) == SP_TRUE
                   && streq(SpPrimitiveArch(component).call_name, "button-release-event")) {
            spDebug(50, "spPrimitiveGetCallbackReasonArch", "slider button released\n");
            call_reason = SP_CR_VALUE_CHANGED;
#endif
        } else if (streq(SpPrimitiveArch(component).call_name, "toggled")) {
            if (spIsRadioButton(component) == SP_TRUE) {
                spBool set;
                if (spGetToggleState(component, &set) == SP_TRUE
                    && set == SP_FALSE) {
                    spDebug(50, "spPrimitiveGetCallbackReasonArch", "return CR_NONE\n");
                    return SP_CR_NONE;
                }
            }
            call_reason = SP_CR_VALUE_CHANGED;
        } else if (streq(SpPrimitiveArch(component).call_name, "changed")) {
            call_reason = SP_CR_VALUE_CHANGED;
        } else if (streq(SpPrimitiveArch(component).call_name, "pressed")) {
            call_reason = SP_CR_BUTTON_PRESS;
        } else if (streq(SpPrimitiveArch(component).call_name, "released")) {
            call_reason = SP_CR_BUTTON_RELEASE;
        } else if (streq(SpPrimitiveArch(component).call_name, "map")) {
            call_reason = SP_CR_MAP;
        } else if (streq(SpPrimitiveArch(component).call_name, "unmap")) {
            call_reason = SP_CR_UNMAP;
        } else {
            call_reason = SP_CR_UNKNOWN;
        }
    }
    
    return call_reason;
}

spBool spPrimitiveGetCallbackMousePositionArch(spComponent component, int *x, int *y)
{
    int lx = 0, ly = 0;
    GdkEvent *event;
    
    if (SpPrimitiveArch(component).event != NULL) {
        event = SpPrimitiveArch(component).event;
        switch (event->type) {
          case GDK_BUTTON_PRESS:
          case GDK_2BUTTON_PRESS:
          case GDK_3BUTTON_PRESS:
          case GDK_BUTTON_RELEASE:
            lx = event->button.x;
            ly = event->button.y;
            break;
          case GDK_MOTION_NOTIFY:
            spDebug(60, "spPrimitiveGetCallbackMousePositionArch",
                    "button motion: %f, %f\n", event->motion.x, event->motion.y);
            
            lx = event->motion.x;
            ly = event->motion.y;
            break;
          default:
            return SP_FALSE;
        }
        if (x != NULL) *x = lx;
        if (y != NULL) *y = ly;
        
        spDebug(60, "spPrimitiveGetCallbackMousePositionArch",
                "x = %d, y = %d\n", lx, ly);
        
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spPrimitiveGetCallbackWheelValueArch(spComponent component, double *delta_x, double *delta_y)
{
#if GTK_CHECK_VERSION(2,0,0)
    GdkEvent *event;
    int horizontal_flag;
    
    if (SpPrimitiveArch(component).event != NULL) {
        event = SpPrimitiveArch(component).event;
        switch (event->type) {
          case GDK_SCROLL:
            horizontal_flag = (event->scroll.state & GDK_SHIFT_MASK);
            if (SpCanvasPart(component).wheel_target_horizontal == SP_TRUE) {
                horizontal_flag = !horizontal_flag;
            }
            
            if (delta_x != NULL) {
                *delta_x = 0.0;
                if (horizontal_flag) {
                    if (event->scroll.direction == GDK_SCROLL_UP) {
                        *delta_x = -1.0;
                    } else if (event->scroll.direction == GDK_SCROLL_DOWN) {
                        *delta_x = 1.0;
                    }
                    if (SpCanvasPart(component).scroll_left_by_wheel_down == SP_TRUE) {
                        *delta_x *= -1.0;
                    }
                } else {
                    if (event->scroll.direction == GDK_SCROLL_LEFT) {
                        *delta_x = -1.0;
                    } else if (event->scroll.direction == GDK_SCROLL_RIGHT) {
                        *delta_x = 1.0;
                    }
                }
            }
            if (delta_y != NULL) {
                *delta_y = 0.0;
                if (horizontal_flag) {
                    if (event->scroll.direction == GDK_SCROLL_LEFT) {
                        *delta_y = -1.0;
                    } else if (event->scroll.direction == GDK_SCROLL_RIGHT) {
                        *delta_y = 1.0;
                    }
                    if (SpCanvasPart(component).scroll_left_by_wheel_down == SP_TRUE) {
                        *delta_y *= -1.0;
                    }
                } else {
                    if (event->scroll.direction == GDK_SCROLL_UP) {
                        *delta_y = 1.0;
                    } else if (event->scroll.direction == GDK_SCROLL_DOWN) {
                        *delta_y = -1.0;
                    }
                }
            }
            return SP_TRUE;
          default:
            break;
        }
    }
#endif
    
    return SP_FALSE;
}

static spKeySym convertKeySym(int key_sym)
{
    if (key_sym >= 0x020 && key_sym <= 0x0ff) {
        return SPK_Character;
    } else {
        switch (key_sym) {
          case SPK_BackSpace:                
          case SPK_Tab:                        
          case SPK_Clear:                
          case SPK_Return:                
          case SPK_Pause:                
          case SPK_ScrollLock:                
          case SPK_Escape:                
          case SPK_Delete:                
          case SPK_Shift:                
          case SPK_Control:                
          case SPK_CapsLock:                
          case SPK_Home:                
          case SPK_Left:                
          case SPK_Up:                        
          case SPK_Right:                
          case SPK_Down:                
          case SPK_Prior:                
          case SPK_Next:                
          case SPK_End:                        
          case SPK_Select:                
          case SPK_Print:                
          case SPK_Execute:                
          case SPK_Insert:                
          case SPK_Cancel:                
          case SPK_Help:                
          case SPK_NumLock:                

          case SPK_F1:                        
          case SPK_F2:                        
          case SPK_F3:                        
          case SPK_F4:                
          case SPK_F5:                        
          case SPK_F6:                        
          case SPK_F7:                        
          case SPK_F8:                        
          case SPK_F9:                        
          case SPK_F10:                        
          case SPK_F11:                        
          case SPK_F12:                
          case SPK_F13:                        
          case SPK_F14:                        
          case SPK_F15:                        
          case SPK_F16:                        
          case SPK_F17:                        
          case SPK_F18:                        
          case SPK_F19:                        
          case SPK_F20:                        
          case SPK_F21:                        
          case SPK_F22:                        
          case SPK_F23:                        
          case SPK_F24:                        
          case SPK_Alt:                        
            return key_sym;
#if GTK_CHECK_VERSION(3,0,0)
          case GDK_KEY_Shift_R:                
          case GDK_KEY_Control_R:                
          case GDK_KEY_Alt_R:
#else
          case GDK_Shift_R:                
          case GDK_Control_R:                
          case GDK_Alt_R:
#endif
            return (key_sym - 0x01);
#if GTK_CHECK_VERSION(3,0,0)
          case GDK_KEY_Menu:                
          case GDK_KEY_Meta_L:                
          case GDK_KEY_Meta_R:
#else
          case GDK_Menu:                
          case GDK_Meta_L:                
          case GDK_Meta_R:
#endif
            return SPK_Alt;
          default:
            break;
        }
    }

    return SPK_Unknown;
}

spBool spPrimitiveGetCallbackKeySymArch(spComponent component, spKeySym *key_sym)
{
    GdkEvent *event;
    spKeySym ks;
    
    if (SpPrimitiveArch(component).event != NULL) {
        event = SpPrimitiveArch(component).event;
        switch (event->type) {
          case GDK_KEY_PRESS:
          case GDK_KEY_RELEASE:
            ks = convertKeySym(event->key.keyval);
            spDebug(60, "spPrimitiveGetCallbackKeySymArch", "ks = %ld\n", ks);
            if (ks != SPK_Unknown /*&& ks != SPK_Character*/) {
                *key_sym = ks;
                return SP_TRUE;
            }
            break;
            
          case GDK_BUTTON_PRESS:
            spDebug(60, "spPrimitiveGetCallbackKeySymArch", "button = %d\n", event->button.button);
            if (event->button.button == 4) {
                *key_sym = SPK_Prior;
                return SP_TRUE;
            } else if (event->button.button == 5) {
                *key_sym = SPK_Next;
                return SP_TRUE;
            }
            break;
            
          default:
            break;
        }
    }
    
    return SP_FALSE;
}

int spPrimitiveGetCallbackKeyStringArch(spComponent component,
                                        char *buf, int buf_size, spBool *overflow)
{
    int len;
    GdkEvent *event;
    spKeySym ks;
    
    if (SpPrimitiveArch(component).event != NULL) {
        event = SpPrimitiveArch(component).event;
        switch (event->type) {
          case GDK_KEY_PRESS:
          case GDK_KEY_RELEASE:
            ks = convertKeySym(event->key.keyval);
            spDebug(60, "spPrimitiveGetCallbackKeyStringArch", "ks = %ld\n", ks);
            if (ks == SPK_Character && event->key.length >= 0) {
                if (event->key.length >= buf_size) {
                    if (overflow != NULL) {
                        *overflow = SP_TRUE;
                    }
                    len = buf_size - 1;
                    strncpy(buf, event->key.string, len);
                } else {
                    if (overflow != NULL) {
                        *overflow = SP_FALSE;
                    }
                    len = event->key.length;
                    strcpy(buf, event->key.string);
                }
                return len;
            }
          default:
            break;
        }
    }

    return -1;
}

void spPrimitivePartInitArch(spComponent component)
{
    SpPrimitiveArch(component).widget = NULL;
    SpPrimitiveArch(component).top_widget = NULL;
    SpPrimitiveArch(component).sub_widget = NULL;
    SpPrimitiveArch(component).adjustment = NULL;
    SpPrimitiveArch(component).pixmap = NULL;
#if GTK_CHECK_VERSION(3,0,0)
    SpPrimitiveArch(component).pixmap_cr = NULL;
#endif
    SpPrimitiveArch(component).state = 0;
    SpPrimitiveArch(component).call_name = NULL;
    SpPrimitiveArch(component).event = NULL;
    SpPrimitiveArch(component).event_mask = 0;
#if GTK_CHECK_VERSION(2,0,0)
    SpPrimitiveArch(component).layout = NULL;
#endif

    SpPrimitiveArch(component).set_width = -1;
    SpPrimitiveArch(component).set_height = -1;
    
    return;
}

void spPrimitiveDestroyArch(spComponent component)
{
    GtkWidget *widget = NULL;

    if (SpPrimitiveArch(component).pixmap != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
        cairo_surface_t *pixmap;

        if (SpPrimitiveArch(component).pixmap_cr != NULL) {
            cairo_destroy(SpPrimitiveArch(component).pixmap_cr);
        }
        SpPrimitiveArch(component).pixmap_cr = NULL;
        
        pixmap = SpPrimitiveArch(component).pixmap;
        SpPrimitiveArch(component).pixmap = NULL;
        cairo_surface_destroy(pixmap);
#else
        GdkPixmap *pixmap;
        pixmap = SpPrimitiveArch(component).pixmap;
        SpPrimitiveArch(component).pixmap = NULL;
        gdk_pixmap_unref(pixmap);
#endif
    }
#if GTK_CHECK_VERSION(2,0,0)
    if (SpPrimitiveArch(component).layout != NULL) {
        g_object_unref(SpPrimitiveArch(component).layout);
        SpPrimitiveArch(component).layout = NULL;
    }
#endif

    if (spIsFrame(component) == SP_TRUE
        && SpFrameArch(component).toplevel != NULL) {
        widget = SpFrameArch(component).toplevel;
        SpFrameArch(component).toplevel = NULL;
        gtk_widget_destroy(widget);
    } else {
        if (SpPrimitiveArch(component).top_widget != NULL) {
            widget = SpPrimitiveArch(component).top_widget;
            SpPrimitiveArch(component).top_widget = NULL;
            gtk_widget_destroy(widget);
        } else {
            if (SpPrimitiveArch(component).sub_widget != NULL) {
                widget = SpPrimitiveArch(component).sub_widget;
                SpPrimitiveArch(component).sub_widget = NULL;
                gtk_widget_destroy(widget);
            }
            if (SpPrimitiveArch(component).widget != NULL) {
                widget = SpPrimitiveArch(component).widget;
                SpPrimitiveArch(component).widget = NULL;
                gtk_widget_destroy(widget);
            }
        }
    }
    
    return;
}

void spPrimitiveMapArch(spComponent component)
{
    if (SpPrimitiveArch(component).top_widget != NULL) {
        gtk_widget_show(SpPrimitiveArch(component).top_widget);
    } else if (SpPrimitiveArch(component).widget != NULL) {
        if (SpPrimitiveArch(component).sub_widget != NULL) {
            gtk_widget_show(SpPrimitiveArch(component).sub_widget);
        }
        gtk_widget_show(SpPrimitiveArch(component).widget);
    }
    
    return;
}

void spPrimitiveUnmapArch(spComponent component)
{
    if (SpPrimitiveArch(component).top_widget != NULL) {
        gtk_widget_hide(SpPrimitiveArch(component).top_widget);
    } else if (SpPrimitiveArch(component).widget != NULL) {
        if (SpPrimitiveArch(component).sub_widget != NULL) {
            gtk_widget_hide(SpPrimitiveArch(component).sub_widget);
        }
        gtk_widget_hide(SpPrimitiveArch(component).widget);
    }
    
    return;
}

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)
GtkWidget *spGetPrevWidgetGtk(spComponent component)
{
    GtkWidget *prev_widget = NULL;
    spComponent prev_component;

    prev_component = component;
    do {
        prev_component = SpGetPrevComponent(prev_component);
    } while (prev_component != NULL && spIsPrimitive(prev_component) == SP_FALSE);
    
    spDebug(80, "spGetPrevWidgetGtk", "%s: prev_component = %ld\n",
            SpGetName(component), (long)prev_component);
    
    if (spIsPrimitive(prev_component) == SP_TRUE) {
        spDebug(80, "spGetPrevWidgetGtk", "name of prev_component = %s\n",
                SpGetName(prev_component));
        if (SpPrimitiveArch(prev_component).top_widget != NULL) {
            prev_widget = SpPrimitiveArch(prev_component).top_widget;
        } else {
            prev_widget = SpPrimitiveArch(prev_component).widget;
        }
        if (prev_widget != NULL
            && SpParentPrimitiveArch(component).widget != gtk_widget_get_parent(prev_widget)) {
            prev_widget = NULL;
        }
    }

    return prev_widget;
}

GtkWidget *spGetNextWidgetGtk(spComponent component)
{
    GtkWidget *next_widget = NULL;
    spComponent next_component;

    next_component = component;
    do {
        next_component = SpGetNextComponent(next_component);
    } while (next_component != NULL && spIsPrimitive(next_component) == SP_FALSE);
    
    spDebug(80, "spGetNextWidgetGtk", "%s: next_component = %ld\n",
            SpGetName(component), (long)next_component);
    
    if (spIsPrimitive(next_component) == SP_TRUE) {
        spDebug(80, "spGetNextWidgetGtk", "name of next_component = %s\n",
                SpGetName(next_component));
        if (SpPrimitiveArch(next_component).top_widget != NULL) {
            next_widget = SpPrimitiveArch(next_component).top_widget;
        } else {
            next_widget = SpPrimitiveArch(next_component).widget;
        }
        if (next_widget != NULL
            && SpParentPrimitiveArch(component).widget != gtk_widget_get_parent(next_widget)) {
            next_widget = NULL;
        }
    }

    return next_widget;
}

spBool spFindPrevNegativeWidthComponent(spComponent component)
{
    spComponent prev_component;

    prev_component = SpGetPrevComponent(component);
    
    while (prev_component != NULL) {
        if (SpComponentPart(prev_component).visible_flag && SpComponentPart(prev_component).width < 0) {
            return SP_TRUE;
        }
        prev_component = SpGetPrevComponent(prev_component);
    }
    
    return SP_FALSE;
}

spBool spFindPrevNegativeHeightComponent(spComponent component)
{
    spComponent prev_component;

    prev_component = SpGetPrevComponent(component);
    
    while (prev_component != NULL) {
        if (SpComponentPart(prev_component).visible_flag && SpComponentPart(prev_component).height < 0) {
            return SP_TRUE;
        }
        prev_component = SpGetPrevComponent(prev_component);
    }
    
    return SP_FALSE;
}

void spUpdateWidgetExpandGtk(spComponent component, GtkWidget *widget)
{
    if (spIsToolBar(component) == SP_TRUE) {
        gtk_widget_set_hexpand(widget, TRUE);
        gtk_widget_set_vexpand(widget, FALSE);
    } else if (spIsToolItem(component) == SP_TRUE) {
        gtk_widget_set_hexpand(widget, FALSE);
        gtk_widget_set_vexpand(widget, FALSE);
    } else {
#if 0
        if (SpGetPrevComponent(component) != NULL) {
            spDebug(80, "spUpdateWidgetExpandGtk", "prev component = %s, %d\n",
                    spGetName(SpGetPrevComponent(component)), SpComponentPart(SpGetPrevComponent(component)).width);
        }
#endif
        if (SpComponentPart(component).width <= 0
            && spFindPrevNegativeWidthComponent(component) == SP_FALSE) {
            spDebug(80, "spUpdateWidgetExpandGtk", "%s: **** hexpand = TRUE ****\n", spGetName(component));
            gtk_widget_set_hexpand(widget, TRUE);
        } else {
            spDebug(80, "spUpdateWidgetExpandGtk", "%s: ---- hexpand = FALSE ----\n", spGetName(component));
            gtk_widget_set_hexpand(widget, FALSE);
        }
        if (/*spIsCanvas(component) == SP_TRUE
              ||*/ (SpComponentPart(component).height <= 0
                && (spIsStatusBar(component) == SP_FALSE && spIsMenuBar(component) == SP_FALSE)
                && spFindPrevNegativeHeightComponent(component) == SP_FALSE)) {
            spDebug(80, "spUpdateWidgetExpandGtk", "%s: **** vexpand = TRUE ****\n", spGetName(component));
            gtk_widget_set_vexpand(widget, TRUE);
        } else {
            spDebug(80, "spUpdateWidgetExpandGtk", "%s: ---- vexpand = FALSE ----\n", spGetName(component));
            gtk_widget_set_vexpand(widget, FALSE);
        }

        if (spIsStatusBar(component) == SP_TRUE) {
            gtk_widget_set_halign(widget, GTK_ALIGN_FILL);        
            gtk_widget_set_valign(widget, GTK_ALIGN_END);        
        }
    }

    return;
}

void spUpdateWidgetMarginGtk(spComponent component, GtkWidget *widget)
{
    GtkWidget *prev;
    GtkWidget *next;
    gint margin_left, margin_right;
    gint margin_top, margin_bottom;

    prev = spGetPrevWidgetGtk(component);
    next = spGetNextWidgetGtk(component);
    spDebug(80, "spUpdateWidgetMarginGtk", "%s: prev widget = %ld, next widget  = %ld, parent orientation = %d\n",
            SpGetClassName(component), (long)prev, (long)next, SpParentComponentPart(component).orientation);

    if (SpParentComponentPart(component).orientation == SP_VERTICAL || prev == NULL) {
        margin_left = SpComponentPart(component).margin_width;
        margin_left = (margin_left > 0 ? margin_left : SP_DEFAULT_MARGIN);
    } else {
        margin_left = SpParentComponentPart(component).spacing / 2;
        spDebug(80, "spUpdateWidgetMarginGtk", "%s: half spacing for left: %d\n", SpGetClassName(component), margin_left);
        
        if (prev != NULL && SpParentComponentPart(component).orientation == SP_HORIZONTAL) {
            gtk_widget_set_margin_right(prev, margin_left);
        }
    }
    gtk_widget_set_margin_left(widget, margin_left);
            
    if (SpParentComponentPart(component).orientation == SP_VERTICAL || next == NULL) {
        margin_right = SpComponentPart(component).margin_width;
        margin_right = (margin_right > 0 ? margin_right : SP_DEFAULT_MARGIN);
    } else {
        margin_right = SpParentComponentPart(component).spacing / 2;
        spDebug(80, "spUpdateWidgetMarginGtk", "%s: half spacing for right: %d\n", SpGetClassName(component), margin_right);
        
        if (next != NULL && SpParentComponentPart(component).orientation == SP_HORIZONTAL) {
            gtk_widget_set_margin_left(next, margin_right);
        }
    }
    gtk_widget_set_margin_right(widget, margin_right);
        
    if (SpParentComponentPart(component).orientation == SP_HORIZONTAL || prev == NULL) {
        margin_top = SpComponentPart(component).margin_height;
        margin_top = (margin_top > 0 ? margin_top : SP_DEFAULT_MARGIN);
    } else {
        margin_top = SpParentComponentPart(component).spacing / 2;
        spDebug(80, "spUpdateWidgetMarginGtk", "%s: half spacing for top: %d\n", SpGetClassName(component), margin_top);
        
        if (prev != NULL && SpParentComponentPart(component).orientation == SP_VERTICAL) {
            gtk_widget_set_margin_bottom(prev, margin_top);
        }
    }
    gtk_widget_set_margin_top(widget, margin_top);
        
    if (SpParentComponentPart(component).orientation == SP_HORIZONTAL || next == NULL) {
        margin_bottom = SpComponentPart(component).margin_height;
        margin_bottom = (margin_bottom > 0 ? margin_bottom : SP_DEFAULT_MARGIN);
    } else {
        margin_bottom = SpParentComponentPart(component).spacing / 2;
        spDebug(80, "spUpdateWidgetMarginGtk", "%s: half spacing for bottom: %d\n", SpGetClassName(component), margin_bottom);
        
        if (next != NULL && SpParentComponentPart(component).orientation == SP_VERTICAL) {
            gtk_widget_set_margin_top(next, margin_bottom);
        }
    }
    gtk_widget_set_margin_bottom(widget, margin_bottom);
        
    spDebug(80, "spUpdateWidgetMarginGtk", "%s: margin = (%d, %d, %d, %d)\n",
            SpGetClassName(component), margin_left, margin_right, margin_top, margin_bottom);
}
#endif

#if !GTK_CHECK_VERSION(2,0,0)
#define gtk_widget_set_size_request gtk_widget_set_usize
#endif

void spSetWidgetSizeGtk(GtkWidget *widget, int width, int height)
{
    gtk_widget_set_size_request(widget, width, height);
    
    return;
}

void spPrimitiveScaleDefaultSizeArch(spComponent component)
{
    return;
}

spBool spPrimitiveSetSizeArch(spComponent component, int width, int height)
{
    int h;
    GtkWidget *widget = NULL;

    spDebug(100, "spPrimitiveSetSizeArch", "%s: width = %d, height = %d\n",
            SpGetClassName(component), width, height);

    if (spIsFrame(component) == SP_TRUE) {
        if (spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
            SpComponentPart(SpFramePart(component).status_bar).current_height = -1;
            if (spGetSize(SpFramePart(component).status_bar, NULL, &h) == SP_TRUE) {
                height += h;
            }
        }
        if (spIsVisible(SpFramePart(component).menu_bar) == SP_TRUE) {
            SpComponentPart(SpFramePart(component).menu_bar).current_height = -1;
            if (spGetSize(SpFramePart(component).menu_bar, NULL, &h) == SP_TRUE) {
                height += h;
            }
        }
        
        spDebug(30, "spPrimitiveSetSizeArch",
                "Frame: x = %d, y = %d, width = %d, height = %d\n",
                SpComponentPart(component).x, SpComponentPart(component).y,
                width, height);
        
#if GTK_CHECK_VERSION(3,0,0)
        gtk_window_resize(GTK_WINDOW(SpFrameArch(component).toplevel), width, height);
#else
        if (SpFramePart(component).window_type != SP_TRANSIENT_WINDOW
            && spIsWidgetMappedGtk(SpFrameArch(component).toplevel)) {
            gdk_window_resize(spGetGdkWindowGtk(SpFrameArch(component).toplevel), width, height);
        } else {
            spSetWidgetSizeGtk(SpFrameArch(component).toplevel, width, height);
        }
#endif        /* !GTK_CHECK_VERSION(3,0,0) */
        
        if (SpComponentPart(component).x != 0 && SpComponentPart(component).y != 0) {
#if GTK_CHECK_VERSION(2,0,0)
            gtk_window_move(GTK_WINDOW(SpFrameArch(component).toplevel),
                            SpComponentPart(component).x,
                            SpComponentPart(component).y);
#else
            gdk_window_move(spGetGdkWindowGtk(SpFrameArch(component).toplevel),
                            SpComponentPart(component).x,
                            SpComponentPart(component).y);
#endif
        }
        
        SpPrimitiveArch(component).set_width = width;
        SpPrimitiveArch(component).set_height = height;
        
        return SP_TRUE;
    } else if (SpPrimitiveArch(component).widget != NULL) {
        if (spIsSubClass(component, SpLabel) == SP_FALSE && SpPrimitiveArch(component).top_widget != NULL) {
            widget = SpPrimitiveArch(component).top_widget;
        } else {
            widget = SpPrimitiveArch(component).widget;
        }

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)
        spUpdateWidgetExpandGtk(component, widget);

        spDebug(30, "spPrimitiveSetSizeArch",
                "%s: width = %d, SpComponentPart(component).width = %d, height = %d, SpComponentPart(component).height = %d\n",
                SpGetClassName(component), width, SpComponentPart(component).width, height, SpComponentPart(component).height);

        if (spIsCanvas(component) == SP_TRUE || spIsSubClass(component, SpContainer) == SP_TRUE) {
            if (spIsCanvas(component) == SP_TRUE
                && (SpCanvasArch(component).h_adjustment != NULL || SpCanvasArch(component).v_adjustment != NULL)) {
                spUpdateCanvasScrollBarArch(component);
            } else {
                gtk_widget_set_size_request(SpPrimitiveArch(component).widget,
                                            SpComponentPart(component).width > 0 ? width : -1,
                                            SpComponentPart(component).height > 0 ? height : -1);
            }
        } else {
            gtk_widget_set_size_request(widget,
                                        SpComponentPart(component).width > 0 ? width : -1,
                                        SpComponentPart(component).height > 0 ? height : -1);
        }
#else /* GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID) */
        if (spIsSubClass(component, SpBox) == SP_TRUE
            && spIsSubClass(SpGetParent(component), SpTabBox) == SP_TRUE) {
            spDebug(10, "spPrimitiveSetSizeArch", "tab child: height = %d\n", height);
        } else if (spIsSubClass(component, SpBox) == SP_TRUE
                   && spIsSubClass(SpGetParent(component), SpFileDialog) == SP_TRUE) {
            spSetWidgetSizeGtk(widget, width, height);
            gtk_widget_queue_draw(widget);
        } else {
            int hoffset = 0, voffset = 0;
#if !GTK_CHECK_VERSION(1,2,0)
            if (spEqClass(component, SpLabel) == SP_TRUE) {
                voffset = 5;
            }
#endif
            
#if !GTK_CHECK_VERSION(3,0,0)
            if (spEqClass(component, SpPushButton) == SP_TRUE
                &&
#if GTK_CHECK_VERSION(2,20,0)
                gtk_widget_get_can_default(SpPrimitiveArch(component).widget)
#else
                GTK_WIDGET_CAN_DEFAULT(SpPrimitiveArch(component).widget)
#endif
                ) {
                int hspace, vspace;

#if GTK_CHECK_VERSION(2,0,0)
                hspace = 0; vspace = 0;
#elif GTK_CHECK_VERSION(1,2,0)
                hspace = GTK_WIDGET(SpPrimitiveArch(component).widget)->style->klass->xthickness + 3;
                vspace = GTK_WIDGET(SpPrimitiveArch(component).widget)->style->klass->ythickness + 3;
#else  /* GTK_CHECK_VERSION(1,2,0) */
                hspace = 5; vspace = 5;
#endif        /* !GTK_CHECK_VERSION(1,2,0) */
                hoffset = -hspace; voffset = -vspace;
                width += 2 * hspace; height += 2 * vspace;
            }
#endif        /* !GTK_CHECK_VERSION(3,0,0) */
            
            spDebug(80, "spPrimitiveSetSizeArch", "%s: x = %d, y = %d, hoffset = %d, voffset = %d\n",
                    SpGetClassName(component), SpComponentPart(component).x, SpComponentPart(component).y, hoffset, voffset);
            if (spIsToolBar(SpGetParent(component)) == SP_FALSE
                && spIsStatusBar(component) == SP_FALSE
                && spIsMenuBar(component) == SP_FALSE) {
                int x, y;

                x = SpComponentPart(component).x + hoffset;
                y = SpComponentPart(component).y + voffset;
                
                if (spIsSubClass(SpGetParent(component), SpContainer) == SP_TRUE
                    && SpParentContainerPart(component).border_on == SP_TRUE) {
                    x -= SpParentComponentPart(component).border_width;
                    y -= SpParentComponentPart(component).border_width;
                }
                
                x = MAX(x, 0);
                y = MAX(y, 0);
                
                spDebug(80, "spPrimitiveSetSizeArch", "set position: (%d, %d), parent widget = %ld, widget = %ld\n",
                        x, y, (long)SpParentPrimitiveArch(component).widget, (long)widget);
                
#if GTK_CHECK_VERSION(3,0,0) && !defined(SP_USE_GTK_GRID)
                if (GTK_IS_FIXED(SpParentPrimitiveArch(component).widget)) {
                    gtk_fixed_move(GTK_FIXED(SpParentPrimitiveArch(component).widget), widget, x, y);
                }
#else
                gtk_widget_set_uposition(widget, x, y);
#endif
            }
            spDebug(80, "spPrimitiveSetSizeArch", "%s: spSetWidgetSizeGtk: width = %d, height = %d\n",
                    SpGetClassName(component), width, height);
            
            spSetWidgetSizeGtk(widget, width, height);
            
            if (spIsComboBox(component) == SP_TRUE) {
#if !GTK_CHECK_VERSION(2,0,0)
                /* set size of combo box children */
                spSetWidgetSizeGtk(GTK_COMBO(SpPrimitiveArch(component).widget)->entry,
                                   width - GTK_COMBO(SpPrimitiveArch(component).widget)->button->allocation.width - 1,
                                   height);
                spSetWidgetSizeGtk(GTK_COMBO(SpPrimitiveArch(component).widget)->button,
                                   -1, height);
#endif
            } else if (spIsCanvas(component) == SP_TRUE) {
                if ((SpCanvasArch(component).h_adjustment != NULL || SpCanvasArch(component).v_adjustment != NULL)) {
                    spUpdateCanvasScrollBarArch(component);
                } else {
                    GtkAllocation allocation;
                    
#if GTK_CHECK_VERSION(2,0,0)
                    gtk_widget_get_allocation(SpPrimitiveArch(component).widget, &allocation);
#else
                    allocation = GTK_WIDGET(SpPrimitiveArch(component).widget)->allocation;
#endif
                    allocation.width = MAX(width, 1);
                    allocation.height = MAX(height, 1);
                    gtk_widget_size_allocate(SpPrimitiveArch(component).widget, &allocation);
                }

                if (spIsVisible(component) == SP_TRUE && SpCanvasPart(component).exposed == SP_TRUE) {
                    spDebug(50, "spPrimitiveSetSizeArch", "call spResizeCanvasCB: client size = %d, %d\n",
                            SpComponentPart(component).client_width, SpComponentPart(component).client_height);
                    spResizeCanvasCB(component, NULL);
                    spDebug(50, "spPrimitiveSetSizeArch", "spResizeCanvasCB done\n");
                }
            } else if (spIsSubClass(component, SpLabel) == SP_TRUE && SpPrimitiveArch(component).top_widget != NULL) {
                spSetWidgetSizeGtk(SpPrimitiveArch(component).top_widget, width, height);
            } else if (spIsSubClass(component, SpStatusBar) == SP_TRUE) {
                if (SpStatusBarArch(component).last_frame != NULL
                    && width - SpStatusBarArch(component).total_size > 1) {
                    spSetWidgetSizeGtk(SpStatusBarArch(component).last_frame,
                                       width - SpStatusBarArch(component).total_size, -1);
                }
            }
        }
#endif        /* !(GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)) */

        spDebug(30, "spPrimitiveSetSizeArch",
                "done: x = %d, y = %d, width = %d, height = %d\n",
                SpComponentPart(component).x, SpComponentPart(component).y, width, height);
        
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool windowGetSize(spComponent component, int *width, int *height)
{
    int h;

    if (SpPrimitiveArch(component).set_width >= 0 && SpPrimitiveArch(component).set_height >= 0) {
        *width = SpPrimitiveArch(component).set_width;
        *height = SpPrimitiveArch(component).set_height;
        return SP_TRUE;
    }

#if 1
    if (SpFrameArch(component).toplevel != NULL) {
        *width = spGetAllocationWidthGtk(SpFrameArch(component).toplevel);
        *height = spGetAllocationHeightGtk(SpFrameArch(component).toplevel);
        spDebug(10, "windowGetSize", "allocation width = %d, height = %d\n", *width, *height);

#if 1
        if (spIsVisible(SpFramePart(component).menu_bar) == SP_TRUE) {
            SpComponentPart(SpFramePart(component).menu_bar).current_height = -1;
            if (spGetSize(SpFramePart(component).menu_bar, NULL, &h) == SP_TRUE) {
                spDebug(10, "windowGetSize", "menu_bar: height = %d\n", h);
                *height -= h;
            }
        }

        if (spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
            SpComponentPart(SpFramePart(component).status_bar).current_height = -1;
            if (spGetSize(SpFramePart(component).status_bar, NULL, &h) == SP_TRUE) {
                spDebug(10, "windowGetSize", "status_bar: height = %d\n", h);
                *height -= h;
            }
        }
#endif
        
        spDebug(10, "windowGetSize", "updated width = %d, height = %d\n", *width, *height);

        return SP_TRUE;
    }
#else
    if (SpPrimitiveArch(component).widget != NULL) {
        *width = spGetAllocationWidthGtk(SpPrimitiveArch(component).widget);
        *height = spGetAllocationHeightGtk(SpPrimitiveArch(component).widget);
        spDebug(10, "windowGetSize", "width = %d, height = %d\n", *width, *height);

        return SP_TRUE;
    }
#endif

    return SP_FALSE;
}

spBool spPrimitiveGetSizeArch(spComponent component, int *width, int *height)
{
    GtkWidget *widget;
    
    spDebug(30, "spPrimitiveGetSizeArch", "in: %s\n", SpGetClassName(component));
    
    if (spIsSubClass(component, SpImage) == SP_TRUE
        || spIsSubClass(component, SpPrintCanvas) == SP_TRUE) {
        *width = SpComponentPart(component).width;
        *height = SpComponentPart(component).height;
        return SP_TRUE;
    } else if (spIsFrame(component) == SP_TRUE) {
        return windowGetSize(component, width, height);
    }

    if (spIsSubClass(component, SpFileDialog) == SP_TRUE) {
        spDebug(10, "spPrimitiveGetSizeArch", "component is SpFileDialog\n");
#if GTK_CHECK_VERSION(3,0,0)
        widget = SpPrimitiveArch(component).widget;
#else
        widget = GTK_FILE_SELECTION(SpPrimitiveArch(component).widget)->main_vbox;
#endif
    } else {
        widget = (SpPrimitiveArch(component).top_widget != NULL ?
                  SpPrimitiveArch(component).top_widget : SpPrimitiveArch(component).widget);
    }
    
    if (widget != NULL) {
        *width = spGetAllocationWidthGtk(widget);
        *height = spGetAllocationHeightGtk(widget);
        spDebug(30, "spPrimitiveGetSizeArch", "original size: width = %d, height = %d\n", *width, *height);

        if (*height <= 1) {
            GtkRequisition requisition;

#if GTK_CHECK_VERSION(3,0,0)
            gtk_widget_get_preferred_size(widget, NULL, &requisition);
            *height = requisition.height; 
#else
            gtk_widget_size_request(widget, &requisition);
            if (SpTopLevelArch(sp_toplevel).menu_proxy_used == SP_FALSE) {
                *height = MAX(requisition.height, SP_DEFAULT_MENU_BAR_HEIGHT); 
            } else {
                /* if requisition.height == 0 (menu bar in ubuntu unity), it's OK (maybe). */
                *height = requisition.height; 
            }
#endif
            spDebug(30, "spPrimitiveGetSizeArch", "requisition size: width = %d, height = %d\n",
                    requisition.width, requisition.height);
        }
        
        if (spIsMenuBar(component) == SP_TRUE || spIsStatusBar(component) == SP_TRUE) {
            if (*width <= 1) {
                *width = SpParentComponentPart(component).current_width;
            }
        }
            
        spDebug(30, "spPrimitiveGetSizeArch", "updated size: width = %d, height = %d\n", *width, *height);
            
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spPrimitiveGetClientSizeArch(spComponent component, int *width, int *height)
{
    GtkWidget *widget;
    
    spDebug(30, "spPrimitiveGetClientSizeArch", "in: %s\n", SpGetClassName(component));

    if (spIsSubClass(component, SpImage) == SP_TRUE
        || spIsSubClass(component, SpPrintCanvas) == SP_TRUE) {
        *width = SpComponentPart(component).width;
        *height = SpComponentPart(component).height;
        return SP_TRUE;
    } else if (spIsFrame(component) == SP_TRUE) {
        return windowGetSize(component, width, height);
    }
        
    if (spIsSubClass(component, SpFileDialog) == SP_TRUE) {
        spDebug(10, "spPrimitiveGetClientSizeArch", "component is SpFileDialog\n");
#if GTK_CHECK_VERSION(3,0,0)
        widget = SpPrimitiveArch(component).widget;
#else
        widget = GTK_FILE_SELECTION(SpPrimitiveArch(component).widget)->main_vbox;
#endif
    } else if (spIsCanvas(component) == SP_TRUE
               && (SpCanvasArch(component).h_adjustment != NULL || SpCanvasArch(component).v_adjustment != NULL)) {
#if GTK_CHECK_VERSION(2,0,0)
        widget = gtk_widget_get_parent(SpPrimitiveArch(component).widget);
#else
        widget = GTK_WIDGET(SpPrimitiveArch(component).widget)->parent;
#endif
    } else {
        widget = SpPrimitiveArch(component).widget;
    }

    if (widget != NULL) {
        *width = spGetAllocationWidthGtk(widget);
        *height = spGetAllocationHeightGtk(widget);
            
        spDebug(30, "spPrimitiveGetClientSizeArch", "width = %d, height = %d\n", *width, *height);
        
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spPrimitiveSetSensitiveArch(spComponent component, spBool flag)
{
    if (SpPrimitiveArch(component).widget != NULL) {
        gtk_widget_set_sensitive(GTK_WIDGET(SpPrimitiveArch(component).widget),
                                 (flag == SP_TRUE ? TRUE : FALSE));

#if GTK_CHECK_VERSION(3,0,0)
#elif !GTK_CHECK_VERSION(1,2,0)
        if (spIsToolItem(component) == SP_TRUE
            && SpPrimitiveArch(component).sub_widget != NULL
            && SpToolItemArch(component).insens_pixmap != NULL) {
            GdkBitmap *bitmap = NULL;

            if (SpParentToolBarArch(component).bitmap != NULL) {
                bitmap = SpToolBarArch(component).bitmap;
            }
            
            if (flag == SP_TRUE) {
                gtk_pixmap_set(GTK_PIXMAP(SpPrimitiveArch(component).sub_widget),
                               SpPrimitiveArch(component).pixmap, bitmap);
            } else {
                gtk_pixmap_set(GTK_PIXMAP(SpPrimitiveArch(component).sub_widget),
                               SpToolItemArch(component).insens_pixmap, bitmap);
            }
        }
#endif

        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spShowToolTipArch(spComponent component)
{
    if (SpPrimitiveArch(component).widget == NULL) return SP_FALSE;
    
#if GTK_CHECK_VERSION(3,0,0)
    gtk_widget_set_tooltip_text(SpPrimitiveArch(component).widget,
                                SpComponentPart(component).description);
#else
    {
        spComponent window;

        window = SpGetWindow(component);

        if (SpFrameArch(window).tooltips == NULL) {
            SpFrameArch(window).tooltips = gtk_tooltips_new();
        }

        gtk_tooltips_set_tip(SpFrameArch(window).tooltips,
                             SpPrimitiveArch(component).widget,
                             SpComponentPart(component).description, NULL);
    }
#endif
    
    return SP_TRUE;
}

#if GTK_CHECK_VERSION(1,2,0) && !GTK_CHECK_VERSION(1,2,10)
static void changeGraphics(void)
{
    spUpdateSystemGraphics();
    return;
}

static void setPixmap(GtkWidget *widget, GdkEvent *event, spComponent component)
{
    if (widget->window != NULL) {
        spDebug(80, "setPixmap", "in\n");
        gdk_window_set_back_pixmap(widget->window, NULL, TRUE);
    }
    return;
}

static void drawBackground(GtkWidget *widget, GdkEvent *event, spComponent component)
{
    if (widget->window != NULL) {
        GtkWidget *parent;
        
        spDebug(80, "drawBackground", "in\n");
        
        gdk_window_set_back_pixmap(widget->window, NULL, TRUE);

        parent = gtk_widget_get_toplevel(widget);
        gtk_paint_flat_box(parent->style,
                           widget->window,
                           GTK_WIDGET_STATE (widget),
                           GTK_SHADOW_NONE,
                           NULL, widget, "background",
                           0, 0, -1, -1);
        
        if (spIsCanvas(component) == SP_TRUE) {
            changeGraphics();
            spRedrawCanvas(component);
        }
    }
    return;
}

/* because GktFixed's paint function always paints the area with gray */
static void fixedDraw(GtkWidget *widget, GdkRectangle *area)
{
    GtkWidget *parent;
    GtkFixedChild *child;
    GdkRectangle child_area;
    GList *children;

    g_return_if_fail(widget != NULL);

    if (GTK_WIDGET_DRAWABLE(widget))
    {
        spDebug(80, "fixedDraw", "in\n");
        parent = gtk_widget_get_toplevel(widget);
        gtk_paint_flat_box(parent->style,
                           widget->window,
                           GTK_WIDGET_STATE(widget),
                           GTK_SHADOW_NONE,
                           area, widget, "background",
                           0, 0, -1, -1);
        
        if (GTK_IS_FIXED(widget)) {
            children = GTK_FIXED(widget)->children;
        } else {
            children = NULL;
        }
        
        while (children)
        {
            child = children->data;
            children = children->next;

            if (GTK_WIDGET_DRAWABLE(child->widget)
                && gtk_widget_intersect(child->widget, area, &child_area)) {
                gtk_widget_draw(child->widget, &child_area);
            }
        }
    }

    return;
}
#endif

void spSetStyleGtk(spComponent component, GtkWidget *widget)
{
#if GTK_CHECK_VERSION(1,2,0) && !GTK_CHECK_VERSION(1,2,10)
    GtkWidgetClass *klass;

    gtk_signal_connect(GTK_OBJECT(widget),
                       "realize", GTK_SIGNAL_FUNC(setPixmap),
                       (gpointer)component);
    gtk_signal_connect(GTK_OBJECT(widget),
                       "style_set", GTK_SIGNAL_FUNC(drawBackground),
                       (gpointer)component);
    
    if (GTK_IS_FIXED(widget)) {
        klass = GTK_WIDGET_CLASS(GTK_OBJECT(widget)->klass);
        klass->draw = fixedDraw;
    }
#endif

    return;
}

void spPrimitiveCreateBorderGtk(spComponent component)
{
    SpComponentPart(component).border_width = 2;
        
    SpPrimitiveArch(component).top_widget = gtk_frame_new(NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(SpPrimitiveArch(component).top_widget),
                              GTK_SHADOW_IN);
    spContainerAddGtk(component, SpPrimitiveArch(component).top_widget);
    if (SpComponentPart(component).visible_flag == SP_TRUE) {
        gtk_widget_show(SpPrimitiveArch(component).top_widget);
    }
    
    return;
}

void spContainerAddGtk(spComponent component, GtkWidget *widget)
{
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)
    
    GtkPositionType position_type;
    GtkWidget *prev_widget = NULL;

    spDebug(80, "spContainerAddGtk", "name = %s, class = %s, spacing_flag = %d\n",
            SpGetName(component), SpGetClassName(component), SpComponentPart(component).spacing_flag);
    
    spUpdateWidgetExpandGtk(component, widget);
    if (spIsSpacingOn(component) == SP_TRUE) {
        spUpdateWidgetMarginGtk(component, widget);
    } else {
        spDebug(80, "spContainerAddGtk", "%s: spacing is not required\n", SpGetClassName(component));
    }

    if (spIsStatusBar(component) == SP_TRUE) {
        position_type = GTK_POS_BOTTOM;
    } else {
        if (SpParentComponentPart(component).orientation == SP_VERTICAL) {
            position_type = GTK_POS_BOTTOM;
        } else {
            position_type = GTK_POS_RIGHT;
        }
    }

    prev_widget = spGetPrevWidgetGtk(component);
    
    if (prev_widget != NULL) {
        gtk_grid_attach_next_to(GTK_GRID(SpParentPrimitiveArch(component).widget),
                                widget,
                                prev_widget, position_type, 1, 1);
    } else {
        gtk_grid_attach(GTK_GRID(SpParentPrimitiveArch(component).widget),
                        widget, 0, 0, 1, 1);
    }
#else
    gtk_container_add(GTK_CONTAINER(SpParentPrimitiveArch(component).widget),
                      widget);
#endif
    
    return;
}

void spPrimitiveContainerAddGtk(spComponent component)
{
    if (SpPrimitiveArch(component).top_widget != NULL) {
#if !GTK_CHECK_VERSION(3,0,0)
        if (spIsCanvas(component) == SP_TRUE
            && (SpCanvasArch(component).h_adjustment != NULL || SpCanvasArch(component).v_adjustment != NULL)) {
            gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(SpPrimitiveArch(component).top_widget),
                                                  SpPrimitiveArch(component).widget);
        } else 
#endif
        {            
            gtk_container_add(GTK_CONTAINER(SpPrimitiveArch(component).top_widget),
                              SpPrimitiveArch(component).widget);
        }
    } else {
        spContainerAddGtk(component, SpPrimitiveArch(component).widget);
    }
    if (SpComponentPart(component).visible_flag == SP_TRUE
        || SpPrimitiveArch(component).top_widget != NULL) {
        gtk_widget_show(SpPrimitiveArch(component).widget);
    }
    
    return;
}

spBool spPrepareEventCallbackArch(spComponent component)
{
    return SP_TRUE;
}

spBool spUnprepareEventCallbackArch(spComponent component)
{
    return SP_TRUE;
}

#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
typedef struct _spEventCallbackDataGtk spEventCallbackDataGtk;

struct _spEventCallbackDataGtk {
    spComponent component;
    spEventType event_type;
    void *call_data;
};

gboolean spEventsCallbackGtk(gpointer data)
{
    spComponent component;
    spEventCallbackDataGtk *event_data = (spEventCallbackDataGtk *)data;

    spDebug(80, "spEventsCallbackGtk", "in\n");
    
    /* required even in 3.0.0+ */
    gdk_threads_enter();
    
    component = event_data->component;
    
    if (spIsPrimitive(component) == SP_TRUE
        && spExistsPrimitiveEventCallback(component) == SP_TRUE) {
        spDebug(80, "spEventsCallbackGtk", "before spPrimitiveCallEventCallback\n");
        spPrimitiveCallEventCallback(component, event_data->event_type, event_data->call_data);
        spDebug(80, "spEventsCallbackGtk", "after spPrimitiveCallEventCallback\n");
    }
    
    /* required even in 3.0.0+ */
    gdk_threads_leave();
    
    spDebug(80, "spEventsCallbackGtk", "done\n");
    
    return TRUE;
}
#else
static spBool spAppendUserEventGtk(spTopLevel toplevel, spComponent component,
                                   spEventType event_type, void *call_data)
{
    spBool flag = SP_FALSE;

    if (spAppendUserEventDefault(&SpTopLevelArch(toplevel).user_event,
                                 component, event_type, call_data) != NULL) {
        flag = SP_TRUE;
    }

    return flag;
}
#endif

spBool spPostUserEventArch(spComponent component, spEventType event_type, void *call_data)
{
    spDebug(80, "spPostUserEventArch", "in\n");
    
#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
    if (g_thread_supported()) {
        spEventCallbackDataGtk *event_data;
        
        event_data = g_malloc(sizeof(spEventCallbackDataGtk));
        event_data->component = component;
        event_data->event_type = event_type;
        event_data->call_data = call_data;
        
        spDebug(80, "spPostUserEventArch", "call g_async_queue_push\n");
        g_async_queue_push(SpTopLevelArch(sp_toplevel).events_queue, event_data);
        spDebug(80, "spPostUserEventArch", "g_async_queue_push done\n");
        g_main_context_wakeup(NULL);
        spDebug(80, "spPostUserEventArch", "done\n");

        return SP_TRUE;
    }
#else
    if (spAppendUserEventGtk(sp_toplevel, component, event_type, call_data) == SP_TRUE) {
        spDebug(80, "spPostUserEventArch", "done\n");
        return SP_TRUE;
    }
#endif
    
    spDebug(80, "spPostUserEventArch", "failed\n");
    
    return SP_FALSE;
}

void *spGetPrimitiveHandleArch(spComponent component)
{
    return SpPrimitiveArch(component).widget;
}
