/*
 *	spProgressBar.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spContainer.h>
#include <sp/spSlider.h>
#include <sp/spGraphics.h>
#include <sp/spDraw.h>
#include <sp/spCanvas.h>
#include <sp/spImage.h>

#include <sp/spProgressBarP.h>

static spParamTable sp_progress_bar_param_tables[] = {
    {SppValue, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spProgressBar, progress_bar.value), "0"},
    {SppMinimum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spProgressBar, progress_bar.minimum), "0"},
    {SppMaximum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spProgressBar, progress_bar.maximum), "100"},
    {SppIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spProgressBar, progress_bar.increment), "10"},
    {SppSegmentedProgressBar, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spProgressBar, progress_bar.segmented_flag), SP_FALSE_STRING},
};

spProgressBarClassRec SpProgressBarClassRec = {
    /* spObjectClassPart */
    {
	SpProgressBar,
	(spObjectClass)&SpPrimitiveClassRec,
	sizeof(spProgressBarRec),
	spArraySize(sp_progress_bar_param_tables),
	sp_progress_bar_param_tables,
	spProgressBarPartInit,
	spProgressBarPartFree,
	SP_FALSE,
	spProgressBarClassPartInit,
	NULL,
	spProgressBarCreate,
	NULL,
	spProgressBarSetParams,
	NULL,
    },
    /* spComponentClassPart */
    {
	SP_TRUE,
	SP_FALSE,
	SP_FALSE,
	NULL,
	NULL,
	spProgressBarSetSize,
	spProgressBarGetSize,
	spProgressBarGetClientSize,
	NULL,
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spPrimitiveClassPart */
    {
	0,
    },
    /* spProgressBarClassPart */
    {
	0,
    },
};

spComponentClass SpProgressBarClass = (spComponentClass)&SpProgressBarClassRec;

void spProgressBarPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpComponentPart(component).spacing_flag = SP_TRUE;
    SpComponentPart(component).margin_width = SP_DEFAULT_MARGIN;
    SpComponentPart(component).margin_height = SP_DEFAULT_MARGIN;
    SpComponentPart(component).spacing = SP_DEFAULT_SPACING;
#ifdef SP_USE_DEFAULT_PROGRESS_BAR
    SpPrimitiveClassRec.component.resize_flag = SP_FALSE;
    SpProgressBarPart(component).canvas = NULL;
    SpProgressBarPart(component).graphics = NULL;
#endif
    
    return;
}

void spProgressBarPartFree(spObject object)
{
    return;
}

void spProgressBarClassPartInit(spObjectClass object_class)
{
#ifdef SP_USE_DEFAULT_PROGRESS_BAR
    spComponentClass component_class = (spComponentClass)object_class;

#if defined(MACOS)
    if (spGetAppearanceVersionMac() >= 0x00000101) {
	return;
    }
#endif
    component_class->component.resize_flag = SP_FALSE;
#endif
    
    return;
}
    
spBool spProgressBarCreate(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spPrimitiveSetDefaultSize(component);
    if (SpComponentPart(component).disable_initial_scaling == SP_FALSE)
	spPrimitiveScaleDefaultSizeArch(component);
    spProgressBarCreateArch(component);

    return SP_TRUE;
}

spBool spProgressBarSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spProgressBarSetParamsArch(component);

    return SP_TRUE;
}

spBool spProgressBarSetSize(spComponent component, int width, int height)
{
#ifdef SP_USE_DEFAULT_PROGRESS_BAR
    if (SpProgressBarPart(component).canvas != NULL) {
	return spSetSize(SpProgressBarPart(component).canvas, width, height);
    }
#endif
    spDebug(80, "spProgressBarSetSize", "width = %d, height = %d\n", width, height);

    return spPrimitiveSetSize(component, width, height);
}

spBool spProgressBarGetSize(spComponent component, int *width, int *height)
{
#ifdef SP_USE_DEFAULT_PROGRESS_BAR
    if (SpProgressBarPart(component).canvas != NULL) {
	return spGetSize(SpProgressBarPart(component).canvas, width, height);
    }
#endif

    return spPrimitiveGetSize(component, width, height);
}

spBool spProgressBarGetClientSize(spComponent component, int *width, int *height)
{
#ifdef SP_USE_DEFAULT_PROGRESS_BAR
    if (SpProgressBarPart(component).canvas != NULL) {
	return spGetClientSize(SpProgressBarPart(component).canvas, width, height);
    }
#endif
    return spPrimitiveGetClientSize(component, width, height);
}

spBool spIsProgressBar(spComponent component)
{
    return spIsSubClass(component, SpProgressBar);
}

spComponent spCreateProgressBar(spComponent parent, const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;

    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spCreateComponentArg(SpProgressBarClass, SpProgressBar, name, parent, args, num_arg);
}

spBool spSetProgressBarValue(spComponent component, int value)
{
    if (spIsProgressBar(component) == SP_FALSE) {
	if (spIsProgressBar(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return SP_FALSE;
	}
    }

    SpProgressBarPart(component).value = value;
    SpProgressBarPart(component).value = MAX(SpProgressBarPart(component).value,
					     SpProgressBarPart(component).minimum);
    SpProgressBarPart(component).value = MIN(SpProgressBarPart(component).value,
					     SpProgressBarPart(component).maximum);
    spProgressBarSetValueArch(component);
    
    return SP_TRUE;
}

spBool spGetProgressBarValue(spComponent component, int *value)
{
    if (spIsProgressBar(component) == SP_FALSE) {
	if (spIsProgressBar(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return SP_FALSE;
	}
    }
    
    if (value != NULL)
	*value = SpProgressBarPart(component).value;
    
    return SP_TRUE;
}

spBool spIncrementProgressBar(spComponent component)
{
    int new_value;
    
    if (spIsProgressBar(component) == SP_FALSE) {
	if (spIsProgressBar(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return SP_FALSE;
	}
    }

    spDebug(50, "spIncrementProgressBar", "value = %d\n", SpProgressBarPart(component).value);
    
    new_value = SpProgressBarPart(component).value + SpProgressBarPart(component).increment;
    new_value = MIN(new_value, SpProgressBarPart(component).maximum);
    
    spDebug(50, "spIncrementProgressBar", "new_value = %d\n", new_value);
    
    spProgressBarIncrementArch(component, new_value);
    
    SpProgressBarPart(component).value = new_value;
    
    return SP_TRUE;
}

#ifdef SP_USE_DEFAULT_PROGRESS_BAR
static spGraphics sp_progress_bar_graphics = NULL;

static spGraphics getProgressBarGraphics(void)
{
    if (sp_progress_bar_graphics == NULL) {
	sp_progress_bar_graphics = spGetSystemGraphics(SP_SYSTEM_COLOR_PROGRESS_BAR);
    }
    
    return sp_progress_bar_graphics;
}

static void drawProgressBar(spComponent component, int value)
{
    int width, height;
    int value_width;
    double range;
    spComponent canvas;
    spGraphics graphics;

    graphics = getProgressBarGraphics();
    canvas = SpProgressBarPart(component).canvas;

    if (graphics != NULL
	&& spGetClientSize(canvas, &width, &height) == SP_TRUE) {
	spDebug(80, "drawProgressBar", "min = %d, max = %d\n",
		SpProgressBarPart(component).minimum, SpProgressBarPart(component).maximum);
	
	range = (double)(SpProgressBarPart(component).maximum - SpProgressBarPart(component).minimum);

	value_width = spRound((double)width * (double)(value - SpProgressBarPart(component).minimum)
			      / range);
	value_width = MIN(value_width, width);
	
	spDebug(80, "drawProgressBar", "value = %d, range = %.0f, value_width = %d\n",
		value, range, value_width);

	if (value_width > 0) {
	    spFillRectangle(canvas, graphics, 0, 0, value_width, height);
	}
	spRefreshCanvas(canvas);
    }
    
    return;
}

static void drawProgressBarCB(spComponent component, spComponent progress_bar)
{
    drawProgressBar(progress_bar, SpProgressBarPart(progress_bar).value);
    
    return;
}

void spProgressBarCreateDefault(spComponent component)
{
    /*SpComponentPart(component).spacing_flag = SP_FALSE;*/
    SpProgressBarPart(component).canvas = spCreateCanvas(SpGetParent(component), "progressBarCanvas",
							 SpComponentPart(component).require_width,
							 SpComponentPart(component).require_height,
							 SppBorderOn, SP_TRUE,
							 SppSpacingOn, SP_TRUE,
							 SppMarginWidth, SpComponentPart(component).margin_width,
							 SppMarginHeight, SpComponentPart(component).margin_height,
							 SppSpacing, SpComponentPart(component).spacing,
							 SppCallbackFunc, drawProgressBarCB,
							 SppCallbackData, component,
							 SppWidth, SpComponentPart(component).width,
							 SppHeight, SpComponentPart(component).height,
							 SppDrawBackground, SP_TRUE,
							 NULL);
    SpComponentPart(component).component = SpProgressBarPart(component).canvas;
    SpGetPropagateComponent(SpProgressBarPart(component).canvas) = component;
    
    return;
}

void spProgressBarSetParamasDefault(spComponent component)
{
    spBool set_flag = SP_FALSE;
    
    if (SpProgressBarPart(SpOldObject(component)).value != SpProgressBarPart(component).value) {
	set_flag = SP_TRUE;
    }
    if (SpProgressBarPart(SpOldObject(component)).minimum != SpProgressBarPart(component).minimum) {
	set_flag = SP_TRUE;
    }
    if (SpProgressBarPart(SpOldObject(component)).maximum != SpProgressBarPart(component).maximum) {
	set_flag = SP_TRUE;
    }
    if (set_flag == SP_TRUE) {
	spDrawImage(SpProgressBarPart(component).canvas);
    }
    
    return;
}

void spProgressBarSetValueDefault(spComponent component)
{
    spDebug(50, "spProgressBarSetValueDefault", "value = %d\n", SpProgressBarPart(component).value);
    spDrawImage(SpProgressBarPart(component).canvas);
    
    return;
}

void spProgressBarIncrementDefault(spComponent component, int new_value)
{
    drawProgressBar(component, new_value);
    
    return;
}
#endif
