/*
 *	spTopLevel.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spString.h>
#include <sp/spMemory.h>
#include <sp/spThread.h>
#include <sp/spPlugin.h>
#include <sp/spDrawingPlugin.h>

#include <sp/spLocaleP.h>
#include <sp/spTopLevelP.h>
#include <sp/spComponentP.h>

spTopLevel sp_toplevel = NULL;

static spParamTable sp_toplevel_param_tables[] = {
    {SppLanguage, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.language), NULL},
    {SppIconName, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.icon_name), NULL},
    {SppInformation, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.information), NULL},
    {SppAltCtrlSwap, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.alt_ctrl_swap), NULL},
    {SppExitPrompt, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.exit_prompt), NULL},
    {SppClosePrompt, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.close_prompt), NULL},
    {SppModifiedClosePrompt, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.modify_prompt), NULL},
    {SppThreadSafe, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.thread_safe), NULL},
    {SppUseWindowMenu, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.use_window_menu), NULL},
    {SppUniqueApplicationId, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.unique_id), NULL},
    {SppApplicationDomainId, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.domain_id), NULL},
    {SppHelpPath, SP_CREATE_ACCESS | SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.help_path), NULL},
    
    {SppTopLevelCallbackFunc, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.call_func), NULL},
    {SppTopLevelCallbackData, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spTopLevel, toplevel.call_data), NULL},

    {SppPermissions, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.permissions), "0"},
    {SppAcceptedPermissions, SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.accepted_permissions), "0"},
    {SppDisabledPermissions, SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.disabled_permissions), "0"},
    {SppAskedPermissions, SP_GET_ACCESS,
         spOffset(spTopLevel, toplevel.asked_permissions), "0"},
};

spTopLevelClassRec SpTopLevelClassRec = {
    /* spObjectClassPart */
    {
	SpTopLevel,
	NULL,
	sizeof(spTopLevelRec),
	spArraySize(sp_toplevel_param_tables),
	sp_toplevel_param_tables,
	spTopLevelPartInit,
	spTopLevelPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spTopLevelCreate,
	spTopLevelDestroy,
	spTopLevelSetParams,
	NULL,
    },
    /* spTopLevelClassPart */
    {
	0,
    },
};

spObjectClass SpTopLevelClass = (spObjectClass)&SpTopLevelClassRec;

void spTopLevelPartInit(spObject object)
{
    spTopLevel toplevel = (spTopLevel)object;
    
    spDebug(60, "spTopLevelPartInit", "in\n");
    SpTopLevelPart(toplevel).argcp = NULL;
    SpTopLevelPart(toplevel).argvp = NULL;
    SpTopLevelPart(toplevel).graphics = NULL;
    SpTopLevelPart(toplevel).last_graphics = NULL;

    SpTopLevelPart(toplevel).alt_ctrl_swap = SP_FALSE;
    SpTopLevelPart(toplevel).exit_prompt = SP_TRUE;
    SpTopLevelPart(toplevel).close_prompt = SP_FALSE;
    SpTopLevelPart(toplevel).modify_prompt = SP_TRUE;
    SpTopLevelPart(toplevel).thread_safe = SP_FALSE;
    SpTopLevelPart(toplevel).use_window_menu = SP_FALSE;
    SpTopLevelPart(toplevel).language = NULL;
    SpTopLevelPart(toplevel).icon_name = NULL;
    SpTopLevelPart(toplevel).information = NULL;
    SpTopLevelPart(toplevel).unique_id = NULL;
    SpTopLevelPart(toplevel).domain_id = NULL;
    SpTopLevelPart(toplevel).help_path = NULL;
    
    SpTopLevelPart(toplevel).call_func = NULL;
    SpTopLevelPart(toplevel).call_data = NULL;
    
    spTopLevelPartInitArch(toplevel);

    return;
}

void spTopLevelPartFree(spObject object)
{
    spTopLevel toplevel = (spTopLevel)object;
    
    spDebug(50, "spTopLevelFree", "in\n");
    
    spTopLevelPartFreeArch(toplevel);
    sp_toplevel = NULL;

    spDebug(50, "spTopLevelFree", "done\n");
    
    return;
}

spBool spTopLevelCreate(spObject object)
{
    spTopLevel toplevel = (spTopLevel)object;

    spDebug(50, "spTopLevelCreate", "in\n");
    
    if (toplevel != NULL && spIsObjectCreated(object) == SP_FALSE) {
	spSetMessageFlag(0);
	spSetLanguage(SpTopLevelPart(toplevel).language);
	spTopLevelCreateArch(toplevel);
	spSetPluginLang(spGetLanguage());
	spAddDefaultDrawingPlugin();
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spTopLevelSetParams(spObject object)
{
    spTopLevel toplevel = (spTopLevel)object;

    if (toplevel != NULL) {
	spTopLevelSetParamsArch(toplevel);
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spTopLevelDestroy(spObject object)
{
    spTopLevel toplevel = (spTopLevel)object;

    if (toplevel != NULL) {
	spTopLevelDestroyArch(toplevel);
	return SP_TRUE;
    }

    return SP_FALSE;
}

int spWaitNextEvent(spTopLevel toplevel)
{
    if (toplevel == NULL) return 0;

    return spWaitNextEventArch(toplevel, SP_TRUE);
}

int spDispatchEvent(spTopLevel toplevel)
{
    if (toplevel == NULL) return 0;

    return spDispatchEventArch(toplevel, SP_TRUE);
}

int spMainLoop(spTopLevel toplevel)
{
    if (toplevel == NULL) spError(1, "You should call spInitialize.\n");

    spDebug(80, "spMainLoop", "in\n");
    
    spSetInitSenseLevel();
    
    return spMainLoopArch(toplevel);
}

static char sp_default_name[SP_MAX_LINE] = "Application";

spTopLevel spInitializeArg(int *argcp, char ***argvp, spArg *args, int num_arg)
{
    char **argv;
    char *progname;

    if (argcp == NULL || argvp == NULL || *argcp <= 0) {
	progname = sp_default_name;
    } else {
	argv = *argvp;
	progname = spGetBaseName(argv[0]);
#if 0
	{
	    int i;
	    for (i = 0; i < *argcp; i++) {
		spDebug(80, "spInitializeArg", "argv[%d] = %s\n", i, argv[i]);
	    }
	}
#endif
    }

    if ((sp_toplevel = (spTopLevel)spAllocObject((spObjectClass)SpTopLevelClass,
						 NULL, progname)) == NULL) {
	return NULL;
    }

    spDebug(60, "spInitializeArg", "spAllocObject done\n");
    
    SpTopLevelPart(sp_toplevel).argcp = argcp;
    SpTopLevelPart(sp_toplevel).argvp = argvp;

    spSetObjectParamsArg((spObject)sp_toplevel, args, num_arg);

    return sp_toplevel;
}

spTopLevel spInitialize(int *argcp, char ***argvp, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    va_start(argp, argvp);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spInitializeArg(argcp, argvp, args, num_arg);
}

spBool spSetTopLevelParamsArg(spTopLevel toplevel, spArg *args, int num_arg)
{
    return spSetObjectParamsArg((spObject)toplevel, args, num_arg);
}

spBool spGetTopLevelParamsArg(spTopLevel toplevel, spArg *args, int num_arg)
{
    return spGetObjectParamsArg((spObject)toplevel, args, num_arg);
}
    
spBool spSetTopLevelParams(spTopLevel toplevel, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (toplevel == NULL) return SP_FALSE;
    
    va_start(argp, toplevel);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spSetTopLevelParamsArg(toplevel, args, num_arg);
}
    
spBool spGetTopLevelParams(spTopLevel toplevel, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (toplevel == NULL) return SP_FALSE;
    
    va_start(argp, toplevel);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spGetTopLevelParamsArg(toplevel, args, num_arg);
}

void *spGetTopLevelInternalCallbackData(spTopLevel toplevel)
{
    if (toplevel != NULL) {
	return SpTopLevelPart(toplevel).call_data;
    } else {
	return NULL;
    }
}

void spQuit(int status)
{
    spDebug(50, "spQuit", "in\n");
    
    spDestroyAllWindow();
    spRemoveDefaultDrawingPlugin();
    spSearchDrawingPluginFile(-1, SP_DRAWING_DEVICE_NONE);
    spQuitArch(status);
    spFreeStringTable();
    spExitArch(status);
    
    spDebug(50, "spQuit", "done\n");
    
    return;
}

spBool spIsTopLevelCreated(void)
{
    if (sp_toplevel == NULL) {
	return SP_FALSE;
    }
    
    return spIsObjectCreated((spObject)sp_toplevel);
}

spBool spIsThreadSafe(void)
{
    if (sp_toplevel == NULL) return SP_FALSE;
    
    return SpTopLevelPart(sp_toplevel).thread_safe;
}

void spThreadEnter(void)
{
    if (sp_toplevel == NULL) return;
    
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) {
	spThreadEnterArch(sp_toplevel);
    }
    return;
}

void spThreadLeave(void)
{
    if (sp_toplevel == NULL) return;
    
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) {
	spThreadLeaveArch(sp_toplevel);
    }
    return;
}

spBool spSuspendMainLock(spBool flag)
{
    if (flag == SP_FALSE || sp_toplevel == NULL) return SP_FALSE;
    
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) {
	return spSuspendMainLockArch(sp_toplevel);
    } else {
	return SP_FALSE;
    }
}

spBool spResumeMainLock(spBool flag)
{
    if (flag == SP_FALSE || sp_toplevel == NULL) return SP_FALSE;
    
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) {
	return spResumeMainLockArch(sp_toplevel);
    } else {
	return SP_FALSE;
    }
}


#define SP_WAIT_THREAD_LOOP_MAX_WAIT_COUNT 16

typedef struct _spWaitThreadLoopData {
    spTopLevel toplevel;
    void *handle;
    volatile spBool quit_flag;
} spWaitThreadLoopData;

spThreadReturn waitThreadFunc(void *data)
{
    long k;
    long ret;
    spWaitThreadLoopData *wtldata = (spWaitThreadLoopData *)data;

    spDebug(80, "waitThreadFunc", "in\n");

    for (k = 0; k < SP_WAIT_THREAD_LOOP_MAX_WAIT_COUNT; k++) {
	ret = spWaitThread(wtldata->handle);
	spDebug(80, "waitThreadFunc", "k = %ld, spWaitThread ret = %ld\n", k, ret);
	if (ret == SP_THREAD_EXIT_SUCCESS
	    || ret == SP_THREAD_EXIT_FAILURE
	    || ret == SP_THREAD_EXIT_CANCELED) {
	    break;
	}
#if defined(MACOS9)
        spYieldThread();
#else
	spMSleep(1);
#endif
    }

    wtldata->quit_flag = SP_TRUE;
    
    spWaitThreadPostDummyEventArch(wtldata->toplevel);

    spDebug(80, "waitThreadFunc", "done\n");
    
    spYieldThread();
    
    return SP_THREAD_RETURN_SUCCESS;
}

spWaitThreadLoopData sp_wtldata = {NULL, NULL, SP_FALSE};

long spWaitThreadLoopDefault(spTopLevel toplevel, void *handle, spBool dispatch_flag)
{
    void *thread;
    spBool suspended_flag;

    spDebug(100, "spWaitThreadLoopDefault", "dispatch_flag = %d\n", dispatch_flag);
    
    sp_wtldata.toplevel = toplevel;
    sp_wtldata.handle = handle;
    sp_wtldata.quit_flag = SP_FALSE;
    if ((thread = spCreateThread(0, 0, waitThreadFunc, (void *)&sp_wtldata)) == NULL) {
        spDebug(10, "spWaitThreadLoopDefault", "spCreateThread failed\n");
	return SP_THREAD_EXIT_FAILURE;
    }

    spDebug(80, "spWaitThreadLoopDefault", "spCreateThread done\n");

    suspended_flag = spSuspendMainLock(SP_TRUE);
    
    while (sp_wtldata.quit_flag == SP_FALSE) {
	spDebug(100, "spWaitThreadLoopDefault", "in loop\n");
	if (dispatch_flag == SP_TRUE) {
	    if (spDispatchEventArch(toplevel, SP_TRUE) == 0) break;
	} else {
	    if (spWaitNextEventArch(toplevel, SP_TRUE) == 0) break;
	}
	spYieldThread();
    }

    spDebug(80, "spWaitThreadLoopDefault", "loop finished\n");

    spWaitThread(thread);
    spDestroyThread(thread);
    spDebug(80, "spWaitThreadLoopDefault", "before spResumeMainLock: suspended_flag = %d\n", suspended_flag);
    spResumeMainLock(suspended_flag);
    
    spDebug(80, "spWaitThreadLoopDefault", "done\n");
    
    return 0;
}

long spWaitThreadLoop(spTopLevel toplevel, void *handle)
{
    if (toplevel == NULL || handle == NULL) {
	spDebug(20, "spWaitThreadLoop", "input arg is invalid\n");
	return SP_THREAD_EXIT_FAILURE;
    }

    return spWaitThreadLoopArch(toplevel, handle);
}

spBool spIsPostEventSupported(spTopLevel toplevel)
{
    if (toplevel == NULL) return SP_FALSE;

    return spIsPostEventSupportedArch(toplevel);
}

unsigned long spGetAcceptedPermissions(spTopLevel toplevel)
{
    if (toplevel == NULL) return 0L;
    
    return SpTopLevelPart(toplevel).accepted_permissions;
}

unsigned long spGetDisabledPermissions(spTopLevel toplevel)
{
    if (toplevel == NULL) return 0L;
    
    return SpTopLevelPart(toplevel).disabled_permissions;
}

spBool spDemandPermissionRequest(spTopLevel toplevel, unsigned long permissions)
{
    if (toplevel == NULL) return SP_FALSE;

#if !defined(ANDROID)
    return SP_TRUE;
#else
    if (SpTopLevelPart(toplevel).call_func == NULL) return SP_FALSE;

    permissions &= ~SpTopLevelPart(toplevel).accepted_permissions;

    if (permissions == 0L) return SP_TRUE;

    return spDemandPermissionRequestArch(toplevel, permissions);
#endif
}

spBool spIsPermissionRequestRequired(spTopLevel toplevel, unsigned long permissions)
{
    if (toplevel == NULL) return SP_FALSE;

#if !defined(ANDROID)
    return SP_FALSE;
#else
    if (SpTopLevelPart(toplevel).call_func == NULL) return SP_FALSE;

    permissions &= ~SpTopLevelPart(toplevel).accepted_permissions;
    
    if (permissions == 0L) {
        return SP_FALSE;
    } else {
        return spIsPermissionRequestRequiredArch(toplevel, permissions);
    }
#endif
}

spBool spGetHelpTopDirDefault(char *dir, int dir_size, char separator)
{
    const char *lang;
    char id[SP_MAX_PATHNAME];
    char path[SP_MAX_PATHNAME];
    char buf[SP_MAX_PATHNAME];

    if (spIsExactName(SpTopLevelPart(sp_toplevel).help_path) == SP_TRUE) {
	spStrCopy(path, sizeof(path), SpTopLevelPart(sp_toplevel).help_path);
    } else {
	sprintf(path, "%s%c%s", spGetApplicationDocDir(), separator,
		SpTopLevelPart(sp_toplevel).help_path);
    }
    spDebug(10, "spGetHelpTopDirDefault", "path = %s\n", path);

    if (spIsDir(path) == SP_FALSE) {
	return SP_FALSE;
    } else {
	lang = spGetLanguage();
	spDebug(10, "spGetHelpTopDirDefault", "lang = %s\n", lang);
    
	if (spGetLanguageId(lang, id, sizeof(id)) == SP_TRUE) {
	    spDebug(10, "spGetHelpTopDirDefault", "lang id = %s\n", id);
	}
	sprintf(buf, "%s%c%s", path, separator, id);
	spDebug(10, "spGetHelpTopDirDefault", "buf = %s\n", buf);
	if (spIsDir(buf) == SP_FALSE) {
	    spStrCopy(dir, dir_size, path);
	} else {
	    spStrCopy(dir, dir_size, buf);
	}
    }
    spDebug(10, "spGetHelpTopDirDefault", "dir = %s\n", dir);

    return SP_TRUE;
}

spBool spGetHelpFileDefault(const char *dir, const char *path, char *file, int file_size, char separator)
{
    if (file == NULL || file_size <= 1 || strnone(path)) {
	return SP_FALSE;
    }
    
    if (spIsExactName(path) == SP_TRUE) {
	spStrCopy(file, file_size, path);
    } else {
	char buf[SP_MAX_PATHNAME];
	
	if (strnone(dir)) {
	    return SP_FALSE;
	}
	sprintf(buf, "%s%c%s", dir, separator, path);
	spStrCopy(file, file_size, buf);
    }

    return SP_TRUE;
}

int spScaleX(int x)
{
#if defined(_WIN32) && !(defined(USE_MOTIF) || defined(GTK))
    return spScaleXWin(sp_toplevel, x);
#else
    return x;
#endif
}

int spScaleY(int y)
{
#if defined(_WIN32) && !(defined(USE_MOTIF) || defined(GTK))
    return spScaleYWin(sp_toplevel, y);
#else
    return y;
#endif
}

int spUnscaleX(int x)
{
#if defined(_WIN32) && !(defined(USE_MOTIF) || defined(GTK))
    return spUnscaleXWin(sp_toplevel, x);
#else
    return x;
#endif
}

int spUnscaleY(int y)
{
#if defined(_WIN32) && !(defined(USE_MOTIF) || defined(GTK))
    return spUnscaleYWin(sp_toplevel, y);
#else
    return y;
#endif
}
