/*
 *	spTopLevel_Xm.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <pwd.h>
#include <X11/Xlib.h>
#include <X11/Xlocale.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/Xm.h>
#include <Xm/MainW.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spLocaleP.h>
#include <sp/spComponentP.h>
#include <sp/spTopLevelP.h>
#include <sp/spGraphicsP.h>
#include <sp/spPrimitiveP.h>

#undef Status			/* to prevent from conflict of 'Status' */
#include <sp/spThread.h>

static char *sp_def_resources[] = {
    "*XmText.translations: #override \
        Ctrl<Key>f: forward-character()\\n \
        Ctrl<Key>b: backward-character()\\n \
        Ctrl<Key>a: beginning-of-line()\\n \
        Ctrl<Key>e: end-of-line()\\n \
        Ctrl<Key>n: next-line()\\n \
        Ctrl<Key>p: previous-line()\\n \
        Ctrl<Key>k: delete-to-end-of-line()\\n \
        Ctrl<Key>h: delete-previous-character()\\n \
        Ctrl<Key>d: delete-next-character()\\n \
        <Btn4Down>: previous-page()\\n \
        <Btn5Down>: next-page()",
    "*XmTextField.translations: #override \
        Ctrl<Key>f: forward-character()\\n \
        Ctrl<Key>b: backward-character()\\n \
        Ctrl<Key>a: beginning-of-line()\\n \
        Ctrl<Key>e: end-of-line()\\n \
        Ctrl<Key>k: delete-to-end-of-line()\\n \
        Ctrl<Key>h: delete-previous-character()\\n \
        Ctrl<Key>d: delete-next-character()",
    "*XmCSText.translations: #override \
        Ctrl<Key>f: forward-character()\\n \
        Ctrl<Key>b: backward-character()\\n \
        Ctrl<Key>a: beginning-of-line()\\n \
        Ctrl<Key>e: end-of-line()\\n \
        Ctrl<Key>n: next-line()\\n \
        Ctrl<Key>p: previous-line()\\n \
        Ctrl<Key>k: delete-to-end-of-line()\\n \
        Ctrl<Key>h: delete-previous-character()\\n \
        Ctrl<Key>d: delete-next-character()\\n \
        <Btn4Down>: previous-page()\\n \
        <Btn5Down>: next-page()",
    "*XmList.translations: #override \
        <Btn4Down>: ListPrevPage()\\n \
        <Btn5Down>: ListNextPage()",
    "*ComboPopup*XmList.translations: #override \
        <Btn4Down>: ListPrevPage()\\n \
        <Btn5Down>: ListNextPage()\\n \
        <Btn2Down>:",
    "*XmText.background: #ffffff",
    "*XmTextField.background: #ffffff",
    "*XmCSText.background: #ffffff",
    "*XmScrolledWindow.background: #c0c0c0",
    "*XmList.background: #ffffff",
    "*background: #c0c0c0",
#if 0
    "*selectionName.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmLabelGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmLabel.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmCascadeButtonGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmCascadeButton.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmPushButtonGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmPushButton.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmToggleButtonGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*XmToggleButton.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-fixed-medium-r-normal--14-*:",
    "*fontList: -*-fixed-medium-r-normal--14-*:",
#else
    "*selectionName.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmLabelGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmLabel.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmCascadeButtonGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmCascadeButton.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmPushButtonGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmPushButton.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmToggleButtonGadget.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*XmToggleButton.fontList: -*-helvetica-medium-r-normal--14-*-iso8859-1;-*-*-medium-r-normal--14-*:",
    "*fontList: -*-*-medium-r-normal--14-*:",
#endif
    "*XmScrollBar*minimum: 0",
    "*XmScrollBar*maximum: 100",
    "*XmScrollBar*sliderSize: 100",
    /* if kinput2 version is old, uncomment following part */
#ifdef USE_OLD_KINPUT2
#if defined(linux) && defined(MOTIF_STATIC)
    "*preeditType: none",
#endif
#endif
    NULL,
};

static char *sp_topname = NULL;
static char **sp_resources = NULL;
static long sp_num_resource = 0;

static void setResources(char **resources)
{
    long i;
    long num;
    long offset;

    if (resources == NULL) return;
    
    for (num = 0;; num++) {
	if (resources[num] == NULL) {
	    break;
	}
    }
    offset = sp_num_resource;
    sp_num_resource += num;
    
    sp_resources = xrealloc(sp_resources, sp_num_resource + 1, char *);
    
    for (i = 0; i <= num; i++) {
	sp_resources[i + offset] = resources[i];
    }
    
    spDebug(50, "setResources", "done\n");
    
    return;
}

void spSetResources(char **resources)
{
    if (sp_resources == NULL) {
	setResources(sp_def_resources);
    }
    setResources(resources);
    
    return;
}

void spTopLevelPartInitArch(spTopLevel toplevel)
{
    SpTopLevelArch(toplevel).widget = NULL;
    SpTopLevelArch(toplevel).display = NULL;
    SpTopLevelArch(toplevel).screen = NULL;
    SpTopLevelArch(toplevel).resources = NULL;

    SpTopLevelArch(toplevel).user_event_atom = None;
    SpTopLevelArch(toplevel).main_mutex = NULL;
    SpTopLevelArch(toplevel).main_thread_lock_flag = SP_FALSE;
    SpTopLevelArch(toplevel).main_thread_lock_mutex = NULL;
    
    return;
}

void spTopLevelPartFreeArch(spTopLevel toplevel)
{
    return;
}

void spTopLevelCreateArch(spTopLevel toplevel)
{
    XGCValues values;
    char **miss, *def;
    int n_miss;

#if 1
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	if (!XtToolkitThreadInitialize() || XInitThreads() == 0) {
	    SpTopLevelPart(toplevel).thread_safe = SP_FALSE;
	} else {
	    SpTopLevelArch(toplevel).main_mutex = spCreateMutex(NULL);
	    
	    SpTopLevelArch(toplevel).main_thread_lock_flag = SP_FALSE;
	    SpTopLevelArch(toplevel).main_thread_lock_mutex = spCreateMutex(NULL);
	}
    }
#else
    SpTopLevelPart(toplevel).thread_safe = SP_FALSE;
#endif
#else
    /* The main loop in a thread seems not to work correctly,
       so I decided not to support threads for Motif. */
    SpTopLevelPart(toplevel).thread_safe = SP_FALSE;
#endif
    
    spSetResources(SpTopLevelArch(toplevel).resources);
    
    /* initialize XToolkit */
    SpTopLevelArch(toplevel).widget =
	XtAppInitialize(&(SpTopLevelArch(toplevel).app_context), sp_topname,
			NULL, 0,
			SpTopLevelPart(toplevel).argcp,
			*(SpTopLevelPart(toplevel).argvp),
			sp_resources,
			NULL, 0);
    SpTopLevelArch(toplevel).display = XtDisplay(SpTopLevelArch(toplevel).widget);
    SpTopLevelArch(toplevel).screen = XtScreen(SpTopLevelArch(toplevel).widget);
    SpTopLevelArch(toplevel).colormap =
	DefaultColormapOfScreen(SpTopLevelArch(toplevel).screen);
    
    spDebug(50, "spTopLevelCreateArch", "initialize done\n");
    
    SpTopLevelArch(toplevel).font_set =
	XCreateFontSet(SpTopLevelArch(toplevel).display,
		       /*"-*-fixed-medium-r-normal--14-*"*/"-*-*-medium-r-normal--14-*",
		       &miss, &n_miss, &def);
    if (SpTopLevelArch(toplevel).font_set == NULL) {
	SpTopLevelArch(toplevel).font_set =
	    XCreateFontSet(SpTopLevelArch(toplevel).display,
			   "-*-*-*-*-*--*-*",
			   &miss, &n_miss, &def);
    }
    spDebug(50, "spTopLevelCreateArch", "create font done\n");
    
    values.foreground = BlackPixelOfScreen(SpTopLevelArch(toplevel).screen);
    values.background = WhitePixelOfScreen(SpTopLevelArch(toplevel).screen);
    SpTopLevelArch(toplevel).fg_gc =
	XCreateGC(SpTopLevelArch(toplevel).display,
		  RootWindowOfScreen(SpTopLevelArch(toplevel).screen),
		  GCForeground | GCBackground, &values);
    /*values.foreground = WhitePixelOfScreen(SpTopLevelArch(toplevel).screen);*/
    values.foreground = spGetColorPixelXm("#c0c0c0");
    values.background = BlackPixelOfScreen(SpTopLevelArch(toplevel).screen);
    SpTopLevelArch(toplevel).bg_gc =
	XCreateGC(SpTopLevelArch(toplevel).display,
		  RootWindowOfScreen(SpTopLevelArch(toplevel).screen),
		  GCForeground | GCBackground, &values);
    spDebug(50, "spTopLevelCreateArch", "setup gc done\n");

    SpTopLevelArch(toplevel).user_event_atom
	= XInternAtom(SpTopLevelArch(toplevel).display, "_SP_USER_EVENT", False);
    
    return;
}

void spTopLevelSetParamsArch(spTopLevel toplevel)
{
    return;
}
    
void spTopLevelDestroyArch(spTopLevel toplevel)
{
    return;
}

void spQuitArch(int status)
{
    return;
}

void spExitArch(int status)
{
    spExit(status);
    return;
}

spBool spLockMainMutexXm(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spDebug(100, "spLockMainMutexXm", "waiting...\n");
	XFlush(SpTopLevelArch(toplevel).display);
	spLockMutex(SpTopLevelArch(toplevel).main_mutex);
	spDebug(100, "spLockMainMutexXm", "++++locked++++\n");
	
	return SP_TRUE;
    }
#endif

    return SP_FALSE;
}

spBool spUnlockMainMutexXm(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	XFlush(SpTopLevelArch(toplevel).display);
	spUnlockMutex(SpTopLevelArch(toplevel).main_mutex);
	spDebug(100, "spUnlockMainMutexXm", "----unlocked----\n");
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

spBool spLockMainThreadMutexXm(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spBool flag = SP_FALSE;
	
	spLockMutex(SpTopLevelArch(toplevel).main_thread_lock_mutex);
	if (SpTopLevelArch(toplevel).main_thread_lock_flag == SP_FALSE) {
	    spLockMainMutexXm(toplevel);
	    SpTopLevelArch(toplevel).main_thread_lock_flag = SP_TRUE;
	    flag = SP_TRUE;
	}
	spUnlockMutex(SpTopLevelArch(toplevel).main_thread_lock_mutex);

	return flag;
    }
#endif
    
    return SP_FALSE;
}

spBool spUnlockMainThreadMutexXm(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spBool flag = SP_FALSE;
	
	spLockMutex(SpTopLevelArch(toplevel).main_thread_lock_mutex);
	if (SpTopLevelArch(toplevel).main_thread_lock_flag == SP_TRUE) {
	    SpTopLevelArch(toplevel).main_thread_lock_flag = SP_FALSE;
	    spUnlockMainMutexXm(toplevel);
	    flag = SP_TRUE;
	}
	spUnlockMutex(SpTopLevelArch(toplevel).main_thread_lock_mutex);

	return flag;
    }
#endif

    return SP_FALSE;
}

int spWaitNextEventArch(spTopLevel toplevel, spBool lock_flag)
{
    XEvent event;
    
    spDebug(100, "spWaitNextEventArch", "in\n");
    XtAppNextEvent(SpTopLevelArch(toplevel).app_context, &event);
    spDebug(100, "spWaitNextEventArch", "XtAppNextEvent done\n");
    if (lock_flag == SP_TRUE) spLockMainThreadMutexXm(toplevel);
    XtDispatchEvent(&event);
    if (lock_flag == SP_TRUE) spUnlockMainThreadMutexXm(toplevel);
    spDebug(100, "spWaitNextEventArch", "done\n");
    
    return 1;
}

int spDispatchEventArch(spTopLevel toplevel, spBool lock_flag)
{
    int flag = 0;
    
    while (XtAppPending(SpTopLevelArch(toplevel).app_context)) {
	spWaitNextEventArch(toplevel, lock_flag);
	flag = 1;
    }
    
    return flag;
}

int spMainLoopArch(spTopLevel toplevel)
{
    while (XtAppGetExitFlag(SpTopLevelArch(toplevel).app_context) == False) {
	spDebug(100, "spMainLoop", "call spWaitNextEventArch\n");
	spWaitNextEventArch(toplevel, SP_TRUE);
	spDebug(100, "spMainLoop", "spWaitNextEventArch done\n");
    }
    
    return 0;
}

spBool spThreadEnterArch(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spLockMainMutexXm(toplevel);
	
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

spBool spThreadLeaveArch(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && XmVersion >= 2001
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spUnlockMainMutexXm(toplevel);
	
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

spBool spSuspendMainLockArch(spTopLevel toplevel)
{
    return spUnlockMainThreadMutexXm(toplevel);
}

spBool spResumeMainLockArch(spTopLevel toplevel)
{
    return spLockMainThreadMutexXm(toplevel);
}

spBool spWaitThreadPostDummyEventArch(spTopLevel toplevel)
{
    return SP_TRUE;
}

long spWaitThreadLoopArch(spTopLevel toplevel, void *handle)
{
    return spWaitThreadLoopDefault(toplevel, handle, SP_FALSE);
}

spBool spIsPostEventSupportedArch(spTopLevel toplevel)
{
    return SP_TRUE;
}
